#include <stdio.h>
#include <helpers.h>
#include <wobject.h>

class Circle : public Wobject
{

  public:
      friend Wobject *Circle_process_mouse(int *x = NULL, int *y = NULL);

      // ADD CONSTRUCTORS & DESTRUCTORS IF NEEDED
      Circle(int size,int pos_x,int pos_y);

      Wobject *process_mouse(int *x = NULL,int *y = NULL);
  private:
        // overload inquire_mouse function
      int inquire_mouse(int x, int y);
      void draw(BITMAP *onto,int from_x, int from_y, int to_x, int to_y);


      static int _ident; // identification integer
};


Circle::Circle(int size,int pos_x,int pos_y)
      : Wobject("Circle",size,size,pos_x,pos_y)
{
    see_through = TRUE;
}




void Circle::draw(BITMAP *onto, int f_x, int f_y,int t_x, int t_y)
{
    f_x = f_y =t_x =t_y = 0; //unused
    circlefill(onto,get_geoinfo()->pos_x+get_geoinfo()->screen_size_x/2,get_geoinfo()->pos_y+get_geoinfo()->screen_size_y/2,get_geoinfo()->screen_size_x/2 -1,makecol(255,255,255));
    circle(onto,get_geoinfo()->pos_x+get_geoinfo()->screen_size_x/2,get_geoinfo()->pos_y+get_geoinfo()->screen_size_y/2,get_geoinfo()->screen_size_x/2 -1,makecol(0,0,0));
}


// IDENTIFICATION FUNCTIONS, DO NOT ALTER
int Circle::inquire_mouse(int pos_x,int pos_y)
{
    GeoInfo const *g = get_geoinfo();
    int r = g->size_x/2;
    pos_x -= r;
    pos_y -= r;
    if ((pos_x*pos_x)+(pos_y*pos_y) <= (r*r))
        return TRUE;

    return FALSE;

}




int Circle::_ident = -1;


// NOTE this mouse function doesn't work if the circles are inside a parent window
// because it uses mouse_x and mouse_y directly instead of first transforming
// them to the coordinates of the parent window

Wobject *Circle::process_mouse(int *x_out, int *y_out)
{
    Wobject *win; 
    int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y; // helper variables
    int flag; // is the border clicked?

    //get the object the mouse is currently in( if flag is set it is in a border)
    //out_x, out_y will be the coordinates relative to that object 
    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);

    //store coordinates relative to this object if the user wants them
    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;
    if (!win)
       return NULL;
    if (mouse_b)
    {
        win->raise();
        mouse_was_x = mouse_x;
        mouse_was_y = mouse_y;
        pos_was_x = win->get_geoinfo()->pos_x;
        pos_was_y = win->get_geoinfo()->pos_y;
        while (mouse_b)
            (Circle *)win->move(pos_was_x+mouse_x-mouse_was_x,pos_was_y+mouse_y-mouse_was_y);
    }
    return win;
}


int main()
{
    allegro_init();
    install_timer();
    install_mouse();
    install_keyboard();
    allegro_message("this example demonstrates how to define other than rectangular objects\n"
           "shadow is turned off, because the shadow is always rectangular.\n"
           "press any key...");

    if(set_gfx_mode(GFX_SAFE,800,600,0,0))
       fatal("unable to set 800x600x8 graphics mode\n");
    
    set_palette(desktop_palette);
    show_mouse(screen);
    WinSystem::Buffered = TRUE;
    WinSystem::BackgroundColor = makecol(100,100,100);

    // shadow doesn't work for non-square objects(uncomment and compile
    // to see what's wrong with it
    WinSystem::ShadowStyle = WHAM_NO_SHADOW;

    // UseBoxes makes no difference for non-square objects
    WinSystem::UseBoxes = TRUE;

    Circle *c = new Circle(100,100,100);
    Circle *d = new Circle(200,100,100);
    Circle *e = new Circle(50,100,100);
    c->show();
    d->show();
    e->show();
    while(!key[KEY_ESC])
        Wham_Secretary.process_mouse();
    delete WHAM;
    return 0;
}
END_OF_MAIN();
