/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/

#ifndef _billwin_h_
#define _billwin_h_
#include "helpers.h"
#include "wobject.h"


#define BILLWIN_STATE_MINIMISED 1
#define BILLWIN_STATE_MAXIMISED 2
#define BILLWIN_STATE_MEDIUMISED 3
#define BILLWIN_MENU_SYSBOX     1
#define BILLWIN_MINIM_SYSBOX    2
#define BILLWIN_MAXIM_SYSBOX    3
#define BILLWIN_STYLE_RESCALE   1

class BillWin : public Wobject
{
    public:
      
      // constructor /destructor
        BillWin(int size_x,int size_y,int pos_x,int pos_y,char *name,int flags = 0);
        BillWin(int size_x,int size_y,int pos_x,int pos_y,char *name, WinSystem *its_manager,int flags = 0);
        ~BillWin();
        
        // statics that must be set ONCE at the start of the program
        static int BorderSize;
        // statics that can be changed anytime
        static int Text_Color;
        static int Bar_Color;
        static int Black_Color;
        static int Dark_Color;
        static int Middle_Color;
        static int Light_Color;
        static int White_Color;
        static int Border_Color;
        //info
        int get_state();

        // get various objects
        BITMAP *bmp();     // returns the user area of the object
        // user area
        void set_user_area(int pos_x, int pos_y, int size_x, int size_y);
        //resize etc
        int resize(int nw_size_x, int nw_size_y);

        // extra functions        
        int printf(char *format,...); // prints text on the current cursor position
        void set_icon(BITMAP *icon);
        void set_state(int state);

        // process input /messages
        Wobject *process_mouse(int *x_out = NULL, int *y_out = NULL);

    private:
        
        static int number_of_instances;
        void draw(BITMAP *onto, int x_from, int y_from, int x_to, int y_to); // draws (a part of) the object the coordinates are screen coordinates
       
        static int draw_border(BITMAP *onto, Wobject *the_object,int from_x, int from_y, int to_x, int to_y);

        BITMAP *window;                 // buffer
        BITMAP *user;                   // user area(sub bitmap of window)

        int must_delete_bitmap;
        void create(BITMAP *bmp); // creates a object with bitmap as buffer
        void destroy();            // just does what it says
        int text_cursor_x,text_cursor_y;
        int state;
        int last_state;
        int med_size_x;
        int med_size_y;
        int med_pos_x;
        int med_pos_y;
        void draw_sysbox(int which, int checked); 
        static BITMAP *standard_icon;
        BITMAP *icon;
        int flags;
        void sys_menu();
};



#endif
