/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include "alguicon.h"
#include <signal.h>

static int default_mouse_x() { return mouse_x; }
static int default_mouse_y() { return mouse_y; }

extern BITMAP *_mouse_screen;
int AlGuiContainer::Border_Color = 255;
int AlGuiContainer::Title_Color = 0;
int AlGuiContainer::nr_of_instances = 0;
int AlGuiContainer::Refresh_Rate = 1000;
int whamgui_raise_style = RAISE_IMMEDIATE;


AlGuiContainer *AlGuiContainer::current_guicon = NULL;
AlGuiContainer **AlGuiContainer::guicon_list = NULL;
int AlGuiContainer::guicon_list_size = 0;
int AlGuiContainer::max_guicon_list_size = 0;





void AlGuiContainer::register_guicontainer(AlGuiContainer *guicon)
{

   #ifdef WTRACE
	  if (wham_trace) trace(" AlGuiContainer::%s %p", __PRETTY_FUNCTION__,guicon);
   #endif

   guicon_list_size++;
   if (guicon_list_size >= max_guicon_list_size)
   {
      max_guicon_list_size +=5;
      guicon_list = (AlGuiContainer **)realloc(guicon_list, max_guicon_list_size * sizeof(AlGuiContainer *));
      if (!guicon_list)
         out_of_mem("memory exhausted in GUI administration");
   }
   guicon_list[guicon_list_size-1] = guicon;
}

void AlGuiContainer::unregister_guicontainer(AlGuiContainer *guicon)
{
   #ifdef WTRACE
	  if (wham_trace) trace(" AlGuiContainer::%s %p", __PRETTY_FUNCTION__,guicon);
   #endif

   for (int i=0;i<guicon_list_size;i++)
   {
      if (guicon_list[i] == guicon)
      {
         for (int j=i;j<guicon_list_size-1;j++)
             guicon_list[j] = guicon_list[j+1];
         guicon_list_size--;
         return;
      }
   }
}

GUIPROCTYPE AlGuiContainer::find_proc_for_dialog(DIALOG *d, AlGuiContainer **obj)
{
   #ifdef WTRACE
	  if (wham_trace) trace(" AlGuiContainer::find_proc_for_dialog %p", d);
   #endif

   if (current_guicon && current_guicon->dialog <= d && current_guicon->dialog + current_guicon->dialog_size > d)
   {
      *obj = current_guicon;
//      trace("found in %s", current_guicon->get_geoinfo()->name);
      return current_guicon->original_procs[d - current_guicon->dialog];
   }
      
   // we lost track, this is possible in some complicated multi gui constructions
   // search the list of guicontainers
//   trace("lost track search all");
   for (int i=0;i<guicon_list_size;i++)
   {
      if (d >= guicon_list[i]->dialog && d < (guicon_list[i]->dialog + guicon_list[i]->dialog_size))
      {
         if (obj)
            *obj = guicon_list[i];
   #ifdef WTRACE
	  if (wham_trace)
     {
      trace(" AlGuiContainer::find_proc_for_dialog DONE return %d from %s", d - guicon_list[i]->dialog, guicon_list[i]->get_geoinfo()->name);
      trace("      and that is ...  %p",guicon_list[i]->original_procs[(d -  guicon_list[i]->dialog)]);
     }
   #endif
   
         return guicon_list[i]->original_procs[(d -  guicon_list[i]->dialog)];
      }
   }
   #ifdef WTRACE
	  if (wham_trace) trace(" AlGuiContainer::find_proc_for_dialog DONE, not found");
   #endif

   return NULL;
}

AlGuiContainer::AlGuiContainer(int pos_x,int pos_y, int size_x, int size_y, char *name, DIALOG *d,int _syspriority)
	  : Window(name,pos_x,pos_y,size_x,size_y,AlGuiContainer::draw_border)
{
	#ifdef WTRACE
   	if (wham_trace) trace(" %s (%d) AlGuiContainer::constructor1", get_geoinfo()->name, get_geoinfo()->tag);
   #endif
   register_guicontainer(this);
	create(d, _syspriority);
   in_changing = 0;
}


AlGuiContainer::AlGuiContainer(int pos_x,int pos_y, int size_x, int size_y, char *name, DIALOG *d, int (*_draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y),int _syspriority)
	  : Window(name,pos_x,pos_y,size_x,size_y,_draw_border)
{
	#ifdef WTRACE
   	if (wham_trace) trace(" %s (%d) AlGuiContainer::constructor2 ", get_geoinfo()->name, get_geoinfo()->tag);
   #endif
   register_guicontainer(this);
	create(d, _syspriority);
   in_changing = 0;
}


DIALOG *AlGuiContainer::change_dialog(DIALOG *d)
{
   #ifdef WTRACE
	  if (wham_trace) trace(" %s (%d) AlGuiContainer::change_dialog ", get_geoinfo()->name, get_geoinfo()->tag);
   #endif
	DIALOG *ret = dialog;
   int was_running = it_is_running;
   int was_active = active_dialog == dialog;

//   trace("in_proc %d",in_proc);
   
   if (in_proc)
   {
      screen = WHAM->hw_screen;
      WHAM = NULL;
      allegro_exit();
      fprintf(stderr,"change_dialog() called from within it's own procedure\n");
      exit(1);
   }
   in_changing = TRUE;
//   trace("start destroy");
   stop();
   destroy();
   FONT *ofont = algui_font;
   create(d, syspriority);
   if (was_running)
      start();
   algui_font = ofont;
   
   if (was_active)
      active_dialog = dialog;
   // show changes on screen
   solid_mode();
   update();
   
   in_changing = FALSE;
   return ret;
}


AlGuiContainer::~AlGuiContainer()
{
   #ifdef WTRACE
	  if (wham_trace) trace(" %s (%d) AlGuiContainer::%s ", get_geoinfo()->name, get_geoinfo()->tag,__PRETTY_FUNCTION__);
   #endif
	destroy();
   unregister_guicontainer(this);
}



void AlGuiContainer::create(DIALOG *d, int _syspriority)
{
   #ifdef WTRACE
	  if (wham_trace) trace(" %s (%d) AlGuiContainer::%s ", get_geoinfo()->name, get_geoinfo()->tag,__PRETTY_FUNCTION__);
   #endif
   in_proc = 0;
   algui_font = font;
   syspriority = _syspriority;
	it_is_running = FALSE;
   delete_when_done = FALSE;
	dialog = d;
	magic_number = GUICON_MAGIC_NUMBER;
	clear_to_color(bmp(),gui_bg_color);
	set_lock_scale(TRUE);

   dialog_size= 0;
}


void AlGuiContainer::destroy()
{
   #ifdef WTRACE
	  if (wham_trace) trace(" %s (%d) AlGuiContainer::%s ", get_geoinfo()->name, get_geoinfo()->tag,__PRETTY_FUNCTION__);
   #endif

//   trace("destroy imnproc %d", in_proc);

   if (in_proc)
   {
      if (WHAM->is_being_deleted)
         return;
      screen = WHAM->hw_screen;
      WHAM = NULL;
      allegro_exit();
      fprintf(stderr,"dialog destroyed from within one of its own procedures");
      exit(1);
   }

   if (!in_proc)
   {
//   trace("destroy stop");
   	stop();
   }
   
   
//   trace("destroy cleanup");


}

int AlGuiContainer::d_proc(int msg, DIALOG *d, int c)
{
//  return 0;
   BITMAP *oscreen = screen;


   #ifdef WTRACE
	  if (wham_trace && msg != MSG_IDLE) trace("%s msg %d, d %p, c %d",__PRETTY_FUNCTION__,msg,d,c);
   #endif
   int ret;
   GUIPROCTYPE proc;
   AlGuiContainer *obj;
   

   proc = find_proc_for_dialog(d, &obj);


   #ifdef WTRACE
		if (wham_trace && msg != MSG_IDLE) trace("  %s                           call %d STARTS HERE",obj->get_geoinfo()->name,obj->in_proc);
	#endif
   // call the thing
   
   if (msg == MSG_DRAW)
   {
      oscreen = screen;
      screen = obj->bmp();
   }
   obj->in_proc++;

   ret = proc(msg,d,c);

   obj->in_proc--;
   
   if (msg == MSG_DRAW)
   {
      screen = oscreen;
   	obj->update();
   }
//   	obj->request(W_REQUEST_UPDATE);
      
      
   #ifdef WTRACE
	  if (wham_trace && msg != MSG_IDLE) trace("%s: call %d DONE HERE",obj->get_geoinfo()->name,obj->in_proc);
   #endif
   return ret;

}


void AlGuiContainer::start(int _focus_obj)
{
	#ifdef WTRACE
       if (wham_trace) trace (" %s (%d) AlGuicon : start", get_geoinfo()->name, get_geoinfo()->tag);
	#endif
   if (!dialog) // this one's empty
   	return;
   if (it_is_running)
      return;
      
//      trace("%s setup procs",get_geoinfo()->name);

      for (int i=0;dialog[i].proc;i++)
   	{
         dialog_size++;
	   }
      original_procs = (GUIPROCTYPE *)malloc(dialog_size * sizeof(GUIPROCTYPE));
      if (!original_procs)
         out_of_mem("creating Gui Window");
      
	   for (int i=0;dialog[i].proc;i++)
   	{
         original_procs[i] = dialog[i].proc;
         if (original_procs[i] == AlGuiContainer::d_proc)
            fatal("Fatal error in WHAM dialog management\nendless loop created");
         
         dialog[i].proc = AlGuiContainer::d_proc;
	   }
      
      
//   trace("start in_proc %d",in_proc);
	BITMAP *oscreen = screen;
		screen = bmp();
		diaplayer = init_dialog(dialog,_focus_obj);
      set_clip(screen,0,0,screen->w - 1, screen->h -1);
		screen = oscreen;
      if (syspriority) // don't let others do anything while I'm here
      {

      	raise_and_lock(); // make sure it's visible
         show();
      	WHAM->claim_coordinates(this);
         WHAM->claim_keyboard(this);
     }
     active_dialog = dialog; // make sure this is set correct
     do_dialog(diaplayer); // call AlGuiCon's do dialog once to do initial drawing
	it_is_running = TRUE;
//   trace("started %d", in_proc);
}

int AlGuiContainer::is_running()
{
//   #ifdef WTRACE
//	  if (wham_trace) trace(" %s (%d) AlGuiContainer::%s ", get_geoinfo()->name, get_geoinfo()->tag,__PRETTY_FUNCTION__);
//   #endif
	return it_is_running;
}



int AlGuiContainer::process_keyboard()
{
   #ifdef WTRACE
	  if (wham_trace) trace(" %s (%d) AlGuiContainer::%s ", get_geoinfo()->name, get_geoinfo()->tag,__PRETTY_FUNCTION__);
   #endif
	if (it_is_running)
	{
		if (!(it_is_running = do_dialog(diaplayer))) // this do_dialog is from AlGuiContainer, it installs the mouse position callbacks and calls update_dialog
			shutdown();
	}
	return 0;
}


int AlGuiContainer::process_message(Wobject *from, int msg)
{
   #ifdef WTRACE
//	  if (wham_trace) trace(" %s (%d) AlGuiContainer::%s ", get_geoinfo()->name, get_geoinfo()->tag,__PRETTY_FUNCTION__);
   #endif
//   if (from == this && msg == W_MSG_HIDDEN && it_is_running)
//   	shutdown();
   
	switch(msg)
	{
		case W_MSG_LOSTMOUSE:
			if (it_is_running)
			{
				if (!(it_is_running = do_dialog(diaplayer))) // this do_dialog is from AlGuiContainer, it installs the mouse position callbacks and calls update_dialog
					shutdown();
				}

			break;
	}
	return Window::process_message(from,msg);
}

int AlGuiContainer::get_closer_object()
{
   #ifdef WTRACE
	  if (wham_trace) trace(" %s (%d) AlGuiContainer::%s ", get_geoinfo()->name, get_geoinfo()->tag,__PRETTY_FUNCTION__);
   #endif
	if (!(it_is_running))
		return diaplayer->obj;
	else
		return -1; // is not yet closed
}

void AlGuiContainer::stop()
{
	#ifdef WTRACE
	if (wham_trace) trace (" %s (%d) AlGuicon : stop", get_geoinfo()->name, get_geoinfo()->tag);
	#endif

	if (it_is_running)
   	shutdown();
}

void AlGuiContainer::shutdown()
{
	#ifdef WTRACE
	if (wham_trace) trace (" %s (%d) AlGuicon : shutdown", get_geoinfo()->name, get_geoinfo()->tag);
	#endif
	shutdown_dialog(diaplayer);
	it_is_running = FALSE;

   if (syspriority)
   {
//      trace("shutdown %s: releasing",get_geoinfo()->name);

   	WHAM->release_coordinates(this);
   	WHAM->release_keyboard(this);
      set_lock(FALSE);
      lower();
      request(W_REQUEST_HIDE);
  }
   if (!in_changing) // if in_chaging ==TRUE the dialog is not really stopping
   {
      if (delete_when_done)
      {
            request(W_REQUEST_DELETE);
      }
   }

   if (dialog)
   {
//      trace("%s replace original procs",get_geoinfo()->name);
   	for (int i=0;dialog[i].proc;i++)
   	{
      	dialog[i].proc = original_procs[i];
	   }
      free(original_procs);
   }
//   trace("destroy done");
   dialog_size = 0;
   
}

int AlGuiContainer::draw_border(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y)
{
   #ifdef WTRACE
	  if (wham_trace) trace("AlGuiContainer::%s ",__PRETTY_FUNCTION__);
   #endif
	int
		f_x,f_y,t_x,t_y;
	GeoInfo const *g = the_object->get_geoinfo();
	if (!onto) // return border dimensions to manager
	{
		switch(from_x)
		{
			case WINDOW_UPPER_BORDER:
				return 10;
			break;
			case WINDOW_LOWER_BORDER:
				return 3;
			break;
			case WINDOW_LEFT_BORDER:
				return 3;
			break;
			case WINDOW_RIGHT_BORDER:
				return 3;
			break;
		}

	}
	solid_mode();
	//upper border
	f_x = g->pos_x - 3;
	t_x = g->pos_x + g->screen_size_x + 2;
	f_y = g->pos_y - 10;      
	t_y = g->pos_y;           
							  
	f_x = MAX(f_x, from_x);   
	t_x = MIN(t_x, to_x);     
	f_y = MAX(f_y, from_y);   
	t_y = MIN(t_y, to_y);     
	if (f_x <= t_x && f_y <= t_y) 
		rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
		text_mode(-1);
		textout_centre(onto,font,g->name,g->pos_x + g->screen_size_x / 2, g->pos_y - 9,Title_Color); 
	//lower border                         
	f_x = g->pos_x - 3;                   
	t_x = g->pos_x + g->screen_size_x + 2; 
	f_y = g->pos_y + g->screen_size_y;     
	t_y = g->pos_y + g->screen_size_y + 2; 
											
	f_x = MAX(f_x, from_x);                
	t_x = MIN(t_x, to_x);                  
	f_y = MAX(f_y, from_y);                
	t_y = MIN(t_y, to_y);                  
	if (f_x <= t_x && f_y <= t_y)          
		rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
											
	//left border                          
	f_x = g->pos_x - 3;                   
	t_x = g->pos_x;                        
	f_y = g->pos_y;                        
	t_y = g->pos_y + g->screen_size_y;     
											
	f_x = MAX(f_x, from_x);                
	t_x = MIN(t_x, to_x);                  
	f_y = MAX(f_y, from_y);                
	t_y = MIN(t_y, to_y);                  
	if (f_x <= t_x && f_y <= t_y)          
		rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
											
	//right border                         
	f_x = g->pos_x + g->screen_size_x;     
	t_x = g->pos_x + g->screen_size_x + 2; 
	f_y = g->pos_y;                        
	t_y = g->pos_y + g->screen_size_y;     
											
	f_x = MAX(f_x, from_x);                
	t_x = MIN(t_x, to_x);                  
	f_y = MAX(f_y, from_y);                
	t_y = MIN(t_y, to_y);                  
	if (f_x <= t_x && f_y <= t_y)          
		rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
											
											
	//END OF EXAMPLE CODE
	return 0; //returnvalue currently ignored 
}

Wobject *AlGuiContainer::process_mouse(int *x_out, int *y_out)
{
   #ifdef WTRACE
	  if (wham_trace) trace(" %s (%d) AlGuiContainer::%s ", get_geoinfo()->name, get_geoinfo()->tag,__PRETTY_FUNCTION__);
   #endif
	Wobject *win; 
	int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y,mouse_is_x,mouse_is_y,size_was_x,size_was_y; // helper variables
	int flag; // is the border clicked?
	GeoInfo const *g; // to store geoinfo from object in

	//get the object the mouse is currently in( if flag is set it is in a border)
	//out_x, out_y will be the coordinates relative to that object 
	win = WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);
	//store coordinates relative to this object if the user wants them
	if (x_out)
		*x_out = out_x;
	if (y_out)
		*y_out = out_y;
	if (win != this)  return win;

   g = win->get_geometric_info(); // get info on this object into g





	if (whamgui_raise_style == RAISE_IMMEDIATE) win->raise();
	if (whamgui_raise_style == RAISE_CLICK && mouse_b) 
	{
		win->raise();
	}

	if (!flag && out_x > 0 && out_y > 0 && out_x < g->size_x && out_y < g->size_y) // mouse not on border, so must be inside the window, call dialog functions
	{
		if (it_is_running)
		{
			if (!(it_is_running = do_dialog(diaplayer)))
         {
				shutdown();
         }
		}
	}

   
	if (flag)
	{
		if (out_x <0 || out_y <0)
			set_Mouse_Cursor(WinSystem::Mouse_Cursor1);
		else if (g->scale_is_locked)
			set_Mouse_Cursor(WinSystem::Mouse_Cursor1);
      else
			set_Mouse_Cursor(WinSystem::Mouse_Cursor2);
	}
	else
		set_Mouse_Cursor(WinSystem::Mouse_Cursor0);




	// mouse button 1 is currently pressed down
	if (mouse_b & 1)
	{

		mouse_was_x = mouse_x; //store current mouse position
		mouse_was_y = mouse_y;
		pos_was_x = g->pos_x; //store current object location 
		pos_was_y = g->pos_y;
		size_was_x = g->screen_size_x;//store current object size
		size_was_y = g->screen_size_y;
		
		//if the window has a parent manager, the parent window could be rescaled   
		//so calc the mouse position relative to the parent window
		if (win->get_manager()->get_parent())
			 win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);
		 
		if (flag) // mouse is on a border 
		{
			if (out_x < 0 || out_y <0) // left or upper border -> move
			{
				while (mouse_b & 1)
				{
					mouse_is_x = mouse_x; // get current mouse pos
					mouse_is_y = mouse_y;
					// calc current mouse pos relative to parent win
					if (win->get_manager()->get_parent())
						win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
					// and move the window
					win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);
				}
			}
			else // left or lower border -> rescale, or move if scale_locked
			{
				while (mouse_b & 1)
				{
					mouse_is_x = mouse_x;// get current mouse pos
					mouse_is_y = mouse_y;
					// make mouse pos relative to parent window
					if (win->get_manager()->get_parent())
						win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
					//rescale the object
					if (!g->scale_is_locked)
						win->rescale(size_was_x + mouse_is_x - mouse_was_x,size_was_y + mouse_is_y - mouse_was_y);
					else
						win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);

				}
			}
		}
	}

   /*
	static int counter = 0;
	counter ++;
	if (counter >= Refresh_Rate)
	{
	  win->update();
	  counter = 0;
	}
    */

	if (!flag) 
		return (Wobject *)win;
	else
		return NULL;
}


int AlGuiContainer::g_mouse_x()
{
	int xout,yout,flag;
	Wobject *w;
	w = WHAM->process_coordinates(mouse_x,mouse_y,&xout,&yout,&flag);
	return xout;
}

int AlGuiContainer::g_mouse_y()
{
	int xout,yout,flag;
	Wobject *w;
	w = WHAM->process_coordinates(mouse_x,mouse_y,&xout,&yout,&flag);

	return yout;
}

int AlGuiContainer::do_dialog(DIALOG_PLAYER *dp)
{
	#ifdef WTRACE
	    if (wham_trace) trace (" %s (%d) AlGuicon : do_dialog", get_geoinfo()->name, get_geoinfo()->tag);
	#endif


   FONT *ofont = font;

   font = algui_font;
//	BITMAP *oscreen = screen;
   AlGuiContainer *old_guicon;

	gui_mouse_x = AlGuiContainer::g_mouse_x;
	gui_mouse_y = AlGuiContainer::g_mouse_y;

   solid_mode();
//	screen = bmp();
   active_dialog = dialog; // make sure this is set correct
   old_guicon = current_guicon;
   current_guicon = this;

   if (dialog[0].proc != AlGuiContainer::d_proc)
      trace("warning: dialog changed when it was already started");
      
   int ret = update_dialog(dp);

   current_guicon = old_guicon;
//	screen = oscreen;

   
	gui_mouse_y = default_mouse_y;
	gui_mouse_x = default_mouse_x;
   font = ofont;

	return ret;
}

static int pushdepth = 0;
static BITMAP *oscreen;

void AlGuiContainer::push_gui_screen()
{
	if (!pushdepth)
   {
   	oscreen = screen;
   }
	pushdepth++;
   screen = WHAM->hw_screen;
}


void AlGuiContainer::pop_gui_screen()
{
	pushdepth--;
   if (pushdepth <0)
   {
      delete WHAM;
      WHAM = NULL;
      allegro_exit();
      ::printf("unpaired pop_gui_screen\n");
      ::raise(SIGFPE);
   }
   if (!pushdepth)
   {
   	screen = oscreen;
   }
}


DIALOG const *AlGuiContainer::get_dialog() const
{
	return dialog;
}

void AlGuiContainer::redraw(int fg, int bg)
{

   set_dialog_color(dialog,fg,bg);
   redraw();
}

void AlGuiContainer::redraw()
{
   if (!it_is_running)
   {
      start();
      stop();
      request(W_REQUEST_UPDATE);
      return;
   }
   diaplayer->res |= D_REDRAW; // it will redraw in the next update_dialog

   if (!in_proc)
      do_dialog(diaplayer);
      
   request(W_REQUEST_UPDATE);
   
}
