/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include "backgrnd.h"
Background::Background(BITMAP *_bmp, int _mode)
      : Wobject("Root window",WHAM->get_TheScreen_size_x(), WHAM->get_TheScreen_size_y(),0,0)
{
	lower();
   set_lock(TRUE);
   set_lock_scale(TRUE);
   set_lock_position(TRUE);
   set_lock_shift(TRUE);
	the_bmp = _bmp;
   mode = _mode;
}

Background::Background(int color)
      : Wobject("Root window",WHAM->get_TheScreen_size_x(), WHAM->get_TheScreen_size_y(),0,0)
{
	lower();
   set_lock(TRUE);
   set_lock_scale(TRUE);
   set_lock_position(TRUE);
   set_lock_shift(TRUE);
	the_bmp = NULL;
   mode = color;
}



void Background::draw(BITMAP *onto, int f_x, int f_y,int t_x, int t_y)
{
		WinSystem *manager= get_manager();
      if (!the_bmp)
      {
         solid_mode();
         rectfill(onto,f_x,f_y,t_x,t_y,mode);
         return;
      }
      
		switch(mode & 7) // & 7 because the switch should lookmonly at the lowest 3 bits
      {
      	default:
      	case BACKGROUND_MODE_STRETCH:
         	stretch_blit
            (
            	the_bmp,
               onto,
               0,0,the_bmp->w,the_bmp->h,
               0,0,manager->get_TheScreen_size_x(),manager->get_TheScreen_size_y()
            );
            break;
         case BACKGROUND_MODE_TILE:
            if (mode & BACKGROUND_MODE_MASKED)
            {
                solid_mode();

                rectfill(onto,0,0,manager->get_TheScreen_size_x()-1,manager->get_TheScreen_size_y()-1,manager->BackgroundColor);
            }
         	for (int x = 0;x<manager->get_TheScreen_size_x();x+= the_bmp->w)
            {
            	for (int y = 0;y<manager->get_TheScreen_size_y();y+=the_bmp->h)
               {
               	if (x <= t_x && x + the_bmp->w >= f_x && y <= t_y && y + the_bmp->h >= f_y)
                  {
                     if (mode & BACKGROUND_MODE_MASKED)
                        masked_blit(the_bmp,onto,0,0,x,y,the_bmp->w,the_bmp->h);
                     else
                        blit(the_bmp,onto,0,0,x,y,the_bmp->w,the_bmp->h);
                  }
               }
            }
            break;
         case BACKGROUND_MODE_CENTER:
            solid_mode();
            rectfill(onto,0,0,manager->get_TheScreen_size_x()-1,manager->get_TheScreen_size_y()-1,manager->BackgroundColor);
            
            if (mode & BACKGROUND_MODE_MASKED)
             masked_blit(the_bmp,onto,0,0,
            	(manager->get_TheScreen_size_x() - the_bmp->w)/2,
               (manager->get_TheScreen_size_y() - the_bmp->h)/2,
               the_bmp->w,the_bmp->h);
            else
             blit(the_bmp,onto,0,0,
            	(manager->get_TheScreen_size_x() - the_bmp->w)/2,
               (manager->get_TheScreen_size_y() - the_bmp->h)/2,
               the_bmp->w,the_bmp->h);
            break;
      }
}


Wobject *Background::process_mouse(int *x_out, int *y_out)
{
    Wobject *win; 
    int out_x,out_y;
    int flag; // is the border clicked?

    //get the object the mouse is currently in( if flag is set it is in a border)
    //out_x, out_y will be the coordinates relative to that object 
    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);

    //store coordinates relative to this object if the user wants them
    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;
    return (Wobject *)win;
}


void Background::set(BITMAP *_bmp, int _mode)
{

   mode = _mode;
   the_bmp = _bmp;
   update();

}
