/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/

#include <stdio.h>
#include <strings.h>
#include "../include/framewin.h"

extern BITMAP *_mouse_screen;

int FrameWin::count = -1;
SystemBox **FrameWin::systemboxes = NULL;
int FrameWin::number_of_system_boxes = 0;
int FrameWin::Border_Color = 2;
int FrameWin::Focus_Color = -1;
int FrameWin::Title_Color = 0;
int FrameWin::max_hor_borderlength = 0;
int FrameWin::Move_Style = FrameWin_OPAQUE;
BITMAP *FrameWin::upper_border = NULL;

FrameWin::FrameWin(int _size_x,int _size_y,int _pos_x, int _pos_y, char *_name)
    : Window(_name,_size_x,_size_y,_pos_x,_pos_y , FrameWin::draw_border)
{
    create();
}

FrameWin::FrameWin(int _size_x,int _size_y,int _pos_x, int _pos_y, char *_name, WinSystem *its_manager)
    : Window(_name,_size_x,_size_y ,_pos_x,_pos_y,its_manager, FrameWin::draw_border)
{
    create();
}


FrameWin::FrameWin(BITMAP *w,int _pos_x, int _pos_y, char *_name)
    : Window(_name,w,_pos_x,_pos_y, FrameWin::draw_border)
{
    create();
}

FrameWin::FrameWin(BITMAP *w,int _pos_x, int _pos_y, char *_name, WinSystem *its_manager)
    : Window(_name,w,_pos_x,_pos_y,its_manager, FrameWin::draw_border)
{
    create();
}

void FrameWin::create()
{
    count++;
    if (!systemboxes)
    {
        systemboxes = (SystemBox **)malloc(sizeof(SystemBox *) * (number_of_system_boxes +1));
        if (!systemboxes)
            out_of_mem("cannot create systembox");
        systemboxes[0] = NULL;

    }
}


FrameWin::~FrameWin()
{

    if (!count)
    {

        destroy_system_boxes();
        if (upper_border)
        {
          destroy_bitmap(upper_border);
          upper_border = NULL;
          max_hor_borderlength = 0;
        }
    }
    count--;
}

void FrameWin::add_system_box(BITMAP *bmp,void (_function)(Wobject *win))
{
    if (!systemboxes)
    {
        number_of_system_boxes = 0;
        systemboxes = (SystemBox **)malloc(sizeof(SystemBox *) * (number_of_system_boxes +1));
        if (!systemboxes)
            out_of_mem("cannot create systembox");
        systemboxes[0] = NULL;
    }
    number_of_system_boxes++;
    systemboxes = (SystemBox **)realloc(systemboxes, sizeof(SystemBox *) * (number_of_system_boxes +1));
    if (!systemboxes)
            out_of_mem("cannot create systembox");
    systemboxes[number_of_system_boxes] = NULL;
    systemboxes[number_of_system_boxes - 1] = new SystemBox(bmp,_function);
    if (!systemboxes[number_of_system_boxes - 1])
            out_of_mem("cannot create systembox");

}

void FrameWin::destroy_system_boxes()
{
    if (systemboxes)
    {
        for (int i=0;systemboxes[i];i++)
        {
            delete systemboxes[i];
        }
        free (systemboxes);
        systemboxes = NULL;
        number_of_system_boxes = 0;
    }
}

int FrameWin::draw_border(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y)
{
    int
        f_x,f_y,t_x,t_y;

    GeoInfo const *g = the_object->get_geoinfo();
    if (!onto) // return border dimensions to manager
    {
        switch(from_x)
        {
            case WINDOW_UPPER_BORDER:
                return text_height(font)+2;
            break;
            case WINDOW_LOWER_BORDER:
                return 3;
            break;
            case WINDOW_LEFT_BORDER:
                return 3;
            break;
            case WINDOW_RIGHT_BORDER:
                return 3;
            break;
        }

    }

    //if (_mouse_screen == screen) trace(" in draw border");
    solid_mode();
    if (Focus_Color < 0)
      Focus_Color = Border_Color;

    int bcolor = (g->has_focus) ? Focus_Color : Border_Color;

    
    
    if (g->screen_size_x > max_hor_borderlength)
    {
    
        max_hor_borderlength = max_hor_borderlength > g->screen_size_x ? max_hor_borderlength : g->screen_size_x;
        if (upper_border)
        {
            destroy_bitmap(upper_border);
        }
        upper_border = create_bitmap(max_hor_borderlength, g->Upper_Border_Size);

        if (!upper_border)
            fatal("out of memory while allocating buffer for border drawing");
            
        clear(upper_border);
//        rectfill(upper_border
    }
    // upper border needs redrawing each time it is used because of title and system boxes
    
    
    clear_to_color(upper_border,bcolor);
//    rectfill(upper_border,0,0,upper_border->w,upper_border->h,bcolor);
    if (g->name)
    {
        text_mode(bcolor);
        textout(upper_border,font,g->name,2,1,Title_Color);
    }
    int s_pos = g->screen_size_x;
    for (int i=0;systemboxes[i];i++)
    {
        s_pos -= systemboxes[i]->sprite->w;
        draw_sprite(upper_border,systemboxes[i]->sprite,s_pos,1);
    }
    f_x = g->pos_x;
    t_x = g->pos_x + g->screen_size_x -1;
    f_y = g->pos_y - g->Upper_Border_Size;
    t_y = g->pos_y;           
                              
    f_x = MAX(f_x, from_x);   
    t_x = MIN(t_x, to_x);     
    f_y = MAX(f_y, from_y);   
    t_y = MIN(t_y, to_y);     

    blit(upper_border,onto,f_x - g->pos_x , f_y - g->pos_y + g->Upper_Border_Size, f_x,f_y, t_x - f_x + 1, t_y - f_y + 1);
        
    f_x = g->pos_x - 2;
    t_x = g->pos_x + g->screen_size_x + 1;
    f_x = MAX(f_x, from_x);   
    t_x = MIN(t_x, to_x);     
    if (f_y == g->pos_y - g->Upper_Border_Size && f_x <= t_x)
        hline (onto,f_x,f_y,t_x,Outline_Color);




    //lower border                         
    f_x = g->pos_x - 3;                   
    t_x = g->pos_x + g->screen_size_x + 2; 
    f_y = g->pos_y + g->screen_size_y;     
    t_y = g->pos_y + g->screen_size_y + 2; 
                                           
    f_x = MAX(f_x, from_x);                
    t_x = MIN(t_x, to_x);                  
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  
    if (f_x <= t_x && f_y <= t_y)          
        rectfill(onto,f_x,f_y,t_x,t_y,bcolor);
    if (t_y == g->pos_y + g->screen_size_y + 2 && f_x <= t_x) 
        hline(onto,f_x,t_y,t_x,Outline_Color);

    //left border                          
    f_x = g->pos_x - 3;                   
    t_x = g->pos_x - 1;
    f_y = g->pos_y - g->Upper_Border_Size + 1;
    t_y = g->pos_y + g->screen_size_y - 1;     
                                           
    f_x = MAX(f_x, from_x);                
    t_x = MIN(t_x, to_x);                  
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  
    if (f_x <= t_x && f_y <= t_y)          
        rectfill(onto,f_x,f_y,t_x,t_y,bcolor);
    
    f_y = g->pos_y - g->Upper_Border_Size;
    t_y = g->pos_y + g->screen_size_y + 1;     
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  
    if (f_x == g->pos_x -3 && f_y <= t_y) 
        vline(onto,f_x,f_y,t_y,Outline_Color);
                                           
    //right border                         
    f_x = g->pos_x + g->screen_size_x;     
    t_x = g->pos_x + g->screen_size_x + 2; 
    f_y = g->pos_y - g->Upper_Border_Size + 1;
    t_y = g->pos_y + g->screen_size_y - 1;     
                                           
    f_x = MAX(f_x, from_x);                
    t_x = MIN(t_x, to_x);                  
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  
    if (f_x <= t_x && f_y <= t_y)          
        rectfill(onto,f_x,f_y,t_x,t_y,bcolor);
    f_y = g->pos_y - g->Upper_Border_Size;
    t_y = g->pos_y + g->screen_size_y + 1;     
    f_y = MAX(f_y, from_y);                
    t_y = MIN(t_y, to_y);                  

    if (t_x == g->pos_x +g->screen_size_x + 2 && f_y <= t_y) 
        vline(onto,t_x,f_y,t_y,Outline_Color);
                                           
                                           

    return 0;
}






Wobject *FrameWin::process_mouse(int *x_out, int *y_out)
{
    Wobject *win;
    int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y,mouse_is_x,mouse_is_y,size_was_x,size_was_y;
    int tmp_x,tmp_y;
    int flag;
    int s_pos;
    GeoInfo const *g;
    int mouse_visible = mouse_is_visible();
    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);
    drawing_mode(DRAW_MODE_XOR,NULL,0,0);
    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;
    
    if (win != this)
    {
        set_Mouse_Cursor(WinSystem::Mouse_Cursor0);
        return win;
    }
    if (flag)
    {
        if (out_x <0 || out_y <0)
            set_Mouse_Cursor(WinSystem::Mouse_Cursor1);
        else
            set_Mouse_Cursor(WinSystem::Mouse_Cursor2);
    }
    else
        set_Mouse_Cursor(WinSystem::Mouse_Cursor0);


    g = win->get_geometric_info();
    s_pos = g->screen_size_x;
    mouse_was_x = mouse_x;
    mouse_was_y = mouse_y;
    pos_was_x = g->pos_x;
    pos_was_y = g->pos_y;
    size_was_x = g->screen_size_x;
    size_was_y = g->screen_size_y;
    tmp_x = mouse_was_x;
    tmp_y = mouse_was_y;
    if (win->get_manager()->get_parent())
        win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);

    if ((mouse_b & 2) && !flag)
    {
        pos_was_x = g->viewport_pos_x;
        pos_was_y = g->viewport_pos_y;
        while (mouse_b & 2)
        {
            mouse_is_x = mouse_x;
            mouse_is_y = mouse_y;
            if (win->get_manager()->get_parent())
                win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
    
            win->move_viewport(pos_was_x - mouse_is_x + mouse_was_x,pos_was_y - mouse_is_y + mouse_was_y);
        }
    }
    if (mouse_b & 1)
    {
        if (flag)
        {
            if (out_x < 0 || out_y <0)
            {
                                
                if (out_y <0)
                {
                    for (int i = 0;((FrameWin *)win)->systemboxes[i];i++)
                    {
                        s_pos -= ((FrameWin *)win)->systemboxes[i]->sprite->w;
                        if (out_x > s_pos)
                        {
                            if (((FrameWin *)win)->systemboxes[i]->function)
                                ((FrameWin *)win)->systemboxes[i]->function(((FrameWin *)win));
                            return NULL;
                        }
                    }
                }
              
                if (((FrameWin *)win)->Move_Style == FrameWin_CLEAR)
                {
                    if (mouse_visible) mouse_off();
                    rect(screen,pos_was_x, pos_was_y , pos_was_x + size_was_x  - 1, pos_was_y + size_was_y  - 1,128);
                    if (mouse_visible) mouse_on();
                }
                while (mouse_b & 1)
                {
                    mouse_is_x = mouse_x;
                    mouse_is_y = mouse_y;
                    if (win->get_manager()->get_parent())
                        win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);

                    if (((FrameWin *)win)->Move_Style == FrameWin_CLEAR && ((tmp_x != mouse_x)||(tmp_y != mouse_y)))
                    {
                        if (mouse_visible) mouse_off();
                        rect
                        (
                            screen,
                            pos_was_x + tmp_x - mouse_was_x, 
                            pos_was_y + tmp_y - mouse_was_y, 
                            pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                            pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                            128
                        );
                        tmp_x = mouse_is_x;
                        tmp_y = mouse_is_y;
                        rect
                        (
                            screen,
                            pos_was_x + tmp_x - mouse_was_x, 
                            pos_was_y + tmp_y - mouse_was_y, 
                            pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                            pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                            128
                        );
                        if (mouse_visible) mouse_on();
                            

                    }
                    else if (((FrameWin *)win)->Move_Style == FrameWin_OPAQUE)
                        win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);
                }

                if (((FrameWin *)win)->Move_Style == FrameWin_CLEAR)
                {
                    if (mouse_visible) mouse_off();
                    rect
                    (
                        screen,
                        pos_was_x + tmp_x - mouse_was_x, 
                        pos_was_y + tmp_y - mouse_was_y, 
                        pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                        pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                        128
                    );
                    if (mouse_visible) mouse_on();
                        win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);
                }

            }
            else
            {
                if (((FrameWin *)win)->Move_Style == FrameWin_CLEAR)
                {
                    if (mouse_visible) mouse_off();
                        rect(screen,pos_was_x, pos_was_y , pos_was_x + size_was_x  - 1, pos_was_y + size_was_y  - 1,128);
                    if (mouse_visible) mouse_on();
                }

                while (mouse_b & 1)
                {
                    mouse_is_x = mouse_x;
                    mouse_is_y = mouse_y;
                    if (win->get_manager()->get_parent())
                        win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);

                    if (((FrameWin *)win)->Move_Style == FrameWin_CLEAR && ((tmp_x != mouse_x)||(tmp_y != mouse_y)))
                    {
                        if (mouse_visible) mouse_off();
                        rect
                        (
                            screen,
                            pos_was_x,
                            pos_was_y, 
                            pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                            pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                            128
                        );
                        tmp_x = mouse_x;
                        tmp_y = mouse_y;
                        rect
                        (
                            screen,
                            pos_was_x, 
                            pos_was_y, 
                            pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                            pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                            128
                        );
                        if (mouse_visible) mouse_on();
                            

                    }
                    else if (((FrameWin *)win)->Move_Style == FrameWin_OPAQUE)
                        win->rescale(size_was_x + mouse_is_x - mouse_was_x,size_was_y + mouse_is_y - mouse_was_y);
                }
                if (((FrameWin *)win)->Move_Style == FrameWin_CLEAR)
                {
                    if (mouse_visible) mouse_off();
                    rect
                    (
                        screen,
                        pos_was_x , 
                        pos_was_y , 
                        pos_was_x + size_was_x  - 1 + tmp_x - mouse_was_x, 
                        pos_was_y + size_was_y  - 1 + tmp_y - mouse_was_y,
                        128
                    );
                    if (mouse_visible) mouse_on();
                        win->rescale(size_was_x + mouse_is_x - mouse_was_x,size_was_y + mouse_is_y - mouse_was_y);
                }
            }
        }
        else
            win->raise();
    }
    solid_mode();
    if (!flag) 
        return (Wobject *)win;
    else
        return NULL;
}

int FrameWin::Outline_Color = 1;
