/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include <allegro.h>
#include "../include/panel.h"
#include "../include/helpers.h"
extern BITMAP * _mouse_screen;
static char buffer[1024];

#ifdef ALLEGRO_DOS
#include <io.h>
#endif
#include <stdio.h>
#include <unistd.h>

FILE *DEBUG_OUTPUT_FILE = NULL;
void trace(char *format, ...);

int mouse_is_visible()
{


	if (WHAM)
   {
//   	fprintf(stderr,"W %p  %p\n", _mouse_screen, WHAM->TheScreen);
   	return is_same_bitmap(_mouse_screen, WHAM->TheScreen);
   }
//  	fprintf(stderr,"NW %p  %p\n", _mouse_screen, screen);
   return is_same_bitmap(_mouse_screen, screen);
}


static MOUSE_STATE *mouse_state_array = NULL;
static int mouse_state_array_size = 0;
static int mouse_state_array_pos = -1;

void mouse_off(int xfrom,int yfrom, int xto , int yto)
{
 	#ifdef WTRACE
  		if (wham_trace) trace("mouse off (visible:%d)",mouse_is_visible());
  	#endif

   if (!WHAM)
   {
   	show_mouse(NULL);
   	return;
   }
  	if (xto == -1)
   		xto = WHAM->get_TheScreen_size_x();

  	if (yto == -1)
      	yto = WHAM->get_TheScreen_size_y();
   
   xto += 16;
   yto += 16;
   xfrom -=16;
   yfrom -= 16;

   mouse_state_array_pos++;
   if (mouse_state_array_pos >= mouse_state_array_size)
   {
   	mouse_state_array_size +=50;
   	mouse_state_array = (MOUSE_STATE *)realloc(mouse_state_array,mouse_state_array_size * sizeof(MOUSE_STATE));
   }
   int mouse_visible = mouse_is_visible();
   mouse_state_array[mouse_state_array_pos].is_visible = mouse_visible;
   mouse_state_array[mouse_state_array_pos].is_frozen = freeze_mouse_flag;

   if (mouse_visible)
   {
   	if (mouse_x >= xfrom && mouse_y >= yfrom && mouse_x <= xto && mouse_y <= yto)
      {
//        trace("hidin");
   		BITMAP *tmpscreen = screen;
  			if (WHAM->TheScreen)
	   		screen = WHAM->TheScreen;
			scare_mouse();
   		screen = tmpscreen;
      }
      else
      {
//      	trace("freeze");
         freeze_mouse_flag = TRUE;
      }
   }
}



void mouse_on()
{
    #ifdef WTRACE
     if (wham_trace) trace("mouse on");
    #endif

    if (mouse_state_array_pos < 0)
    {
    	show_mouse(WHAM->TheScreen);
      return;
    }
    
    freeze_mouse_flag = mouse_state_array[mouse_state_array_pos].is_frozen;

    if (mouse_state_array[mouse_state_array_pos].is_visible && !mouse_is_visible())
    {
		BITMAP *tmpscreen = screen;
   	if (WHAM->TheScreen)
	   	screen = WHAM->TheScreen;
		unscare_mouse();
   	screen = tmpscreen;
    }
	mouse_state_array_pos--;
}

void freeze_mouse()
{
    #ifdef WTRACE
     if (wham_trace) trace("freeze mouse function called");
    #endif
    freeze_mouse_flag = TRUE;
}

void thaw_mouse()
{
    #ifdef WTRACE
     if (wham_trace) trace("thaw mouse ");
    #endif
    freeze_mouse_flag = FALSE;
}




void out_of_mem(char *reason)
{
//   delete WHAM;
   WHAM = NULL;

   #ifdef ALLEGRO_DOS
   allegro_exit();
   static char msg[] = "Out of memory:";
   static char msg2[] = "\n\rpossibly garbage left in memory\n\r";
   
   _write(STDERR_FILENO, msg, sizeof(msg)-1);
   _write(STDERR_FILENO, reason, strlen(reason));
   _write(STDERR_FILENO, msg2, sizeof(msg2)-1);
   #endif
    exit(1);
}

void out_of_mem()
{

//   delete WHAM;
   WHAM= NULL;
   allegro_exit();
   #ifdef ALLEGRO_DOS
   static char msg[] = "Out of memory:";

   _write(STDERR_FILENO, msg, sizeof(msg)-1);
   #endif
    exit(1);
}


void fatal(char *format, ...)
{
    #ifdef WTRACE
      if(wham_trace) trace("FATAL CALLED !!!!!!!!!!!!!!!!!\n");
    #endif
    if (WHAM) delete WHAM;
    WHAM = NULL;
    allegro_exit();
    va_list arg;
    va_start(arg,format);
    sprintf(buffer,"FATAL:");
    vsprintf(buffer+6,format,arg);
    fputs(buffer,stderr);
    va_end(arg);
    exit(1);
}

void trace(char *format, ...)
{
    sprintf(buffer,"Warning: ");
    va_list arg;
    va_start(arg, format);
    vsprintf(buffer + 9,format,arg);
    va_end(arg);
    if (!DEBUG_OUTPUT_FILE)
    {
        if (!screen)
        {
           printf(buffer);
           readkey();
           while(keypressed())
              readkey();
           
           return;
        }
      BITMAP *tscreen = WHAM ? WHAM->TheScreen : screen;
	     int mouse_visible = mouse_is_visible();
	     int cl = tscreen->cl;
	     int cr = tscreen->cr;
	     int ct = tscreen->ct;
	     int cb = tscreen->cb;
	     static int color = makecol(255,255,255);

	     color = (color == makecol(255,255,255) ? makecol(255,255,0) : makecol(255,255,255));
    
	     text_mode(0);
    
        if (mouse_visible)
            show_mouse(NULL);

        set_clip(tscreen,0,0,SCREEN_W,SCREEN_H);

        text_mode(makecol(0,0,0));
        textout(tscreen,font,buffer,0,10,color);
//        textout(tscreen,font,buffer,0,SCREEN_H-10,color);
        if (mouse_visible)
            show_mouse(tscreen);

        clear_keybuf();
        readkey();
        while(keypressed())
            readkey();
        if (mouse_visible)
            show_mouse(NULL);
        solid_mode();
        rectfill(tscreen,0,10,SCREEN_W,18,0);
//        rectfill(tscreen,0,SCREEN_H-10,SCREEN_W,SCREEN_H,0);
        if (mouse_visible)
            show_mouse(tscreen);
        set_clip(tscreen,cl,ct,cr,cb);
    }
    else
    {
        fprintf(DEBUG_OUTPUT_FILE,buffer);
        fprintf(DEBUG_OUTPUT_FILE,"\n");
    }
        
}


void ttrace(char *format, ...)
{
	 BITMAP *tscreen = WHAM ? WHAM->TheScreen : screen;
    int mouse_visible = mouse_is_visible();
    int cl = tscreen->cl;
    int cr = tscreen->cr;
    int ct = tscreen->ct;
    int cb = tscreen->cb;
    static int color = makecol(255,255,255);
    color = (color == makecol(255,255,255) ? makecol(255,255,0) : makecol(255,255,255));


    text_mode(0);
    sprintf(buffer,"Warning: ");
    va_list arg;
    va_start(arg, format);
    vsprintf(buffer + 9,format,arg);
    va_end(arg);
    if (!DEBUG_OUTPUT_FILE)
    {        
        if (mouse_visible)
            show_mouse(NULL);
        set_clip(tscreen,0,0,SCREEN_W,SCREEN_H);
        text_mode(makecol(0,0,0));
        textout(tscreen,font,buffer,0,0,color);
        textout(tscreen,font,buffer,0,SCREEN_H-10,color);
        rest(1000);
        solid_mode();
        rectfill(tscreen,0,0,SCREEN_W,8,0);
        rectfill(tscreen,0,SCREEN_H-10,SCREEN_W,SCREEN_H,0);
        if (mouse_visible)
            show_mouse(tscreen);
        set_clip(tscreen,cl,ct,cr,cb);
    }
    else
    {
        fprintf(DEBUG_OUTPUT_FILE,buffer);
        fprintf(DEBUG_OUTPUT_FILE,"\n");
    }
        
}



static Mouse_Cursor *current_Mouse_Cursor;
Mouse_Cursor *set_Mouse_Cursor(Mouse_Cursor *cursor)
{
    int mouse_visible = mouse_is_visible();
    
    Mouse_Cursor *old = current_Mouse_Cursor;
    
    if (current_Mouse_Cursor != cursor)
    {
        if (mouse_visible) mouse_off();
        if (!cursor)
        {
            set_mouse_sprite(NULL);
            set_mouse_sprite_focus(0,0);
        }
        else
        {
            set_mouse_sprite(cursor->bmp);
            set_mouse_sprite_focus(cursor->x,cursor->y);
        }
        if (mouse_visible) mouse_on();
        current_Mouse_Cursor = cursor;
    }
    return old;
}



void w_textout(BITMAP *b,FONT *f, char *c,int x,int y, int color,int mode)
{
    text_mode(mode);
    textout(b,f,c,x,y,color);
}

void w_textout_centre(BITMAP *b,FONT *f,char *c,int x,int y, int color,int mode)
{
    text_mode(mode);
    textout_centre(b,f,c,x,y,color);
}

static char *messages[] =
{
 "UNKOWN_MESSAGE",              //0
 "W_MSG_RAISE ",                //1
 "W_MSG_HIDE  ",                //2
 "W_MSG_SHOW  ",                //3
 "W_MSG_UPDATE",                //4
 "W_MSG_RAISED",                //5
 "W_MSG_HIDDEN",                //6
 "W_MSG_SHOWN ",                //7
 "W_MSG_MOVED ",                //8
 "W_MSG_MOVE_RESCALED",         //9
 "W_MSG_RESIZED",               //10
 "W_MSG_RESCALED",              //11
 "W_MSG_VIEWPORT_RESIZED",      //12
 "W_MSG_CREATED",               //13
 "W_MSG_DESTROYED",             //14
 "W_MSG_USER_AREA_SET",         //15
 "W_MSG_LOWERED",               //16
 "W_MSG_LOWER",                 //17
 "W_MSG_ATTACHED",              //18
 "W_MSG_SHIFTED",               //19
 "W_MSG_MANAGER_DELETE",        //20
 "W_MSG_VIEWPORT_MOVED",        //21
 "W_MSG_GOING_TO_RAISE",        //22
 "W_MSG_GOING_TO_LOWER",        //23
 "W_MSG_UPDATED",               //24
 "W_MSG_LOCKED",                //25
 "W_MSG_UNLOCKED",              //26
 "W_MSG_RAISE_AND_LOCKED",      //27
 "W_MSG_LOSTMOUSE",             //28
 "W_MSG_GOTMOUSE",              //29
 "W_MSG_FROZEN",                //30
 "W_MSG_THAWN",                 //31
 "W_MSG_DESTROY"                //32
};                              

char *message_string(int message)
{
    static char stdmsg[100];
    if (message > -1 || message < -31)
    {
        sprintf(stdmsg, "UNKOWN MESSAGE (%d)",message);
        return stdmsg;
    }
    return messages[-message];
}

static BITMAP *behind_progress = NULL;
static int progress_initialised = 0;

void show_progress(char *message, int now,int min, int max)
{
    static int lastx;
    static char *last_message;
    int x;
    if (now>max && !progress_initialised) return;
    int xpos = SCREEN_W /2 -100;
    int ypos = SCREEN_H /2 -10;
    if (!progress_initialised)
    {
        if (!behind_progress)
            behind_progress = create_bitmap(200,20);
        if (!behind_progress)
            out_of_mem();
        blit(screen,behind_progress,xpos,ypos,0,0,200,20);
        progress_initialised = TRUE;
        rect(screen,xpos,ypos,xpos+199,ypos+19,makecol(50,50,50));
        rect(screen,xpos+1,ypos+1,xpos+199,ypos+19,makecol(150,150,150));
        rectfill(screen,xpos+1,ypos+1,xpos+198,ypos+18,makecol(100,100,100));
        lastx=0;
    }
    if (now >= max)
    {
        blit(behind_progress,screen,0,0,xpos,ypos,200,20);
        progress_initialised = FALSE;
        return;
    }
    x =  197 * (now - min)/max;
    if (x>lastx || message != last_message)
    {

        int mouse_visible = mouse_is_visible();
        if (mouse_visible)
            mouse_off();

        if (message != last_message)
            rectfill(screen,xpos+1,ypos+1,xpos+198,ypos+18,makecol(100,100,100));
        
            
        rectfill (screen,xpos+2,ypos+2,xpos+2 + x,ypos+17,makecol(255,0,0));
        text_mode(-1);
        textout(screen,font,message,xpos+4,ypos+4,makecol(0,0,0));
        if (mouse_visible)
            mouse_on();
        lastx = x;
        last_message = message;
    }
}
                                             
void wham_trans_calc_progress(int pos)
{
    //textout(screen,font,"setting up shading table",0,0,15);
    //rect(screen,0,8,SCREEN_W -1, 16,15);
    //rectfill(screen,1,9,(SCREEN_W-2) * pos / 256 + 1,15,14);
    show_progress("setting up shading table",pos,0,255);
}

void remap_bitmap(BITMAP *bmp,PALETTE old, int method) // currently just one method, plan to add dithering later
{
    int mouse_visible = mouse_is_visible();
    if (mouse_visible)
        mouse_off();
    
    int size = bmp->w * bmp->h;
    int show_progress_bar = (size > 20000);
    int col;
    int h =bmp->h;
    switch (method)
    {
    case 0:
        for (int i = 0;i<bmp->w;i++)
        {
            for (int j=0;j<bmp->h;j++)
            {
                if (show_progress_bar && !((i*h +j) % 100))
                    show_progress("remapping bitmap",i*h +j,0,size);
                col = _getpixel(bmp,i,j);
                if (col)
                {
                    _putpixel 
                    (
                        bmp,
                        i,
                        j,
                        makecol
                        (
                            old[col].r *  256/64,
                            old[col].g  * 256/64,
                            old[col].b  * 256/64
                        )
                    );
                }
            }
        }
    }
    if (show_progress_bar)
                show_progress("remapping bitmap",size,0,size);
    if (mouse_visible)
        mouse_on();
}

void msgbox(char *format,...)
{
    sprintf(buffer,"Warning: ");
    va_list arg;
    va_start(arg, format);
    vsprintf(buffer + 9,format,arg);
    va_end(arg);
    if (!screen)
    {
       printf(buffer);
       readkey();
       while(keypressed())
          readkey();
       
       return;
    }
    if (!WHAM)
    {
       trace(buffer);
       return;
    }
    int size_x = text_length(font,buffer) + 10;
    int size_y = 4 * text_height(font);
    Wobject *message = new Panel(size_x,size_y, (SCREEN_W - size_x) /2, (SCREEN_H - size_y)/2,"BugBox");
    textout_centre(message->bmp(), font,buffer, size_x/2, size_y/2, makecol(0,0,0));

    message->raise_and_lock();
    message->show();
    clear_keybuf();
    while(!keypressed())
    {
      if (message == Wham_Secretary.process_mouse() && mouse_b)
      {
         while(mouse_b);
         break;
      }
    }
    clear_keybuf();
    message->hide();
    delete message;
    
}


static Panel *winlist = NULL;

int helper_update_winlist(void *manager)
{
   if (!winlist)
      return -1;
      
   int nr = ((WinSystem *)manager)->get_number_of_wobjects();
   int locked,poslocked,sizelocked,scalelocked,shiftlocked, vpslocked, vpplocked,vislocked;
   char *name;
   int visible;
   Wobject *w;
   GeoInfo const *g;
   for (int i = 0;i<nr;i++)
   {
      w = ((WinSystem *)manager)->get_wobject_from_hierarchy(i);
      g = w->get_geoinfo();
      name = g->name;
      locked = g->is_locked ? 1 : 0;
      poslocked = g->pos_is_locked ? 1 : 0;
      sizelocked = g->size_is_locked ? 1 : 0;
      scalelocked = g->scale_is_locked ? 1 : 0;
      shiftlocked = g->shift_is_locked ? 1 : 0;
      vpslocked = g->viewport_size_is_locked ?  1 : 0;
      vpplocked = g->viewport_pos_is_locked ?  1 : 0;
      vislocked = g->visibility_is_locked ? 1: 0;
      visible = g->is_displayed ? 1 : 0;
      text_mode(Panel::Background_Color);
      textprintf(winlist->bmp(),font,1, 9 * i + 1, 0, "%d%d%d%d%d%d%d%d  %d %s                      ",locked,poslocked,sizelocked,scalelocked,shiftlocked,vpplocked,vpslocked,vislocked,visible,name);
   }
   textprintf(winlist->bmp(),font,1, 9 * nr + 1, 0, "                                                  ");

   winlist->update();
  return 0;
  
}

void helper_get_winlist(WinSystem *manager)
{
   if (winlist)
   {
      helper_update_winlist(manager);;
      return;
   }
   int nr = manager->get_number_of_wobjects();
   winlist = new Panel(400, (nr + 50) * 10, 5,5,"Object List");
   Wham_Secretary.schedule(helper_update_winlist, manager,1);
   winlist->show();
   winlist->set_lock_size(FALSE);
}

void helper_close_winlist()
{
 if (winlist)
  delete winlist;
 winlist = NULL;
}

BITMAP *hw_screen = NULL;


void arced_rect(BITMAP *onto, int xf, int yf, int xt, int yt, int c)
{
 static BITMAP *arc = NULL;
 static int pattern = 0;
 if (!arc)
 {
   arc = create_bitmap(8,8);
   if (!arc)
      out_of_mem("creating fill pattern bitmap");
 }
 pattern++;
 pattern %=4;
 clear(arc);
 switch (pattern)
 {
  case 0:
    line(arc,0,0,7,7,1);
   break;
  case 2:
    line(arc,0,0,7,3,1);
    line(arc,0,4,7,7,1);
   break;
  case 3:
    line(arc,0,0,3,7,1);
    line(arc,4,0,7,7,1);
   break;
  default:
    line(arc,0,7,7,0,1);
   break;
 }
 drawing_mode(DRAW_MODE_MASKED_PATTERN, arc,0,0);
 rectfill(onto,xf,yf,xt,yt,c);
 solid_mode();
 rect(onto,xf,yf,xt,yt,c);
}
