/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/

#include "helpers.h"


#include "pager.h"
Pager::Pager(int from_x, int from_y,int to_x,int to_y,  int _scale)
	  : Wobject("pager",(to_x - from_x)/_scale,(to_y - from_y)/_scale,5,10,Pager::draw_border)
{
	set_lock_shift(TRUE);
	xfrom = from_x;
	xto = to_x;
	yfrom = from_y;
	yto = to_y;
	scale = _scale;
}

void Pager::draw(BITMAP *onto, int f_x, int f_y,int t_x, int t_y)
{
	WinSystem *m = get_manager();
	int nr = m->get_number_of_wobjects();
	int from_x,from_y,to_x,to_y;
	GeoInfo const *g;
	Wobject *w;
	int color;



	solid_mode();

	// get boundaries
	f_x = MAX(f_x,get_geoinfo()->pos_x);
	f_y = MAX(f_y,get_geoinfo()->pos_y);
	t_x = MIN(t_x,get_geoinfo()->pos_x + get_geoinfo()->screen_size_x - 1);
	t_y = MIN(t_y,get_geoinfo()->pos_y + get_geoinfo()->screen_size_y - 1);
	if (t_x < f_x || t_y <f_y)    
		return;

	rectfill(onto,f_x,f_y,t_x,t_y,Background_Color);
	

	for (int i = nr ;i>=0;i--)
	{

		// get coord on screen
		if (i< nr)
		{
		 w = m->get_wobject_from_hierarchy(i);
		 g = w->get_geoinfo();
       if (!g->is_displayed)
       	continue;

		 if (w == this)
			color = Pager_Color;
		 else if (w == m->get_top_wobject())
			color = Active_Color;
		 else
			color = Passive_Color;

			from_x = g->pos_x - g->Left_Border_Size;
			to_x = g->pos_x + g->screen_size_x + g->Right_Border_Size - 1;
			from_y = g->pos_y - g->Upper_Border_Size;
			to_y = g->pos_y + g->screen_size_y + g->Lower_Border_Size - 1;
		}
		else
		{
			// draw the viewport
		 from_x = 0;
		 from_y = 0;
		 to_x = get_manager()->get_TheScreen_size_x();
		 to_y = get_manager()->get_TheScreen_size_y();
		 color = Background_Color_Screen;
		}
		// substract origin of pager
		from_x -= xfrom;
		from_y -= yfrom;
		to_x   -= xfrom;
		to_y   -= yfrom;

		
		// scale to pager
		from_x /= scale;
		from_y /= scale;
		to_x /= scale;
		to_y /= scale;



		// scale to screen size

		from_x = (int)(from_x * get_geoinfo()->scale_x);
		from_y = (int)(from_y * get_geoinfo()->scale_y);
		to_x = (int)(to_x * get_geoinfo()->scale_x);
		to_y = (int)(to_y * get_geoinfo()->scale_y);


		// translate to screenposition
		from_x += get_geoinfo()->pos_x;
		from_y += get_geoinfo()->pos_y;
		to_x += get_geoinfo()->pos_x;
		to_y += get_geoinfo()->pos_y;


		// check boundaries
		from_x = MAX(f_x, from_x);
		from_y = MAX(f_y, from_y);
		to_x = MIN(to_x, t_x);
		to_y = MIN(to_y, t_y);

		if (to_x >= from_x && to_y >= from_y)
		{
				rectfill(onto,from_x,from_y,to_x,to_y,color);
		}
	}


}


int Pager::draw_border(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y)
{
	int
		f_x,f_y,t_x,t_y;


	if (!onto) // return border dimensions to manager
	{
		switch(from_x)
		{
			case WINDOW_UPPER_BORDER:
				return 5;
			break;
			case WINDOW_LOWER_BORDER:
				return 5;
			break;
			case WINDOW_LEFT_BORDER:
				return 5;
			break;
			case WINDOW_RIGHT_BORDER:
				return 5;
			break;
		}

	}

	GeoInfo const *g = the_object->get_geoinfo();
	solid_mode();
	//upper border
	f_x = g->pos_x - 5;
	t_x = g->pos_x + g->screen_size_x + 4;
	f_y = g->pos_y - 5;
	t_y = g->pos_y;           
							  
	f_x = MAX(f_x, from_x);   
	t_x = MIN(t_x, to_x);     
	f_y = MAX(f_y, from_y);   
	t_y = MIN(t_y, to_y);     
	if (f_x <= t_x && f_y <= t_y) 
		rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
										   
	//lower border                         
	f_x = g->pos_x - 5;
	t_x = g->pos_x + g->screen_size_x + 4;
	f_y = g->pos_y + g->screen_size_y;     
	t_y = g->pos_y + g->screen_size_y + 4;
										   
	f_x = MAX(f_x, from_x);                
	t_x = MIN(t_x, to_x);                  
	f_y = MAX(f_y, from_y);                
	t_y = MIN(t_y, to_y);                  
	if (f_x <= t_x && f_y <= t_y)          
		rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
										   
	//left border                          
	f_x = g->pos_x - 5;
	t_x = g->pos_x;                        
	f_y = g->pos_y;                        
	t_y = g->pos_y + g->screen_size_y;     
										   
	f_x = MAX(f_x, from_x);                
	t_x = MIN(t_x, to_x);                  
	f_y = MAX(f_y, from_y);                
	t_y = MIN(t_y, to_y);                  
	if (f_x <= t_x && f_y <= t_y)          
		rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
										   
	//right border                         
	f_x = g->pos_x + g->screen_size_x;     
	t_x = g->pos_x + g->screen_size_x + 4;
	f_y = g->pos_y;                        
	t_y = g->pos_y + g->screen_size_y;     
										   
	f_x = MAX(f_x, from_x);                
	t_x = MIN(t_x, to_x);                  
	f_y = MAX(f_y, from_y);                
	t_y = MIN(t_y, to_y);                  
	if (f_x <= t_x && f_y <= t_y)          
		rectfill(onto,f_x,f_y,t_x,t_y,Border_Color);
										   
										   
	return 0; //returnvalue currently ignored 
}

Wobject *Pager::process_mouse(int *x_out, int *y_out)
{
	Wobject *win; 
	int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y,mouse_is_x,mouse_is_y,size_was_x,size_was_y; // helper variables
	int flag; // is the border clicked?
	int xclick,yclick; // scaled variables to scroll the screen to
	GeoInfo const *g; // to store geoinfo from object in

	//get the object the mouse is currently in( if flag is set it is in a border)
	//out_x, out_y will be the coordinates relative to that object 
	win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);

	//store coordinates relative to this object if the user wants them
	if (x_out)
		*x_out = out_x;
	if (y_out)
		*y_out = out_y;
	if (win != this)
		return win;

	raise();

	if (flag)
	{
	   if (out_x <0 || out_x <0)
		  set_Mouse_Cursor(WinSystem::Mouse_Cursor1);
	   else
		  set_Mouse_Cursor(WinSystem::Mouse_Cursor2);
	}
	else
	   set_Mouse_Cursor(WinSystem::Mouse_Cursor0);

	// mouse button 1 is currently pressed down
	if (mouse_b & 1)
	{
	   g = win->get_geometric_info(); // get info on this object into g 
	   mouse_was_x = mouse_x; //store current mouse position
	   mouse_was_y = mouse_y;
	   pos_was_x = g->pos_x; //store current object location 
	   pos_was_y = g->pos_y;
	   size_was_x = g->screen_size_x;//store current object size
	   size_was_y = g->screen_size_y;

	   //if the window has a parent manager, the parent window could be rescaled   
	   //so calc the mouse position relative to the parent window
	   if (win->get_manager()->get_parent())
		   win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);

	   if (flag) // mouse is on a border 
	   {
		   if (out_x < 0 || out_y <0) // left or upper border -> move
		   {
			   while (mouse_b & 1)
			   {
				   mouse_is_x = mouse_x; // get current mouse pos
				   mouse_is_y = mouse_y;
				   // calc current mouse pos relative to parent win
				   if (win->get_manager()->get_parent())
					   win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
				   // and move the window
				   win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);
			   }
		   }
		   else // left or lower border -> rescale
		   {
			   while (mouse_b & 1)
			   {
				   mouse_is_x = mouse_x;// get current mosue pos
				   mouse_is_y = mouse_y;
				   // make mouse pos relative to parent window
				   if (win->get_manager()->get_parent())
					   win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);
				   //rescale the object
				   win->rescale(size_was_x + mouse_is_x - mouse_was_x,size_was_y + mouse_is_y - mouse_was_y);
			   }
		   }
	   }
	   else // flag =0 -> clicked in the window
	   {
			WinSystem *m = get_manager();
			win->raise();
			xclick = out_x * scale;
			yclick = out_y * scale;
			xclick += xfrom;
			yclick += yfrom;
			while(mouse_b);
			m->shift_all(-xclick + m->get_TheScreen_size_x()/2,-yclick + m->get_TheScreen_size_y()/2);
	   }
	}
	if (!flag) 
		return (Wobject *)win;
	else
		return NULL;
}

int Pager::process_message(Wobject *from,int msg)
{
	
	switch(msg)
	{
		case  W_MSG_RAISED:
			//fall through
		case  W_MSG_HIDDEN:
			//fall through
		case  W_MSG_SHOWN:
			//fall through
		case  W_MSG_MOVED:
			//fall through
		case  W_MSG_MOVE_RESCALED:
			//fall through
		case  W_MSG_RESIZED:
			//fall through
		case  W_MSG_RESCALED:
			//fall through
		case  W_MSG_VIEWPORT_RESIZED:
			//fall through
		case  W_MSG_CREATED:
			//fall through
		case  W_MSG_DESTROYED:
			//fall through
		case  W_MSG_USER_AREA_SET:  
			//fall through
		case  W_MSG_LOWERED:
			update();
			break;
		default:
			break;
	}

	return  Wobject::process_message(from, msg);
}


int Pager::Border_Color = 0;
int Pager::Passive_Color = 2;
int Pager::Active_Color = 3;
int Pager::Pager_Color = 4;
int Pager::Background_Color = 1;
int Pager::Background_Color_Screen = 5;

