/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/

#include "helpers.h"
#include "secretar.h"
#include <screensv.h>



void ScreensTimer() //just increases some counters every second
{
        Secretary::SecCounter++;
        Secretary::SecCounter %= 3600;
        Secretary::ScreenScounter++;
        Secretary::ScreenScounter %= 32000;
}
END_OF_FUNCTION(ScreensTimer);


int Secretary::initialized = 0;

void Secretary::destroy()
{
    if (initialized)
    {
        if (screensaver)
            set_screen_saver(NULL,0);
        if (schedulerlist)
        {
            while(schedulerlist->next)
                delete schedulerlist->next;
            delete schedulerlist;
        }
        initialized = 0;
    }
    if (number_of_loaded_screensavers)
    {
        for (int i=0;i<number_of_loaded_screensavers;i++)
        {
        	clear_keybuf();
            delete loaded_screensaver_filenames[i];
        }
        free(loaded_screensaver_filenames);
        free(loaded_screensaver_names);
        free(loaded_screensavers);
        number_of_loaded_screensavers = 0;
        loaded_screensaver_names = NULL;
        loaded_screensavers = NULL;
    }
}

void Secretary::create()
{
    initialized++;
    if (!schedulerlist)
    {
        schedulerlist = new Scheduler_List(NULL,NULL,0,NULL);
    }    
}

Secretary::Secretary()
{
    create();

}

Secretary::~Secretary()
{
    destroy();
}

int Secretary::number_of_loaded_screensavers = 0;
char **Secretary::loaded_screensaver_names = NULL;
char **Secretary::loaded_screensaver_filenames = NULL;
void **Secretary::loaded_screensavers = NULL;
char *Secretary::current_screen_saver_name = NULL;

char *Secretary::get_current_screen_saver_name()
{
    return current_screen_saver_name;
}

Wobject *Secretary::mouse_object = NULL;


static int alternate;
int Secretary::process_message(Wobject *from,int msg, Wobject *to)
{
      if (mouse_object == from && mouse_object)
      {
         if (msg == W_MSG_DESTROYED)
            mouse_object = NULL;
      }

      return 0;
}


static int checkkey()
{
 for (int i=0;i<KEY_MAX;i++)
  if (key[i])
     return TRUE;
     
 return keypressed();
}


Wobject *Secretary::process_mouse(int *x, int *y)
{

    int out_x,out_y,flag; // not used but needed for process_coord
    Wobject *obj = WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);
    Wobject *ret = NULL;

    if (!obj || !flag)
       set_Mouse_Cursor(WinSystem::Mouse_Cursor0);


//    text_mode(0);
//    textprintf(screen,font,0,0,makecol(0,255,0),"%d %d",SecCounter, ScreenScounter);
    
    if (screensaver)
    {
        if (checkkey() || (old_mouse_x !=mouse_x) || (old_mouse_y != mouse_y) || mouse_b || screensaver_suspended)
        {
//               ttrace("checkkey() %d ox %d x %d oy %d y %d b %d sus %d",
//                checkkey(),old_mouse_x,mouse_x,old_mouse_y,mouse_y,mouse_b,screensaver_suspended);
                
               ScreenScounter = 0;
        }
        old_mouse_x = mouse_x;
        old_mouse_y = mouse_y;

        if (ScreenScounter>ScreenStime)
        {
            PALETTE p;
            get_palette(p);
            int mouse_visible = mouse_is_visible();
            if (mouse_visible)
	            mouse_off();
            screensaver();
            set_palette(p);
            WHAM->refresh_all_visible_objects();
            if (mouse_visible)
	            mouse_on();
            ScreenScounter = 0;
        }
    }
    
    schedulerlist->do_list();

    if (obj != mouse_object)
    {
        if (mouse_object)
        {
            WHAM->send_message(NULL,W_MSG_LOSTMOUSE,mouse_object);
        }
        mouse_object = obj;
        if (mouse_object)
        {
            WHAM->send_message(NULL,W_MSG_GOTMOUSE,mouse_object);
        }
    }

    if (WHAM->get_keyboard_claimer() && keypressed())
    {
	    WHAM->get_keyboard_claimer()->process_keyboard();
       ScreenScounter = 0;
    }

    if (obj)
    {
        // if there are keys in buffer and there is no claimer, call the keyboard processor
        // of the current object instead of the mouse processor every other loop
        // don't call it every loop, for if the object doesn't eat keys
        // the mouse function would not be called at all anymore
        alternate = !alternate;
        
        if ((!(WHAM->get_keyboard_claimer())) && keypressed() && alternate)
        {
            obj->process_keyboard();
            ScreenScounter = 0;
        }
        else          // else is neccesary becaus the keys could have closed the window
        {
            ret = obj->process_mouse(x, y);
        }
    }
    else if (catch_keys)
    {
         clear_keybuf();
    }

    WHAM->do_requests();
    WHAM->dispatch_delayed_messages();
    
    return ret;
}





void Secretary::set_screen_saver(void (*_screensaver)(),int time, char *name)
{

        if (!init_timer)
        {
            LOCK_VARIABLE(ScreenScounter);
            LOCK_VARIABLE(SecCounter);
            LOCK_FUNCTION(ScreensTimer);
            
            if(install_int_ex(ScreensTimer,SECS_TO_TIMER(1000)))
              trace("failed to allocate timer");
            init_timer++;
        }
        screensaver = _screensaver;
        ScreenStime = time;
        if (!name)
            current_screen_saver_name = "User defined";
        else
            current_screen_saver_name = name;


}




int Secretary::load_screen_saver(char *name,int time)
{



        screensaver = NULL;
        char *screensaver_name = NULL;
        for (int i=0;i<number_of_loaded_screensavers;i++)
        {
            if (!strcmp(name,loaded_screensaver_filenames[i]))
            {
                screensaver = (void (*)())loaded_screensavers[i];
                screensaver_name = loaded_screensaver_names[i];
            }
        }
        // since screensavers cannot be unloaded, I keep a list of
        // already loaded screensavers, to prevent loading them twice
        // which would be a waste of memory
        if (!screensaver)
        {
            screensaver = load_screensaver(name,&screensaver_name);
            if (screensaver)
            {
                number_of_loaded_screensavers++;
                loaded_screensaver_names =(char **)realloc(loaded_screensaver_names, sizeof(char *) * number_of_loaded_screensavers);
                loaded_screensaver_filenames =(char **)realloc(loaded_screensaver_filenames, sizeof(char *) * number_of_loaded_screensavers);
                loaded_screensavers = (void **)realloc(loaded_screensavers, sizeof(void *) * number_of_loaded_screensavers);
                if (!loaded_screensaver_names || !loaded_screensaver_filenames || !loaded_screensavers)
						out_of_mem("loading screensaver");
                
                loaded_screensaver_filenames[number_of_loaded_screensavers - 1] = strdup(name);
                if (!loaded_screensaver_filenames[number_of_loaded_screensavers - 1])
                	out_of_mem("loading screensaver");
                
                current_screen_saver_name = loaded_screensaver_names[number_of_loaded_screensavers - 1] = screensaver_name;
                loaded_screensavers[number_of_loaded_screensavers - 1] = (void *)screensaver;
                
            }
            else
                return -1;

        }

        if (!init_timer)
        {
            LOCK_VARIABLE(ScreenScounter);
            LOCK_VARIABLE(SecCounter);
            LOCK_FUNCTION(ScreensTimer);
            if(install_int(ScreensTimer,1000))
              trace("failed to allocate timer");

            init_timer++;
        }
        ScreenStime = time;
        return 0;
}

void Secretary::reset_screen_saver_time(int time)
{
        ScreenStime = time;
}

void Secretary::set_key_catching(int _catch)
{
    catch_keys = _catch;
}

void Secretary::set_screensaver_suspension(int susp)
{
    screensaver_suspended = susp;
}

void Secretary::schedule(int (*fun)(void *arg),void *arg,int itime)
{
    if (!init_timer)
    {
        LOCK_VARIABLE(ScreenScounter);
        LOCK_VARIABLE(SecCounter);
        LOCK_FUNCTION(ScreensTimer);
        if(install_int_ex(ScreensTimer,SECS_TO_TIMER(1000)))
          trace("failed to allocate timer");
        init_timer++;

        
    }
    // add one to the list by creating it
    Scheduler_List *n; 
        n = new Scheduler_List(fun,arg,itime,schedulerlist);

}


volatile int Secretary::ScreenScounter = 0;
volatile int Secretary::SecCounter = 0;
int Secretary::ScreenStime = 0;
int Secretary::init_timer = 0;

int Secretary::screensaver_suspended= FALSE;

int Secretary::catch_keys = TRUE;

void (*Secretary::screensaver)() = NULL;

int Secretary::old_mouse_x = 0;
int Secretary::old_mouse_y = 0;

Scheduler_List *Secretary::schedulerlist = NULL;


Scheduler_List::Scheduler_List(int (*f)(void *arg),void *_arg,int t,Scheduler_List *head)
{
    if (head)
    {
        next = head->next;
        last = head;
        head->next = this;
        if (next)
            next->last = this;
    }
    else
    {
        next = NULL;
        last = NULL;
    }
    function = f;
    time_increment = t;
    next_execute_time = Wham_Secretary.SecCounter;
    arg = _arg;
}


Scheduler_List::~Scheduler_List()
{
    if (next)
        next->last = last;
    if (last)
        last->next =next;
}
// execute the function of the next, so the first elem never gets executed
// this is so because then I'm able to delete it
void Scheduler_List::do_list()
{
    if (next && next->function && next->next_execute_time <= Wham_Secretary.SecCounter)
    {
        if (next->function(next->arg))
        {
            delete next;    // if returns nonzero, finish
            do_list(); // call do_list again, to call new next function
        }
        else
        {
            next->next_execute_time += next->time_increment;
            next->do_list();
        }
    }
}



//**********************************

Secretary Wham_Secretary;




