/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include <stack.h>
#include <helpers.h>

Stack::Stack(int initial_stacksize)
{
	nr_of_elements = 0;
   initial_size = stacksize = initial_stacksize;
   the_data =(void **)malloc(sizeof(void *) * initial_stacksize);
}

Stack::~Stack()
{
	if (nr_of_elements >0)
      trace("Non empty stack deleted");
   free(the_data);
}

void Stack::push(void *data)
{

#ifdef WTRACE
	if (wham_trace) trace("Stack::push %p",data);
#endif
	nr_of_elements++;
   if (nr_of_elements > stacksize)
   {
	    stacksize += initial_size;
	    the_data = (void **)realloc(the_data, stacksize);
   }
   if (!the_data)
   {
   	out_of_mem("WHAM: internal stack overflow");
   }
   the_data[nr_of_elements-1] = data;
}

void *Stack::pop()
{

	nr_of_elements--;

   if (nr_of_elements < 0)
   	fatal("Wham: internal stack underflow");

#ifdef WTRACE
	if (wham_trace) trace("Stack::pop %p",the_data[nr_of_elements]);
#endif
      
   return the_data[nr_of_elements];
}


void Stack::remove(void *data)
{
     for (int i=nr_of_elements - 1;i>=0;i++)
     {
         if (data == the_data[i])
         {
            for (int j=i;j< nr_of_elements -1;j++)
                the_data[j] = the_data[j+1];
            nr_of_elements--;
            return;
         }
     }
}
int Stack::get_size()
{
//   trace("stack %d",nr_of_elements);
	return nr_of_elements;
}
