/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/



#include <wham_all.h>
#include <whaminit.h>

void cleanup_wham()
{
    if (WHAM)
    {
        delete WHAM;
    }
}

static void rgbmap_progress(int pos)
{
    show_progress("setting rgb table",pos,0,255);
}


int cfg_color_depth;
void wham_init_from_config(char *configfile)
{
    allegro_init();
    set_config_file(configfile);
    install_timer();
    install_mouse();
    install_keyboard();
    cfg_color_depth = get_config_int(NULL,"color_depth",8);
    
    int xres = get_config_int(NULL,"x_resolution",800);
    int yres = get_config_int(NULL,"y_resolution",600);
    set_color_depth(cfg_color_depth);
    if (set_gfx_mode(GFX_AUTODETECT,xres,yres,0,0))
        fatal("Unable to enter graphics mode:%s",allegro_error);
//    if (cfg_color_depth == 8)
//    {
        static PALETTE p;
        char **cols;
        char buffer[34];
        int nr;
        for (int i=0;i<256;i++)
        {
            sprintf(buffer,"%d",i);
            cols = get_config_argv("Palette",buffer,&nr);
            if (nr != 3)
                fatal("error reading pallete from inifile at color %d",i);
            p[i].r = atoi(cols[0])/ 4;
            p[i].g = atoi(cols[1])/ 4;
            p[i].b = atoi(cols[2])/ 4;
        }
        set_palette(p);

        rgb_map = (RGB_MAP *)malloc(sizeof(RGB_MAP));
        create_rgb_table(rgb_map,p,rgbmap_progress);

//    }
    mouse_on();
    WinSystem::UseBoxes = get_config_int("WinSystem","Use_Boxes",TRUE);

    WinSystem::UseSpecialSegvHandler = get_config_int("WinSystem","Use_Segv_Handler",FALSE);


    WinSystem::InitialBufferSize = get_config_int("WinSystem","Initial_Buffersize",25);
    WinSystem::Buffered = get_config_int("WinSystem","Buffered",TRUE);
    WinSystem::ShadowStyle = get_config_int("WinSystem","Shadow_Style",WHAM_NO_SHADOW);       
    WinSystem::ShadowColor = get_config_int("WinSystem","Shadow_Color",100);       
    WinSystem::ShadowSize = get_config_int("WinSystem","Shadow_Size",10);       
    if ( WinSystem::BackgroundColor < 0)
    {
        int nr;
        char **bgcols = get_config_argv("WinSystem","Background_Color",&nr);
        if (nr !=3)
        {
            fatal("error reading inifile:\n[WinSystem]\n Background_Color = <r> <g> <b>\n expected");
        }
        WinSystem::BackgroundColor = makecol(atoi(bgcols[0]),atoi(bgcols[1]),atoi(bgcols[2]));
    }
    WinSystem::Buffer_Increment =  get_config_int("WinSystem","Buffer_Increment",10);         
    WinSystem::Full_Remove = get_config_int("WinSystem","Full_Remove",TRUE);
    WinSystem::Show_Effect = get_config_int("WinSystem","Effect",0);
    WinSystem::Show_Effect_Param = get_config_int("WinSystem","Effect_Parameter",1);
    WinSystem::Vsynced = get_config_int("WinSystem","Vsynced",FALSE);

    WHAM = new WinSystem;
    atexit(cleanup_wham);
}




void wham_total_init( int pause)
{
    allegro_init();
    install_timer();
    install_keyboard();
    install_mouse();
    if (set_gfx_mode(GFX_SAFE,320,200,0,0))
    {
         fatal("%s\n",allegro_error);
    }

    // build a nice palette
    RGB *c = new RGB;

    if (!c)
        out_of_mem("cannot create RGB struct");
    for (int r=0;r<6;r++)    
    for (int g=0;g<6;g++)
    for (int b=0;b<6;b++)
    {
        c->r = r*10;
        c->g = g*10;
        c->b = b*10;
        set_color(r + 6 * g +36 * b,c);
    }
    c->r = c->g = c->b =0;
    for (int i=166;i<251;i++)
        set_color(i,c);

    //white
    c->r = 63;
    c->g = 63;
    c->b = 63;
    set_color(251,c);
    //lgray
    c->r = 45;
    c->g = 45;
    c->b = 45;
    set_color(252,c);
    //gray
    c->r = 35;
    c->g = 35;
    c->b = 35;
    set_color(253,c);
    //dgray
    c->r = 25;
    c->g = 25;
    c->b = 25;
    set_color(254,c);
    //black
    c->r = 0;
    c->g = 0;
    c->b = 0;
    set_color(255,c);
    delete c;


    mouse_on();




    WinSystem::InitialBufferSize = 40;  
    WinSystem::Buffer_Increment = 40;
    WinSystem::Buffered = TRUE;
    WinSystem::BackgroundColor= 5;
    WinSystem::TheScreen = screen;
    WinSystem::Minimum_Size_x = 40;
    WinSystem::Minimum_Size_y = 32;
    WinSystem::UseBoxes = TRUE;
    WinSystem::ShadowStyle = WHAM_NO_SHADOW;
    WinSystem::ShadowColor = 255;
    WinSystem::ShadowSize  = 10;
    BorderWin::RemapPalette = TRUE;
    BorderWin::DatafileName = "whamdat.dat";
    
    BorderWin::UpperBorderBitmap = NULL;
    BorderWin::LowerBorderBitmap = NULL;
    BorderWin::RightBorderBitmap = NULL;
    BorderWin::LeftBorderBitmap = NULL;
    BorderWin::UpperLeftCorner = NULL;
    BorderWin::UpperLeftCorner = NULL;
    BorderWin::UpperLeftCorner = NULL;
    BorderWin::UpperLeftCorner = NULL;



    WHAM = new WinSystem();
    if (!WHAM)
        out_of_mem("cannot create window manager");

    BillWin::Black_Color = 255;
    BillWin::Dark_Color = Panel::Dark_Outline_Color = 254;
    BillWin::Middle_Color = Panel::Background_Color  = 253;
    BillWin::Light_Color = Panel::Light_Outline_Color = 252;
    BillWin::White_Color = 251;
    BillWin::BorderSize = 2;
    BillWin::Border_Color = 253;
    FrameWin::Border_Color = 255;
    FrameWin::Title_Color = 251;
    FrameWin::Outline_Color = 252;



    //atexit(cleanup_wham);
    // commercial
    /*
    Window *commercial = new Window("WHAM",400,200,SCREEN_W/2 - 200, SCREEN_H/2 - 100);
    if (!commercial)
        out_of_mem("cannot create intro window");
    rect(commercial->bmp(),0,0,399,199,251);
    commercial->set_user_area(1,1,398,198);
    commercial->show();
    for (int i=0;i<225;i++)
    {
        circle(commercial->bmp(),200,100,i,i % 166);
        w_textout_centre(commercial->bmp(), font, "WHAM windowing system",200,100,251, -1);
        commercial->update();
    }

    while(!keypressed() && !mouse_b && pause);
    delete commercial;
    */
    pause = 0;
}

void wham_standardinit(char *datafilename)
{
    WinSystem::InitialBufferSize = 40;  
    WinSystem::Buffer_Increment = 40;
    WinSystem::Buffered = TRUE;
    WinSystem::BackgroundColor= 1;
    WinSystem::TheScreen = screen;
    WinSystem::Minimum_Size_x = 32;
    WinSystem::Minimum_Size_y = 32;
    WinSystem::ShadowStyle = WHAM_NO_SHADOW;
    WinSystem::UseBoxes = TRUE;
    WinSystem::ShadowColor = 128;
    WinSystem::ShadowSize  = 4;
    BorderWin::RemapPalette = TRUE;
    if (datafilename)
        BorderWin::DatafileName = datafilename;
    else
        BorderWin::DatafileName = "whamdat.dat";
    
    BorderWin::UpperBorderBitmap = NULL;
    BorderWin::LowerBorderBitmap = NULL;
    BorderWin::RightBorderBitmap = NULL;
    BorderWin::LeftBorderBitmap = NULL;
    BorderWin::UpperLeftCorner = NULL;
    BorderWin::UpperLeftCorner = NULL;
    BorderWin::UpperLeftCorner = NULL;
    BorderWin::UpperLeftCorner = NULL;
    WHAM = new WinSystem();
    if (!WHAM)
        out_of_mem("cannot create window manager");
    //atexit(cleanup_wham);
}



