/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include "whammenu.h"
#include <ctype.h>

/* when a menu contains a child menu, it is taken out
 * and a WHM_MENU_INFO struct is created, this is put in the dp field
 * and wham_handle_childmenu is used as callback function
 * where the global value wham_current_menu is used to give access to the current menu
 * element, from where the dp struct can be obtained to get the original child menu etc.
 * currently the border drawing function is ignored, it just uses popup_menu
 */
static int default_mouse_x() { return mouse_x; }
static int default_mouse_y() { return mouse_y; }


class WHAM_MENU_INFO
{
  public:
   WHAM_MENU_INFO(int (*proc)(void), void *dp, MENU *child, MENU *parent, WHAM_MENU_INFO * dellist);
   ~WHAM_MENU_INFO();
   int (*proc)(void);
   void *dp;
   MENU *child;
   MENU *parent;
   WHAM_MENU_INFO *next; // by keeping a pointer to the next one, I can just delete the last one
} ;

WHAM_MENU_INFO::WHAM_MENU_INFO(int (*_proc)(void), void *_dp, MENU *_child, MENU *_parent, WHAM_MENU_INFO *_dellist)
{
   proc = _proc;
   dp = _dp;
   parent = _parent;
   child = _child;
   next = _dellist;
}

WHAM_MENU_INFO::~WHAM_MENU_INFO()
{
 if (next)
    delete next;
}

static int wham_menu_x,wham_menu_y;
static MENU *wham_current_menu; // allegro's active_menu points to the active element, this points to the whol;e array of MENU structs

static int wham_handle_child_menu()
{
   int ret = 0, xpos, ypos;
   WHAM_MENU_INFO *i;
   i = (WHAM_MENU_INFO *)(active_menu->dp);
   active_menu->dp = i->dp;
   if (i->proc)
      ret = i->proc();
   active_menu->dp = i;

   // find out the coordinates to pop up the child

   xpos = wham_menu_x + 20;
   ypos = wham_menu_y + (active_menu - wham_current_menu) * text_height(font);

   if (!i->proc)
      ret = do_menu(i->child,xpos,ypos);
   else
      do_menu(i->child,xpos,ypos);
      
   return ret;

}

int wham_do_menu(MENU *menu, int x, int y)
{
   int ret, old_menu_x, old_menu_y;
   int menu_height;
   
   MENU *old_current_menu;
//   WHAM_MENU_INFO *dellist = NULL;

   int (*old_gui_mouse_x)(), (*old_gui_mouse_y)();

   old_gui_mouse_x = gui_mouse_x;
   old_gui_mouse_y = gui_mouse_y;
   gui_mouse_y = default_mouse_y;
   gui_mouse_x = default_mouse_x;

   menu_height = 0;
   for (MENU *m = menu;m->text;m++)
   {
      menu_height += text_height(font) + 4;
      /*
      if (m->child)
      {

         dellist = new WHAM_MENU_INFO(m->proc,m->dp,m->child,m,dellist);
         m->proc = wham_handle_child_menu;
         m->child = NULL;
         m->dp = dellist;
      } */
   }

   old_menu_x = wham_menu_x;
   old_menu_y = wham_menu_y;
   old_current_menu = wham_current_menu;
   
   wham_current_menu = menu;

   if (y>SCREEN_H/2)
      y -= menu_height;
   
   wham_menu_x = x;
   wham_menu_y = y;
   
   ret = do_menu(menu,x,y);

   wham_current_menu = old_current_menu;
   wham_menu_x = old_menu_x;
   wham_menu_y = old_menu_y;
   gui_mouse_x = old_gui_mouse_x;
   gui_mouse_y = old_gui_mouse_y;

//   delete dellist;
   
   WHAM->refresh_all_visible_objects();
   
   return ret;
}

int d_wham_menu_proc(int msg, DIALOG *d, int c)
{
   return d_menu_proc(msg,d,c);
}


int (*wham_menu_draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y) = NULL;



