/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include <stdio.h>

#include <stdarg.h>
#include "window.h"

extern BITMAP *_mouse_screen;

Window::Window(char *name,int size_x,int size_y,int pos_x,int pos_y)
    : Wobject(name,size_x,size_y,pos_x,pos_y)
{
	 BITMAP *bmp = NULL;
	 if (Allocate_From_VRAM)
		bmp = create_video_bitmap(size_x,size_y);
      
    if (!bmp)
    	bmp = create_bitmap(size_x, size_y);
      
    if (!bmp) out_of_mem("creating Window bitmap");
    
    must_delete_bitmap = TRUE;
    create(bmp);

}

Window::Window(char *name,int size_x,int size_y,int pos_x,int pos_y, WinSystem *its_manager)
      : Wobject(name,size_x,size_y,pos_x,pos_y,its_manager)
{
	 BITMAP *bmp = NULL;
	 if (Allocate_From_VRAM)
		bmp = create_video_bitmap(size_x,size_y);
      
    if (!bmp)
    	bmp = create_bitmap(size_x, size_y);

    if (!bmp) out_of_mem("creating Window bitmap");

    must_delete_bitmap = TRUE;
    create(bmp);
}

Window::Window(char *name,BITMAP *bitmap,int pos_x,int pos_y)
      : Wobject(name,bitmap->w,bitmap->h,pos_x,pos_y)
{
    must_delete_bitmap = FALSE;
    create(bitmap);
}

Window::Window(char *name,BITMAP *bitmap, int _pos_x, int _pos_y, WinSystem *_its_manager)
    : Wobject(name,bitmap->w,bitmap->h,_pos_x,_pos_y,_its_manager)
{
    must_delete_bitmap = FALSE;
    create(bitmap);

}

Window::Window(char *name,int size_x,int size_y,int pos_x,int pos_y,int (*draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y))
    : Wobject(name,size_x, size_y, pos_x, pos_y, draw_border)
{
    BITMAP *bmp = NULL;
    
	 if (Allocate_From_VRAM)
		bmp = create_video_bitmap(size_x,size_y);
      
    if (!bmp)
    	bmp = create_bitmap(size_x, size_y);

    if (!bmp) out_of_mem("creating Window bitmap");

    must_delete_bitmap = TRUE;
    create(bmp);
}

Window::Window(char *name,int size_x,int size_y,int pos_x,int pos_y, WinSystem *its_manager,int (*draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y))
    : Wobject(name, size_x, size_y, pos_x, pos_y, its_manager,draw_border)
{
    BITMAP *bmp = NULL;
    
	 if (Allocate_From_VRAM)
		bmp = create_video_bitmap(size_x,size_y);
      
    if (!bmp)
    	bmp = create_bitmap(size_x, size_y);

    if (!bmp) out_of_mem("creating Window bitmap");

    must_delete_bitmap = TRUE;
    create(bmp);
}

Window::Window(char *name,BITMAP *bitmap,int pos_x,int pos_y,int (*draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y)) // size of bitmap is used for the size of the windo
    : Wobject(name,bitmap->w,bitmap->h, pos_x, pos_y,draw_border)
{
    must_delete_bitmap = FALSE;
    create(bitmap);

}

Window::Window(char *name,BITMAP *bitmap, int _pos_x, int _pos_y, WinSystem *_its_manager,int (*draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y))
    : Wobject(name,bitmap->w,bitmap->h, _pos_x, _pos_y, _its_manager,draw_border)
{
    must_delete_bitmap = FALSE;
    create(bitmap);

}

Window::~Window()
{
    destroy();
}

void Window::create(BITMAP *bmp)
{

    window = bmp;
    if (must_delete_bitmap) clear(window); // it is a new one
    user = create_sub_bitmap(window, 0, 0, get_geoinfo()->size_x, get_geoinfo()->size_y);
    text_cursor_x = text_cursor_y = 1;
    background_color = 0;
    pal = 0;
    free_palette_on_deletion = FALSE;
}

void Window::destroy()
{
//   trace("destroy bitmap");
    if (user) destroy_bitmap(user);            // free memory of buffers
//   trace("destroy window");
    if (must_delete_bitmap)
        destroy_bitmap(window);

//   trace("destroy pal");
    if (free_palette_on_deletion && pal)
       free(pal);
}


/*************************************************************
 * draws the portion of the window that is inside the rectangle
 * given by the coordinates
 *************************************************************/
void Window::draw(BITMAP *onto, int from_x, int from_y, int to_x, int to_y)
{


    #ifdef WTRACE
        if (wham_trace) trace("Window::draw win %d",get_geoinfo()->tag);
    #endif

    int
        w_from_x,
        w_from_y,
        w_to_x,
        w_to_y;

    //readkey();
    //adjust coordinates on screen
    GeoInfo const * g = get_geoinfo();
    BITMAP *tmpfrom = window;
    if (g->screen_size_x <=0 || g->screen_size_y <= 0) // this window is so small' it not even visible
    	return;
    
    from_x = MAX(g->pos_x,from_x);
    from_y = MAX(g->pos_y,from_y);
    to_x = MIN(to_x,g->pos_x + g->screen_size_x - 1);
    to_y = MIN(to_y,g->pos_y + g->screen_size_y - 1);
    
    // get coordinates in wobject
    w_from_x = g->viewport_pos_x + (int)((from_x - g->pos_x) * (float)g->viewport_size_x / (float)g->screen_size_x);
    w_to_x = g->viewport_pos_x + (int)((to_x - g->pos_x) * (float)g->viewport_size_x / (float)g->screen_size_x);
    w_from_y = g->viewport_pos_y + (int)((from_y - g->pos_y) * (float)g->viewport_size_y / (float)g->screen_size_y);
    w_to_y = g->viewport_pos_y + (int)((to_y - g->pos_y) * (float)g->viewport_size_y / (float)g->screen_size_y);
    

    
    if (w_to_x >= w_from_x && w_to_y >= w_from_y)
    {
        if (pal && bitmap_color_depth(onto) !=8)
           select_palette(pal);
        if (g->screen_size_x != g->viewport_size_x || g->screen_size_y != g->viewport_size_y)
        {
            // to avoid wiggling we need to stretch_blit the whole window
            // since it winsys sets the clipping rectangle of 'onto' this is not
            // as inefficient as it seems to be
            from_x = g->pos_x;
            to_x = g->pos_x + g->screen_size_x -1;
            from_y = g->pos_y;
            to_y = g->pos_y + g->screen_size_y -1;
            //if (_mouse_screen == screen) trace("8");
            if ((bitmap_color_depth(window) != bitmap_color_depth(onto) || is_screen_bitmap(window)))
            {
                tmpfrom = WHAM->get_scratch_bmp(g->size_x,g->size_y,bitmap_color_depth(onto));
                blit
                (
                    window,
                    tmpfrom,
                    g->viewport_pos_x,
                    g->viewport_pos_y,
                    g->viewport_pos_x,
                    g->viewport_pos_y,
                    g->viewport_size_x,
                    g->viewport_size_y
                );
            }
            stretch_blit
            (
                tmpfrom,
                onto,
                g->viewport_pos_x,
                g->viewport_pos_y,
                g->viewport_size_x,
                g->viewport_size_y,
                from_x,
                from_y,
                to_x - from_x + 1,
                to_y - from_y + 1
            );
        }
        else
        {
            //if (_mouse_screen == screen) trace("9");
            blit
            (
                window,
                onto,
                w_from_x,
                w_from_y,
                from_x,
                from_y,
                to_x - from_x + 1,
                to_y - from_y + 1

            );
        }
    }

}



/*************************************************************
 * draws the point
 * given by the coordinates
 *************************************************************/
void Window::point(BITMAP *onto, int x, int y) // onto should be screen or buffer
{


    #ifdef WTRACE
        if (wham_trace) trace("Window::point win %d",get_geoinfo()->tag);
    #endif

    int
        w_x,
        w_y;

    //readkey();
    //adjust coordinates on screen
    GeoInfo const * g = get_geoinfo();
    
    
    // get coordinates in wobject
    w_x = g->viewport_pos_x + (int)((x - g->pos_x) * (float)g->viewport_size_x / (float)g->screen_size_x);
    w_y = g->viewport_pos_y + (int)((y - g->pos_y) * (float)g->viewport_size_y / (float)g->screen_size_y);
    

    
    if (w_x >= g->viewport_pos_x && w_y >= g->viewport_pos_y && w_x < g->viewport_pos_x + g->viewport_size_x && w_y < g->viewport_pos_y + g->viewport_size_y )
    {
       //if (_mouse_screen == screen) trace("9");
       putpixel
       (
          onto,
          x,
          y,
          getpixel
          (
            window,
            w_x,
            w_y
          )
       );
    }

}




/*************************************************************
 * set the size & position of the user area
 *************************************************************/
void Window::set_user_area(int _pos_x,int _pos_y,int _size_x, int _size_y)
{
    #ifdef WTRACE
        if (wham_trace) trace("Window::set_user_area");
    #endif

    // get everything within boundaries with some ternair monsters
    _pos_x = _pos_x < 0 ? 0 : _pos_x >= get_geoinfo()->size_x ? get_geoinfo()->size_x-1 : _pos_x;
    _pos_y = _pos_y < 0 ? 0 : _pos_y >= get_geoinfo()->size_y ? get_geoinfo()->size_y-1 : _pos_y;
    _size_x = _size_x < 0 ? 0 : _size_x + _pos_x >= get_geoinfo()->size_x ? get_geoinfo()->size_x - _pos_x: _size_x;
    _size_y = _size_y < 0 ? 0 : _size_y + _pos_y >= get_geoinfo()->size_y ? get_geoinfo()->size_y - _pos_y: _size_y;

    Wobject::set_user_area(_pos_x, _pos_y,_size_x,_size_y);

    destroy_bitmap(user);
    user = create_sub_bitmap(window,  get_geoinfo()->user_pos_x,  get_geoinfo()->user_pos_y, get_geoinfo()->user_size_x, get_geoinfo()->user_size_y);
}





/*************************************************************
 * resize the bitmap, note that is means making a new bitmap
 * and copying the old one
 * the information inside the window is not rescaled
 * use rescale for that
 *************************************************************/
int Window::resize(int nw_size_x,int nw_size_y)
{
    #ifdef WTRACE
        if (wham_trace) trace("Window::resize");
    #endif

    if (get_geoinfo()->size_is_locked) return FALSE;



    Wobject::resize(nw_size_x,nw_size_y);
    nw_size_x = get_geoinfo()->size_x;    
    nw_size_y = get_geoinfo()->size_y;    

    BITMAP *nw = create_bitmap(nw_size_x, nw_size_y);
    if (!nw) 
         out_of_mem("creating Window bitmap for resizing");

    blit(window, nw, 0,0,0,0,nw_size_x, nw_size_y);


    destroy_bitmap(user);
    if (must_delete_bitmap)
        destroy_bitmap(window);
    window = nw;
    user = create_sub_bitmap(window,  get_geoinfo()->user_pos_x,  get_geoinfo()->user_pos_y, get_geoinfo()->user_size_x, get_geoinfo()->user_size_y);
    

    Wobject::notify_children();


    return TRUE;// return success
}



BITMAP *Window::bmp() 
{
    return user;
}


int Window::printf(char *format,...)
{
    static char buffer[2048];

    va_list arg;
    va_start(arg,format);
    int ret = vsprintf(buffer,format,arg);
    va_end(arg);

    char *ent = buffer;
    char *head = buffer;
    int nr_of_enters = 0;
    while(*ent)
    {
        if (*ent == 10) nr_of_enters++;
        ent++;
    }
    ent = buffer;
    text_mode(background_color);
    solid_mode();
    for (int i=0;i<nr_of_enters;i++)
    {
        // look up enter
        while(*ent++ != 10);
        *(--ent) = 0;
        textout(user,font,head,text_cursor_x,text_cursor_y,Text_Color);
        head = ++ent;
        text_cursor_x = 1;
        text_cursor_y += text_height(font) + 2;
        if (text_cursor_y>get_geoinfo()->size_y - text_height(font))
        {
            if (Scroll_Behaviour == WINDOW_SCROLL_OVERLAY)
                text_cursor_y=1;
            else
            {
                text_cursor_y -= (text_height(font )+ 2);
                blit(user,user,0,text_height(font)+2,0,0,user->w, text_cursor_y );
                rectfill(user,0,text_cursor_y ,user->w -  1,text_cursor_y +  (text_height(font) + 2),background_color);
            }
        }
    }
    textout(user,font,head,text_cursor_x,text_cursor_y, Text_Color);
    text_cursor_x += text_length(font, head);

    return ret;
}

int Window::Text_Color = 255;
int Window::Scroll_Behaviour = WINDOW_SCROLL_OVERLAY;
int Window::Allocate_From_VRAM = 0;

Wobject *Window::process_mouse(int *x_out, int *y_out)
{
    Wobject *win;
    int out_x,out_y;
    int flag;
    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);
    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;
    if (win != this) return win;

    if (win && (mouse_b & 1))
    {
        win->raise();
    }
    return (Wobject *)win;
}

void Window::use_palette(RGB *_pal, int free_on_deletion)
{
   pal = _pal;
   free_palette_on_deletion = free_on_deletion;
}


RGB *Window::get_used_palette()
{
   return pal;
}

int Window::process_message(Wobject *from, int msg)
{

   return Wobject::process_message(from,msg);
}
