/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/


// wobject_order_list[i] : i = tag, array gives place in hierarchy for each window
// order_wobject_list[i] : i = place in hierarchy, array gives tag

#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <new.h>
#include "../include/helpers.h"
#include "../include/winsys.h"
#include "../include/wversion.h"
#ifdef ALLEGRO_DOS
#include <io.h>
#endif

static char wham_signal_msg[] = "Closed off Wham & Allegro \r\n";
static char wham_ctrlc_msg[] = "Closed off Wham & Allegro \r\nctrl-c pressed \r\n";
static char wham_sigsegv_msg[] = "Closed off Wham & Allegro, possibly garbage left in memory. \r\n";
static char wham_new_msg[] = "Out of memory in 'new' call. \r\n";

int wham_trace = TRUE;
extern int _color_depth;  // from allegro file graphics.c

static int nested_signal = FALSE;

static void signal_handler(int num)
{
   if (nested_signal)
   {
      WHAM = NULL; // make sure the atexit function doesn't try to remove it
      if (nested_signal > 1)
      {
         if (nested_signal > 2) // this is pretty hopeless
         {
            signal(num, SIG_DFL);// set default action
            raise(num);          // raise it again
         }
         // severe panic, try to remove at least the keyboard handler
         remove_keyboard();
         signal(num, SIG_DFL); // raise default
         raise(num);
      }
      // oops, try to remove allergo at least
      allegro_exit();
      #ifdef ALLEGRO_DOS
             _write(STDERR_FILENO, wham_signal_msg, sizeof(wham_signal_msg) - 1);
      #endif
      write(STDERR_FILENO, wham_signal_msg, sizeof(wham_signal_msg) - 1);
      signal(num, SIG_DFL);
      raise(num);
      
   }
   nested_signal++;

   if (hw_screen)
      screen = hw_screen;
      
//   trace("GOT SIGNAL %d!!!!!!!!!!!!!!",num);
// delete WHAM;
   WHAM = NULL; // exit straight away without deleting wham
   allegro_exit();

   write(STDERR_FILENO, wham_signal_msg, sizeof(wham_signal_msg) - 1);

   signal(num, SIG_DFL);
   raise(num);
}
END_OF_FUNCTION(signal_handler);


// the idea is that when a sigsegv occurs, the system might be corrupted
// so don't try to clean up wham
// but call allegro_exit and get out.

// after all this doesn't seem neccesary since it works ok using the 
// above signal handler, so give the user an option to enable it;

int WinSystem::UseSpecialSegvHandler;


static void sigsegv_handler(int num)
{

    // don't delete wham and make sure no at_exit function tries to
   WHAM = NULL;

   allegro_exit();

   write(STDERR_FILENO, wham_sigsegv_msg, sizeof(wham_sigsegv_msg) - 1);

   signal(num, SIG_DFL);
   raise(num);
}
END_OF_FUNCTION(sigsegv_handler);


static void ctrlc_handler(int num)
{
   num = 0;

   if (WHAM->hw_screen)
      screen = WHAM->hw_screen;

   text_mode(makecol(0,0,0));
   textout_centre(screen,font,"CONTROL-C pressed!",SCREEN_W/2, SCREEN_H/2,makecol(255,255,255));
   rest(500);

   
   delete WHAM;
   WHAM = NULL;
   set_gfx_mode(GFX_TEXT,0,0,0,0);
// allegro_exit();

   write(STDERR_FILENO,wham_ctrlc_msg , sizeof(wham_ctrlc_msg) - 1);
    exit(1);
}
END_OF_FUNCTION(ctrlc_handler);


static void my_new_handler()
{
   if (WHAM->hw_screen)
      screen = WHAM->hw_screen;

   delete WHAM;
   WHAM = NULL;

   allegro_exit();

    write(STDERR_FILENO,wham_new_msg , sizeof(wham_new_msg) - 1);
    exit(1);
}
END_OF_FUNCTION(my_new_handler);


extern BITMAP *_mouse_screen;

static BITMAP *create_buffer(int w,int h)
{
   BITMAP *buffer;
   if (gfx_capabilities & GFX_HW_VRAM_BLIT)
   {
      if ( (gfx_capabilities & GFX_HW_MEM_BLIT) || WinSystem::UseBoxes)
      {  // try to get a VRAM buffer

         buffer = create_video_bitmap(SCREEN_W,SCREEN_H);
         if (!buffer)
            buffer = create_bitmap(w,h);
      }
      else
         buffer = create_bitmap(w,h);
      
   }
   else
      buffer = create_bitmap(w,h);

   return buffer;
}


/*************************************************************
 * the initial size of the bookkeeping tables
 *************************************************************/
int WinSystem::InitialBufferSize = 100;

/*************************************************************
 * the size of the increments used when the buffer is exceeded
 * (if the above default is used: when more than 100 wobjects
 * are opened at once)
 *************************************************************/
int WinSystem::Buffer_Increment = 100;

/*************************************************************
 * the size of the increments used when the delayed message mailbin if full
 *************************************************************/
int WinSystem::Mailbin_Increment = 100;

/*************************************************************
 * if this is set WinSystem will do all operations on a buffer
 * an blit the result to the screen
 *************************************************************/
int WinSystem::Buffered = 0;

/*************************************************************
 * flag to count the number of WinSystems
 *************************************************************/
int WinSystem::winsys_count = 0;



/*************************************************************
 * version
 *************************************************************/
char *WinSystem::version= WHAM_VERSION_STRING;


BITMAP *WinSystem::hw_screen;

/*************************************************************
 * background color
 *************************************************************/
int WinSystem::BackgroundColor= -1;

/*************************************************************
 * when set, lets winsys wait before each screen draw
 * only works in buffered mode
 *************************************************************/
int WinSystem::Vsynced=FALSE;

/*************************************************************
 * pointers to the bitmaps used for output
 *************************************************************/
BITMAP *WinSystem::TheScreen = NULL;


/*************************************************************
 * obsolete, maintained for campatibility reasons
 * the value is ignored
 *************************************************************/
int WinSystem::Turn_Off_Mouse = 0;


/*************************************************************
 * minimum sizes, must be >0
 *************************************************************/
int WinSystem::Minimum_Size_x = 32;
int WinSystem::Minimum_Size_y = 32;


/*************************************************************
 * shadow stuff
 *************************************************************/
int WinSystem::ShadowStyle = WHAM_NO_SHADOW;
int WinSystem::ShadowColor = 128;
int WinSystem::ShadowSize  = 4;

/************************************************************
 * use boxlist boxes
 ***************************************************************/
int WinSystem::UseBoxes = FALSE;


/************************************************************
 * first draw a solid rect before rebuilding screen
 ***************************************************************/
int WinSystem::Full_Remove = TRUE;

/************************************************************
 * cool effects, only work in buffered mode
 ***************************************************************/
int WinSystem::Show_Effect = 0;
int WinSystem::Show_Effect_Param = 10;




/************************************************************
 * if true message system it working, else it is ignored
 ***************************************************************/
int WinSystem::Message_System_Active = TRUE;


/************************************************************
 * scratch buffer 
 ***************************************************************/
BITMAP *WinSystem::scratch_bmp = NULL;

/*************************************************************
 * claimer
 *************************************************************/
Stack *WinSystem::coordinate_claimer_stack = NULL;
Stack *WinSystem::keyboard_claimer_stack = NULL;
Wobject *WinSystem::current_coordinate_claimer = NULL;
Wobject *WinSystem::current_keyboard_claimer = NULL;

/*************************************************************
 * constructor
 *************************************************************/
WinSystem::WinSystem()
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem constructor 1");
    #endif
    parent = NULL;
    create();
}

WinSystem::WinSystem(Wobject *_parent)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem constructor 2");
    #endif

    parent = _parent;
    if (parent->Child_manager)
        fatal("WHAM: Multiple child wobject managers tied to one wobject.");
    parent->Child_manager = this;
    
    create();
}


static int __set_gfxmode_ex(int card,int w,int h, int depth)
{
      int ret;
      set_color_depth(depth);
      set_gfx_mode(GFX_TEXT,0,0,0,0);
      ret = set_gfx_mode(card,w,h,0,0);

      return ret;
      
}



static void _get_current_video_config(int *card, int *w, int *h, int *depth)
{
    *depth = bitmap_color_depth(screen); // record current color depth
    *card = -1;
    for (int i=0;_gfx_driver_list[i].driver;i++) // loop through drivers searching current one
    {
      if (_gfx_driver_list[i].driver  == gfx_driver)
         *card = i;
    }
    
    if (*card == -1) // could not determine which driver is in use now
    {
      *w = 320;
      *h = 200;
      *depth = 8;
    }
    else
    {
      *w = SCREEN_W;
      *h = SCREEN_H;
    }


}

int WinSystem::switch_gfx_mode(int card, int w , int h, int color_depth)
{
    if (!is_same_bitmap(this_managers_screen, screen))
        return - 1;  // doesn't output to hardware screen

    int cur_card, cur_w,cur_h,cur_depth;
    int ret = 0;
    PALLETE old; 
    get_palette(old);

    show_mouse(NULL);


    remove_mouse();


    _get_current_video_config(&cur_card,&cur_w,&cur_h,&cur_depth);

    if (color_depth)
        set_color_depth(color_depth);

    if (must_delete_TheScreen)
       destroy_bitmap(TheScreen);
    

    if (set_gfx_mode(card,w,h,0,0))
    {
     ret = -1;
     set_color_depth(cur_depth);
     set_gfx_mode(cur_card,cur_w,cur_h,0,0);
    }

    set_palette(old);
    clear_to_color(screen,_BackgroundColor);
    ::hw_screen = screen;
    hw_screen = screen;
    TheScreen = create_video_bitmap(SCREEN_W, SCREEN_H);
    must_delete_TheScreen = TRUE;
    show_video_bitmap(TheScreen);
    this_managers_screen = TheScreen;
    TheScreen_size_x = SCREEN_W;
    TheScreen_size_y = SCREEN_H;
    
    if (_Buffered)
    {
        destroy_bitmap(buffer);
        buffer = create_buffer(TheScreen_size_x,TheScreen_size_y);
        if (!buffer)
            out_of_mem("creating drawing buffer");
        clear_to_color(buffer,_BackgroundColor);
    }
    refresh_all_visible_objects();
    install_mouse();
    show_mouse(screen);

    return ret;
}

char *WinSystem::get_version()
{
    return version;
}

void WinSystem::let_user_choose_gfx_mode()
{
/*
    if (!is_same_bitmap(this_managers_screen, screen))
        return;  // doesn't output to hardware screen

    int card ,w,h,coldepth;
    int cur_card = 0,cur_w=800,cur_h=600,cur_depth=8;
    char tbuffer[1024];
    
    _get_current_video_config(&cur_card,&cur_w,&cur_h,&cur_depth);
    if (cur_card < 0)
       cur_card = 0;
    PALLETE old; 
    get_palette(old);
    gui_bg_color = makecol(255,255,255);
    gui_fg_color = makecol(0,0,0);

    while (!gfx_mode_select_ex(&card,&w,&h, &coldepth)); // get initial choice
    show_mouse(NULL); // make SURE it's not displayed

    if (must_delete_TheScreen)
       destroy_bitmap(TheScreen);

    remove_mouse();
    
    while (__set_gfxmode_ex(card,w,h,coldepth))
    {
        set_color_depth(cur_depth);
        set_gfx_mode(cur_card,cur_w,cur_h,0,0);
        clear_to_color(screen,_BackgroundColor);
        set_palette(old);
        sprintf(tbuffer,"can't switch to %s,%dx%dx%d",card >0 ? _gfx_driver_list[card].driver->name : "autodetect",w,h,coldepth);
        install_mouse();
        show_mouse(screen);
        alert(tbuffer,allegro_error,"select new mode","OK",NULL,13,0);
        while (!gfx_mode_select_ex(&card,&w,&h,&coldepth));
        show_mouse(NULL);
        remove_mouse();
        set_color_depth(coldepth);
    }
    set_palette(old);
    clear_to_color(screen,_BackgroundColor);

    TheScreen = create_video_bitmap(SCREEN_W,SCREEN_H);
    must_delete_TheScreen = TRUE;
    show_video_bitmap(TheScreen);
    this_managers_screen = TheScreen;
    TheScreen_size_x = SCREEN_W;
    TheScreen_size_y = SCREEN_H;
    hw_screen = screen;
    ::hw_screen = screen;
    
    if (_Buffered)
    {
        destroy_bitmap(buffer);
        buffer = create_buffer(TheScreen_size_x,TheScreen_size_y);
        if (!buffer) 
            out_of_mem("creating drawing buffer");
        clear_to_color(buffer,_BackgroundColor);
    }
    install_mouse();
    refresh_all_visible_objects();
    show_mouse(screen);
*/
}

static int signal_is_initialized = FALSE;

void WinSystem::create()
{
    if (!signal_is_initialized)
    {
        // set signal handler
        LOCK_FUNCTION(signal_handler);
        LOCK_FUNCTION(sigsegv_handler);
        LOCK_FUNCTION(ctrlc_handler);
        LOCK_FUNCTION(my_new_handler);
        LOCK_VARIABLE(wham_signal_msg);
        LOCK_VARIABLE(wham_sigsegv_msg);
        LOCK_VARIABLE(wham_ctrlc_msg);
        LOCK_VARIABLE(WHAM);
        LOCK_VARIABLE(WinSystem::Mouse_Cursor0);
        LOCK_VARIABLE(WinSystem::Mouse_Cursor1);
        LOCK_VARIABLE(WinSystem::Mouse_Cursor2);
        LOCK_VARIABLE(WinSystem::Mouse_Cursor3);
        LOCK_VARIABLE(WinSystem::Mouse_Cursor4);
        LOCK_VARIABLE(WinSystem::Mouse_Cursor5);
        LOCK_VARIABLE(WinSystem::Mouse_Cursor6);
        LOCK_VARIABLE(WinSystem::Mouse_Cursor7);
        LOCK_VARIABLE(WinSystem::Mouse_Cursor8);
        LOCK_VARIABLE(WinSystem::Mouse_Cursor9);
        signal(SIGABRT, signal_handler);
        signal(SIGFPE,  signal_handler);
        signal(SIGILL,  signal_handler);
        if (!UseSpecialSegvHandler)
            signal(SIGSEGV, signal_handler);
        else
            signal(SIGSEGV, sigsegv_handler);

        signal(SIGTERM, signal_handler);
        signal(SIGINT,  ctrlc_handler);
        signal(SIGKILL, signal_handler);
        signal(SIGQUIT, signal_handler);
        set_new_handler(my_new_handler);
        signal_is_initialized = TRUE;
        hw_screen = screen;
        ::hw_screen = screen;
    }

    set_config_string("#wham","version",version);
    
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::create");
    #endif
    is_being_deleted = FALSE;

    if (!(Wham_Secretary.initialized))
      Wham_Secretary.create();
    if (!hw_screen && screen)
       hw_screen = screen;

    winsys_count++;
    must_delete_TheScreen = FALSE;
    if (!TheScreen)
    {
      must_delete_TheScreen = TRUE;

      TheScreen = create_video_bitmap(SCREEN_W,SCREEN_H);

      if (!TheScreen)
         fatal("Could not allocate display: insufficient video memory");

      show_video_bitmap(TheScreen);
    }
     show_mouse(screen);
    _Buffered = Buffered;            // if on, do all operations on a buffer and blit the result to TheScreen
    _ShadowStyle = ShadowStyle;        // determines the type of shadowing used
    _ShadowColor = ShadowColor;        // determines the type of shadowing used
    _ShadowSize = (_ShadowStyle ? ShadowSize : 0);         // determines the type of shadowing used
    _BackgroundColor = (BackgroundColor > 0) ? BackgroundColor : 1;
    if (_ShadowStyle == WHAM_SLOW_SHADOW && bitmap_color_depth(TheScreen) != 8)
      fatal("Translucent shadows only supported in 8 bit videomode");
      
    this_managers_screen = TheScreen;
    if (parent)
        this_managers_screen = parent->bmp();

    if (!this_managers_screen)
        fatal("WinSystem: manager tied to a wObject which is not fit for drawing onto");
    // filter out nonsense values
    if (!is_same_bitmap(this_managers_screen, screen))
    {
        _Buffered = FALSE;
    }

    #ifdef DEBUGBOXLIST
      _Buffered = FALSE;
    #endif

    wobject_list = (Wobject **)malloc(sizeof(Wobject *) * InitialBufferSize);
    wobject_order_list = (int *)malloc(sizeof(int) * InitialBufferSize);
    order_wobject_list = (int *)malloc(sizeof(int) * InitialBufferSize);
    if (!(wobject_list && wobject_order_list && order_wobject_list))
        out_of_mem("allocating administration tables, reduce InitialBuffersize");

    mailbin_size = Mailbin_Increment;
    mailbin = (Message *)malloc(mailbin_size * sizeof(Message));
    if (!mailbin)
      fatal("Could not allocate Message buffer");
    
    buffersize = InitialBufferSize;
    for (int i=0;i<buffersize;i++)
    {
        wobject_order_list[i] = 0;
        order_wobject_list[i] = 0;
        wobject_list[i] = NULL;
    }
    number_of_wobjects = 0;

    if (this_managers_screen == screen)
    {
        TheScreen_size_x = SCREEN_W;
        TheScreen_size_y = SCREEN_H;
    }
    else
    {
        TheScreen_size_x = this_managers_screen->w;
        TheScreen_size_y = this_managers_screen->h;
    }

    if (_Buffered)
    {
        buffer = create_buffer(TheScreen_size_x,TheScreen_size_y);
        if (!buffer)
            out_of_mem("creating drawing buffer");
        clear_to_color(buffer,_BackgroundColor);
    }
    mouse_off();
    clear_to_color(this_managers_screen,_BackgroundColor);
    if (_ShadowStyle == WHAM_SLOW_SHADOW)
    {
        if (!color_map)
        {
            color_map = new COLOR_MAP;
            if (!color_map)
                out_of_mem("creating shading colormap");
            PALETTE the_palette;
            get_palette(the_palette);
            create_light_table(color_map,the_palette,0,0,0,wham_trans_calc_progress);
            must_delete_colormap = TRUE;
        }
    }
    shadow_pattern = NULL;
    if (_ShadowStyle == WHAM_PATTERN_SHADOW)
    {
        shadow_pattern = create_bitmap(2,2);
        if (!shadow_pattern)
            out_of_mem("allocating 16 bytes");
        if (_color_depth == 8)
        {
            _putpixel(shadow_pattern,0,0,255);
            _putpixel(shadow_pattern,1,1,255);
            _putpixel(shadow_pattern,0,1,0);
            _putpixel(shadow_pattern,1,0,0);
        }
        else
        {
            putpixel(shadow_pattern,0,0,makecol(255,255,255));
            putpixel(shadow_pattern,1,1,makecol(255,255,255));
            putpixel(shadow_pattern,0,1,makecol(255,0,255));
            putpixel(shadow_pattern,1,0,makecol(255,0,255));
        }

    }
    clear_to_color(this_managers_screen,_BackgroundColor);
    
    mouse_on();
    x_shifted = y_shifted = 0;
    maxshiftx = maxshifty = minshiftx = minshifty = 0;

    nr_of_messages_in_mailbin = 0;
    
    if (!keyboard_claimer_stack)
       keyboard_claimer_stack = new Stack;

    if (!coordinate_claimer_stack)
       coordinate_claimer_stack = new Stack;
    
}


void WinSystem::parent_bitmap_has_changed()
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::parent_bitmap_has_changed");
    #endif

    this_managers_screen = parent->bmp();
    BITMAP *s = this_managers_screen;
    TheScreen_size_x = this_managers_screen->w;
    TheScreen_size_y = this_managers_screen->h;
    clear_to_color(s,_BackgroundColor);

    // redraw whole 'screen'
    rebuild_screen(s,0,0,TheScreen_size_x - 1, TheScreen_size_y - 1, number_of_wobjects -1);
    // no need for update, for it is never buffered if it has a parent screen
}



/*************************************************************
 * destructor
 *************************************************************/
WinSystem::~WinSystem()
{
  #ifdef WTRACE
    if (wham_trace)
      trace("WinSystem destructor");
  #endif

    is_being_deleted = TRUE;
    // close all wobjects
    broadcast_message(NULL,W_MSG_MANAGER_DELETE);

   if (WHAM->hw_screen)
      screen = WHAM->hw_screen;

    while (wobject_list[order_wobject_list[0]])
    {
         // don't bother removing visible windows
         wobject_list[order_wobject_list[0]]->geoinfo.is_displayed = FALSE;
//         trace("WINSYS is goin to del %s",wobject_list[order_wobject_list[0]]->geoinfo.name);
         delete wobject_list[order_wobject_list[0]];
    }

//    trace("deleted all objects now free remaining mem");
    free(wobject_list);
    free(wobject_order_list);
    free(order_wobject_list);
    //trace("free buffer");
    if (_Buffered)
        destroy_bitmap(buffer);
    winsys_count--;

    //trace("free mailbin");
    free(mailbin);
    if (this == WHAM)
    {

        if (must_delete_colormap)
        {
            delete color_map;
            color_map = NULL;
        }
        //trace("destroy wham secretary");
        Wham_Secretary.destroy();
        //trace("destroy scratch bitmap");
        if (scratch_bmp)
                destroy_bitmap(scratch_bmp);

        //trace("cleanup claimer stacks");
        while (current_keyboard_claimer)
        {
            current_keyboard_claimer =(Wobject *)(keyboard_claimer_stack->pop());
        }
        while (current_coordinate_claimer)
        {
            current_coordinate_claimer =(Wobject *)(coordinate_claimer_stack->pop());
        }
        delete keyboard_claimer_stack;
        delete coordinate_claimer_stack;
    }
    //trace("ok done");
}

/*************************************************************
 * return a new unused tag and register the wobject
 * also sets the wobject to be th top wobject
 *************************************************************/
int WinSystem::get_new_tag(Wobject *w)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::get_new_tag");
    #endif

    int number_of_locked_top_wobjects = get_number_of_locked_top_wobjects();
    int tag;
    if (number_of_wobjects >= buffersize - 1)
    {
        buffersize += Buffer_Increment;
        wobject_list = (Wobject **)realloc(wobject_list, sizeof(Wobject *) * buffersize);
        wobject_order_list = (int *)realloc(wobject_order_list, sizeof(int) * buffersize);
        order_wobject_list = (int *)realloc(order_wobject_list, sizeof(int) * buffersize);

        if (!(wobject_list && wobject_order_list && order_wobject_list))
            out_of_mem("enlarging administration tables, too many windows");

        for (int i = buffersize - Buffer_Increment;i<buffersize;i++)
            wobject_list[i] = NULL;


    }

    // the new wobject is always on top, unless the top wobject is locked
    // shift all other wobjects down in hierarchy
    for (int i=number_of_wobjects ;i > number_of_locked_top_wobjects;i--)
    {
        order_wobject_list[i] = order_wobject_list[i - 1];
        wobject_order_list[order_wobject_list[i]] = i;
    }
    tag = get_free_tag();

    wobject_list[tag] = w;
    wobject_order_list[tag] = number_of_locked_top_wobjects;
    order_wobject_list[number_of_locked_top_wobjects] = tag;
    number_of_wobjects++;
    order_wobject_list[number_of_wobjects] = 0;
    wobject_order_list[0] = number_of_wobjects;
    return tag;
}


int WinSystem::get_number_of_locked_top_wobjects()
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::get_number_of_locked_top_wobjects");
    #endif

    int i=0;
    while(1)
    {
        if (i >= number_of_wobjects)
            return 0;
        if (!wobject_list[order_wobject_list[i]]->geoinfo.is_locked)
            return i;
        i++;
    };
}


int WinSystem::get_first_free_bottom_place()
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::get_number_of_locked_bottom_wobjects");
    #endif

    int i=number_of_wobjects - 1;
    while(1)
    {
        if (i < 0)
            return 0;
        if (!wobject_list[order_wobject_list[i]]->geoinfo.is_locked)
            return i;
        i--;
    };
}


/*************************************************************
 * remove this wobject from the system(does not delete the
 * wobject itself)
 *************************************************************/
void WinSystem::destroy_tag(int tag)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::destroy_tag %d",tag);
    #endif


    number_of_wobjects--;

    for (int i =  wobject_order_list[tag]; i < number_of_wobjects; i++)
    {
        //trace("shifting %s from %d to %d",wobject_list[order_wobject_list[i+1]]->get_geoinfo()->name,i+1,i);
        order_wobject_list[i] = order_wobject_list[i+1];
        wobject_order_list[order_wobject_list[i]] = i;
    }
    //trace("till h");
    order_wobject_list[number_of_wobjects] = 0;
    wobject_order_list[0] = number_of_wobjects;
    wobject_list[tag] = NULL;
    //trace("anfd kljk");
}

/*************************************************************
 * remove the wobject from the screen
 *************************************************************/
void WinSystem::remove_wobject(int tag)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::remove_wobject");
    #endif
    if (is_being_deleted)
       return;

    int from_x,from_y,to_x,to_y;
    
    BITMAP *s = _Buffered ? buffer : this_managers_screen;

    Wobject *the_wobject = wobject_list[tag];
    if (the_wobject->frozen)
        return;

    if (!the_wobject->geoinfo.is_displayed)
        return;
    

    from_x = the_wobject->geoinfo.pos_x;
    from_y = the_wobject->geoinfo.pos_y;
    to_x = the_wobject->geoinfo.pos_x + the_wobject->geoinfo.screen_size_x;
    to_y = the_wobject->geoinfo.pos_y + the_wobject->geoinfo.screen_size_y;
    

    if (_ShadowStyle)
    {
        to_x += _ShadowSize;
        to_y += _ShadowSize;
    }

    from_x -= the_wobject->geoinfo.Left_Border_Size;
    from_y -= the_wobject->geoinfo.Upper_Border_Size;
    to_x += the_wobject->geoinfo.Right_Border_Size;
    to_y += the_wobject->geoinfo.Lower_Border_Size;

    from_x = MAX(from_x,0);
    from_y = MAX(from_y,0);
    to_x = MIN(to_x, TheScreen->w);
    to_y = MIN(to_y, TheScreen->h);

    if (from_x > to_x || from_y > to_y)
    {
//       trace("remove %d %d %d %d  NOT",from_x,from_y,to_x,to_y);
       return;
    }
    if (!_Buffered)
    {
        mouse_off(from_x, from_y,to_x,to_y);
    }

    if (Full_Remove)
    {
        drawing_mode(DRAW_MODE_SOLID,0,0,0);
        if (!UseBoxes)
        {
            //if (_mouse_screen == screen) if (wham_trace) trace("1");
            rectfill(s,from_x,from_y,to_x-1 ,to_y-1 ,_BackgroundColor);
        }
        else
            the_wobject->my_boxlist->draw_the_boxes(s,from_x,from_y,to_x,to_y,_BackgroundColor);
    }
    the_wobject->geoinfo.is_displayed = FALSE;
    if (!UseBoxes)
        // display all the wobjects from bottom to top
        rebuild_screen(s,from_x,from_y,to_x,to_y, number_of_wobjects - 1);
    else
    {
        // redisplay all object below this one
        update_boxlists(wobject_order_list[tag]);
        rebuild_screen_using_boxes(s,from_x,from_y,to_x,to_y,number_of_wobjects -1,wobject_order_list[tag]);
    }
    
    // this TRUE is correct, you only come here when the window
    // is displayed , but it was set temporarily to false for the 
    // update_boxlist, the administration is done elsewhere
    the_wobject->geoinfo.is_displayed = TRUE;
    
    if (!_Buffered)
    {
       mouse_on();
    }

}
/*************************************************************
 * remove part of the wobject from the screen
 *************************************************************/
void WinSystem::remove_wobject(int tag, int xfrom,int yfrom, int xto, int yto)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::remove_wobject");
    #endif
    if (is_being_deleted)
       return;

    int from_x,from_y,to_x,to_y;
    
    BITMAP *s = _Buffered ? buffer : this_managers_screen;

    Wobject *the_wobject = wobject_list[tag];
    if (the_wobject->frozen)
        return;

    if (!the_wobject->geoinfo.is_displayed)
        return;
    

    from_x = the_wobject->geoinfo.pos_x;
    from_y = the_wobject->geoinfo.pos_y;
    to_x = the_wobject->geoinfo.pos_x + the_wobject->geoinfo.screen_size_x;
    to_y = the_wobject->geoinfo.pos_y + the_wobject->geoinfo.screen_size_y;
    

    if (_ShadowStyle)
    {
        to_x += _ShadowSize;
        to_y += _ShadowSize;
    }

    from_x -= the_wobject->geoinfo.Left_Border_Size;
    from_y -= the_wobject->geoinfo.Upper_Border_Size;
    to_x += the_wobject->geoinfo.Right_Border_Size;
    to_y += the_wobject->geoinfo.Lower_Border_Size;

    from_x = MAX(from_x,xfrom);
    from_y = MAX(from_y,yfrom);
    to_x = MIN(to_x,xto);
    to_y = MIN(to_y,yto);

    from_x = MAX(from_x,0);
    from_y = MAX(from_y,0);
    to_x = MIN(to_x, TheScreen->w);
    to_y = MIN(to_y, TheScreen->h);

    if (from_x > to_x || from_y > to_y)
    {
//       trace("remove pt %d %d %d %d  NOT",from_x,from_y,to_x,to_y);
       return;
    }


    if (!_Buffered && !UseBoxes)
    {
         mouse_off(from_x,from_y,to_x,to_y);
    }

    if (Full_Remove)
    {
        drawing_mode(DRAW_MODE_SOLID,0,0,0);
        if (!UseBoxes)
        {
            //if (_mouse_screen == screen) if (wham_trace) trace("3");
            rectfill(s,from_x,from_y,to_x - 1,to_y - 1,_BackgroundColor);
        }
        else
        {
            the_wobject->my_boxlist->draw_the_boxes(s,from_x,from_y,to_x,to_y,_BackgroundColor);
        }
    }
    the_wobject->geoinfo.is_displayed = FALSE;
    if (!UseBoxes)
    {
        // display all the wobjects from bottom to top
        rebuild_screen(s,from_x,from_y,to_x,to_y, number_of_wobjects - 1);
    }
    else
    {
        // redisplay all object below this one
        update_boxlists(wobject_order_list[tag]);

        if (_ShadowStyle == 0)
        {
         rebuild_screen_using_boxes(s,from_x,from_y,to_x,to_y, number_of_wobjects - 1 ,wobject_order_list[tag]);
        }
        else
        {
         rebuild_screen_using_boxes(s,from_x,from_y,to_x,to_y, number_of_wobjects - 1 , 0);
        }
        
    }
    
    // this TRUE is correct, you only come here when the window
    // is displayed , but it was set temporarily to false for the 
    // update_boxlist
    the_wobject->geoinfo.is_displayed = TRUE;
    
    if (!_Buffered && !UseBoxes)
    {
         mouse_on();
    }
}

/*************************************************************
 * display/update the wobject on the screen
 *************************************************************/
void WinSystem::display_wobject(int tag, int mode)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::display_wobject");
    #endif

    
    int from_x,from_y,to_x,to_y;
    BITMAP *s = _Buffered ? buffer : this_managers_screen;

    Wobject *the_wobject = wobject_list[tag];

    if (!the_wobject)
    {
      trace("tried to display non existing object %d",tag);
      return;
    }
    if (the_wobject->frozen)
        return;
    if (the_wobject->is_being_deleted)
    {
        #ifdef WTRACE
            if (wham_trace) trace("WinSystem::is_being_deleted:do_not_display");
        #endif
        return;
    }
    
    from_x = the_wobject->geoinfo.pos_x;
    from_y = the_wobject->geoinfo.pos_y;
    to_x = the_wobject->geoinfo.pos_x + the_wobject->geoinfo.screen_size_x;
    to_y = the_wobject->geoinfo.pos_y + the_wobject->geoinfo.screen_size_y;

    from_x = MAX(from_x,0);
    from_y = MAX(from_y,0);
    to_x = MIN(to_x, TheScreen->w);
    to_y = MIN(to_y, TheScreen->h);



    int wobject_to_start_redrawing = wobject_order_list[tag];
    int was_displayed = the_wobject->geoinfo.is_displayed;
    the_wobject->geoinfo.is_displayed = TRUE;
        
    if (!was_displayed)
    {
        if (UseBoxes)
        {
            update_boxlists(wobject_to_start_redrawing);
        }
    }

        // see_through means all the objects below this one need 
        // to redrawn to ensure no remains of higher lying objects
        // are visible
    if (the_wobject->see_through && from_x < to_x && from_y < to_y)
    {

            wobject_to_start_redrawing = number_of_wobjects -  1;
            drawing_mode(DRAW_MODE_SOLID,0,0,0);
            if (Full_Remove)
            {
               if (_ShadowStyle || !UseBoxes)
               {
                  if (!_Buffered)
                     mouse_off(from_x,from_y,to_x,to_y);
                  rectfill(s,from_x,from_y,to_x -1,to_y-1,_BackgroundColor);
                  if (!_Buffered)
                     mouse_on();
               }
               else
               {
                  the_wobject->my_boxlist->draw_the_boxes(s,from_x,from_y,to_x,to_y,_BackgroundColor);
               }
            }
    }
    if (mode == DISPLAY_MODE_FULL || mode == DISPLAY_MODE_BORDER)
    {
        from_x -= the_wobject->geoinfo.Left_Border_Size;
        from_y -= the_wobject->geoinfo.Upper_Border_Size;
        to_x += the_wobject->geoinfo.Right_Border_Size;
        to_y += the_wobject->geoinfo.Lower_Border_Size;
        from_x = MAX(from_x,0);
        from_y = MAX(from_y,0);
        to_x = MIN(to_x, TheScreen->w);
        to_y = MIN(to_y, TheScreen->h);

        // how many wobjects need redrawing depends on the shadow style 
        // complex shadowing means redrawing 
        // all the wobjects in the area of the shadow
        // no shadow or simple shadow means only the wobjects above the current one need redrawing
        if (_ShadowStyle)
        {

            if (_ShadowStyle >0)
            {
                if (!_Buffered || from_x < to_x || from_y < to_y)
                    mouse_off(from_x,from_y,to_x+_ShadowSize,to_y+_ShadowSize);

                wobject_to_start_redrawing = number_of_wobjects -  1;
                drawing_mode(DRAW_MODE_SOLID,0,0,0);
                if (Full_Remove) 
                {
                    rectfill(s,to_x + 1,from_y,to_x + _ShadowSize-1,to_y-1,_BackgroundColor);
                    rectfill(s,from_x,to_y + 1,to_x + _ShadowSize-1,to_y + _ShadowSize-1,_BackgroundColor);
                }
                if (!_Buffered || from_x < to_x || from_y < to_y)
                   mouse_on();
            }
            to_x += _ShadowSize;
            to_y += _ShadowSize;
        }
    } // end if DRAW_MODE_FULL

    if (!UseBoxes)    // display all the wobjects from this one to top
    {
        if (!_Buffered)
            mouse_off(from_x,from_y,to_x,to_y);
        rebuild_screen(s,from_x,from_y,to_x,to_y, wobject_to_start_redrawing);
        if (!_Buffered)
            mouse_on();
    }
    else // or be a little smarter
    {
        if (_ShadowStyle>0 || the_wobject->see_through)
        {
           if (_ShadowStyle)
           {
            rebuild_screen_using_boxes(s,from_x,from_y,to_x,to_y,number_of_wobjects -1,0);
           }
           else
           {
            rebuild_screen_using_boxes(s,from_x,from_y,to_x,to_y, number_of_wobjects - 1,wobject_order_list[tag]);
            redraw_higher_lying_see_through_objects(s,from_x,from_y,to_x,to_y,the_wobject);
           }
        }
        else
        {
            the_wobject->my_boxlist->draw_the_boxes(s,from_x,from_y,to_x,to_y);
            if (_ShadowStyle)
                redraw_higher_lying_objects(s,from_x,from_y,to_x,to_y,the_wobject);
            else
                redraw_higher_lying_see_through_objects(s,from_x,from_y,to_x,to_y,the_wobject);
        }
    }
    the_wobject->geoinfo.is_displayed = was_displayed;
}

/*************************************************************
 * display/update part of the wobject on the screen
 * coordinates are screen coordinates ,not window coordinates
 *************************************************************/
void WinSystem::display_wobject(int tag, int from_x, int from_y, int to_x, int to_y)
{
  #ifdef WTRACE
        if (wham_trace) trace("WinSystem::display_wobject_partial");
  #endif

    BITMAP *s = _Buffered ? buffer : this_managers_screen;
    Wobject *the_wobject = wobject_list[tag];
    GeoInfo g = the_wobject->geoinfo;    
    
    if (the_wobject->frozen)
        return;
    
//  if (!g.is_displayed)
//      trace("window %s is partially drawn while it was not visible",g.name);
    
    from_x = MAX(from_x,0);
    from_y = MAX(from_y,0);
    to_x = MIN(to_x, TheScreen->w);
    to_y = MIN(to_y, TheScreen->h);

    if (from_x >= to_x || from_y >= to_y)
    {
//       trace("display pt %d %d %d %d  NOT",from_x,from_y,to_x,to_y);
       return;
    }

    int wobject_to_start_redrawing = wobject_order_list[tag];

    if (the_wobject->see_through)
    {
        wobject_to_start_redrawing = number_of_wobjects -  1;
        if (Full_Remove)
        {
            if (_ShadowStyle || !UseBoxes)
            {
              drawing_mode(DRAW_MODE_SOLID,0,0,0);
              if (!_Buffered)
               mouse_off(from_x,from_y,to_x,to_y);
              rectfill(s,from_x,from_y,to_x -1,to_y-1,_BackgroundColor);
              if (!_Buffered)
               mouse_on();
            }
            else
            {

                  the_wobject->my_boxlist->draw_the_boxes(s,from_x,from_y,to_x,to_y,_BackgroundColor);
            }

        }
    }

    // display all the wobjects from this one to top, or bottom to top

    if (!UseBoxes)
    {
        if (!_Buffered)
          mouse_off(from_x, from_y,to_x,to_y);
        rebuild_screen(s,from_x,from_y,to_x,to_y, wobject_to_start_redrawing);
        if (!_Buffered)
          mouse_on();

    }
    else
    {
        if (the_wobject->see_through)
        {
         if (_ShadowStyle == 0)
         {
            rebuild_screen_using_boxes(s,from_x,from_y,to_x,to_y, number_of_wobjects -1,wobject_order_list[tag]);
            redraw_higher_lying_see_through_objects(s,from_x,from_y,to_x,to_y,the_wobject);
         }
         else
         {
            rebuild_screen_using_boxes(s,from_x,from_y,to_x,to_y, number_of_wobjects -1, 0);
         }
        }
        else
        {
         if (_ShadowStyle == 0)
         {
            the_wobject->my_boxlist->draw_the_boxes(s,from_x,from_y,to_x,to_y);
            redraw_higher_lying_see_through_objects(s,from_x,from_y,to_x,to_y,the_wobject);
         }
         else
         {
            rebuild_screen_using_boxes(s,from_x,from_y,to_x,to_y, number_of_wobjects -1, 0);
         }
            
        }
    }

    update_screen
    (
        from_x,
        from_y,
        to_x - from_x,
        to_y - from_y
    );
}

/*************************************************************
 * look for a free position in the wobject-lists
 * never using tag 0 for that is a dummy for non existing windows
 * it is used as sentinel in the order_wobject_list
 *************************************************************/
int WinSystem::get_free_tag()
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::get_free_tag");
    #endif

    int i=0;
    while(wobject_list[++i]);
    return i;
}


/*************************************************************
 * lower this wobject in the lists(does not do any displaying)
 *************************************************************/
void WinSystem::lower(int tag)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::lower");
    #endif
    int new_pos;

    new_pos = get_first_free_bottom_place();
    int 
        its_position_in_order = wobject_order_list[tag];
    

    // move up the other wobjects
    for (int i=wobject_order_list[tag] ;i<new_pos;i++)
    {
        order_wobject_list[i] = order_wobject_list[i + 1];
        wobject_order_list[order_wobject_list[i]] = i;
    }
    wobject_order_list[tag] = new_pos;
    order_wobject_list[new_pos] = tag;

    
    GeoInfo const *g = wobject_list[tag]->get_geoinfo();
    BITMAP *s = _Buffered ? buffer : this_managers_screen;
    if (g->is_displayed)
    {
        if (_ShadowStyle > 0)
        {
            if (!_Buffered)
               mouse_off
               (
                g->pos_x + g->Right_Border_Size + g->screen_size_x,
                g->pos_y - g->Upper_Border_Size,
                g->pos_x + g->Right_Border_Size + g->screen_size_x + _ShadowSize,
                g->pos_y + g->Lower_Border_Size + g->screen_size_y + _ShadowSize
               );
            // erase shadow without redrawing objects
            rectfill
            (
                s,
                g->pos_x + g->Right_Border_Size + g->screen_size_x,
                g->pos_y - g->Upper_Border_Size,
                g->pos_x + g->Right_Border_Size + g->screen_size_x + _ShadowSize - 1,
                g->pos_y + g->Lower_Border_Size + g->screen_size_y + _ShadowSize - 1,
                _BackgroundColor
            );
            if (!_Buffered)
               mouse_off
               (
                g->pos_x - g->Left_Border_Size,
                g->pos_y + g->Lower_Border_Size + g->screen_size_y,
                g->pos_x + g->Right_Border_Size + g->screen_size_x,
                g->pos_y + g->Lower_Border_Size + g->screen_size_y + _ShadowSize
               );
            
            rectfill
            (
                s,
                g->pos_x - g->Left_Border_Size,
                g->pos_y + g->Lower_Border_Size + g->screen_size_y,
                g->pos_x + g->Right_Border_Size + g->screen_size_x - 1,
                g->pos_y + g->Lower_Border_Size + g->screen_size_y + _ShadowSize - 1,
                _BackgroundColor
            );
            mouse_on();
            mouse_on(); // balance mouse_off/mouse_on pairs
            
        }
        
        if (UseBoxes) 
        {
            update_boxlists(its_position_in_order);
            rebuild_screen_using_boxes
            (
                s,
                g->pos_x - g->Left_Border_Size,
                g->pos_y - g->Upper_Border_Size,
                g->pos_x + g->Right_Border_Size + g->screen_size_x + _ShadowSize,
                g->pos_y + g->Lower_Border_Size + g->screen_size_y + _ShadowSize,
                number_of_wobjects - 1,
                0
            );
        }
        else
        {
            rebuild_screen // rebuild_screen handles mouse on/off
            (
                s,
                g->pos_x - g->Left_Border_Size,
                g->pos_y - g->Upper_Border_Size,
                g->pos_x + g->Right_Border_Size + g->screen_size_x + _ShadowSize,
                g->pos_y + g->Lower_Border_Size + g->screen_size_y + _ShadowSize,
                number_of_wobjects - 1
            );
        }
    }
}


/*************************************************************
 * raise this wobject in the lists(does not do any displaying)
 *************************************************************/
void WinSystem::raise(int tag)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::raise");
    #endif
    int new_pos;

    new_pos = get_number_of_locked_top_wobjects();

    // move down the other wobjects
    for (int i=wobject_order_list[tag] ;i>new_pos;i--)
    {
        order_wobject_list[i] = order_wobject_list[i - 1];
        wobject_order_list[order_wobject_list[i]] = i;
    }
    wobject_order_list[tag] = new_pos;
    order_wobject_list[new_pos] = tag;
}
/*************************************************************
 * raise this wobject in the lists(does not do any displaying)
 * even if it's locked, and always completely to the top, 
 * ignoring any top_locked windows
 *************************************************************/
void WinSystem::raise_guaranteed(int tag)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::raise_guaranteed");
    #endif

    // move down the other wobjects
    for (int i=wobject_order_list[tag] ;i>0;i--)
    {
        order_wobject_list[i] = order_wobject_list[i - 1];
        wobject_order_list[order_wobject_list[i]] = i;
    }
    wobject_order_list[tag] = 0;
    order_wobject_list[0] = tag;
}

/*************************************************************
 * some info functions
 *************************************************************/
Wobject *WinSystem::get_wobject(int tag)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::get_wobject");
    #endif

    if (tag >=  buffersize)
        return NULL;
    return wobject_list[tag];
}


Wobject *WinSystem::get_parent()
{
    return parent;
}



int WinSystem::get_TheScreen_size_x()
{
    return TheScreen_size_x;
}

int WinSystem::get_order(int tag)
{
    if (!wobject_list[tag])
        fatal("call for non-existing wobject");
    return wobject_order_list[tag];
}

int WinSystem::get_TheScreen_size_y()
{
    return TheScreen_size_y;
}

/*************************************************************
 * this function returns the wobject in which these coordinates
 * lie  or NULL if they lie within no wobject
 * relative coordinates in respect to the  origin of the user 
 * area are stored in the output_values.
 * these values are scaled to the original size of the wobject
 * (the size of the memory buffer, not the size on screen)
 *************************************************************/
Wobject *WinSystem::process_coordinates(int x,int y,int *out_x = NULL, int *out_y = NULL, int *flag = NULL)
{

    Wobject 
        *w,
        *w2 = NULL;
    int ox,oy;

    if (flag)
       *flag = 0;

    if (current_coordinate_claimer)
    {
         if (out_x)
                  *out_x = x;
         if (out_y)
                  *out_y = y;
         if (out_x && out_y)
                  current_coordinate_claimer->transform_coordinates(out_x,out_y);
         return current_coordinate_claimer;
    }
    
    for (int i=0;i<number_of_wobjects;i++)
    {
        w = wobject_list[order_wobject_list[i]];
        if 
        (   
            w->geoinfo.pos_x <= x && 
            w->geoinfo.pos_y <= y && 
            w->geoinfo.pos_x + w->geoinfo.screen_size_x > x &&
            w->geoinfo.pos_y + w->geoinfo.screen_size_y > y &&
            w->geoinfo.is_displayed
        )
        {
            ox = x - w->geoinfo.pos_x;
            oy = y - w->geoinfo.pos_y;
            ox = (int)(ox * (float)w->geoinfo.viewport_size_x/(float)w->geoinfo.screen_size_x);
            oy = (int)(oy * (float)w->geoinfo.viewport_size_y/(float)w->geoinfo.screen_size_y);
            ox += w->geoinfo.viewport_pos_x;
            oy += w->geoinfo.viewport_pos_y;
            
            if (!(w->inquire_mouse(ox,oy)))  // ask the window if these mouse coordinates
                continue;                 // (relative to the window's origin) are part of the window
            ox -= w->geoinfo.user_pos_x;
            oy -= w->geoinfo.user_pos_y;
            if (out_x)
                        *out_x =ox;
            if (out_y)
                        *out_y =oy;
            if (w->Child_manager)
            {
                w2 = w->Child_manager->process_coordinates(ox,oy,out_x,out_y,flag);
                if  (w2)
                {
                    w = w2;// return the child wobject
                }
                else
                {
                     if (out_x)
                        *out_x =ox;
                     if (out_y)
                        *out_y =oy;
                }
            }

            return w;
        }
        else if 
        (   
            w->geoinfo.pos_x - w->geoinfo.Left_Border_Size<= x && 
            w->geoinfo.pos_y - w->geoinfo.Upper_Border_Size<= y && 
            w->geoinfo.pos_x + w->geoinfo.screen_size_x + w->geoinfo.Right_Border_Size > x &&
            w->geoinfo.pos_y + w->geoinfo.screen_size_y + w->geoinfo.Lower_Border_Size > y &&
            w->geoinfo.is_displayed
        )
        {
         if (flag)
            *flag = 1;
         if (out_x)
            *out_x = x - (w->geoinfo.pos_x );
         if (out_y)
            *out_y = y - (w->geoinfo.pos_y );
            return w;
        }

    }
    if (out_x)
       *out_x = x;
    if (out_y)
       *out_y = y;
    return NULL;
}

/*************************************************************
 * this function does nothing when in unbuffered mode
 *************************************************************/
void WinSystem::update_screen(int xp,int yp,int xw,int yh)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::update_screen");
    #endif

    if (_Buffered)
    {
        mouse_off(xp,yp,xp+xw,yp+yh);
        if (Vsynced && this_managers_screen == screen)
            vsync();
        //if (_mouse_screen == screen) if (wham_trace) trace("1");
        blit(buffer,this_managers_screen,xp,yp,xp,yp,xw,yh);
      mouse_on();
    }
    if (parent)
    {
//     trace("parent _update");
//      parent->update();
        parent->update(xp,yp,xw+1,yh+1); //wqaarom +1, nog eens uitzoeken
    }
}


/*************************************************************
 * this function does nothing when in unbuffered mode
 * else it updates the screen slowly using effect effect
 *************************************************************/


void WinSystem::update_screen(int xp,int yp,int xw,int yh, int effect)
{
    #ifdef WTRACE
        if (wham_trace) trace("WinSystem::update_screen w effect");
    #endif

    if (_Buffered)
    {
        int xf,xt,yf,yt;
        xf= MAX(0,xp);
        yf= MAX(0,yp);
        xt= MIN(xp + xw, SCREEN_W);
        yt= MIN(yp + yh, SCREEN_H);
        if (xt < xf || yt < yf) return;

        xp = xf;
        yp = yf;
        xw = xt - xf;
        yh = yt - yf;

        int xx,yy;
        mouse_off(xp,yp,xp+xw,yp+yh);
        
        switch(effect)
        {
            case W_EFFECT_HLINES:
               for (int i=0;i<10;i++)
               {
                    for(int j=yp;j<yp + yh;j+=10)
                    {
                        if (Vsynced && this_managers_screen == screen)
                            vsync();

                        blit(buffer,this_managers_screen,xp,j + i,xp,j+i,xw,1);
                    }
                    rest(Show_Effect_Param);
               }
            break;
            case W_EFFECT_VLINES:
               for (int i=0;i<10;i++)
               {
                    for(int j=xp;j<xp + xw;j+=10)
                    {
                        if (Vsynced && this_managers_screen == screen)
                            vsync();

                        blit(buffer,this_managers_screen,j+i,yp,j+i,yp,1,yh);
                    }
                    rest(Show_Effect_Param);
               }
            break;
            case W_EFFECT_DOTS:
               for (int i=0;i<10;i++)
               {
                    for(int j=0;j<10;j++)
                    {
                        if (Vsynced && this_managers_screen == screen)
                            vsync();
                        for (int ii=xp;ii<xp + xw;ii+=10)
                            for (int jj=yp;jj<yp + yh;jj+=10)
                                putpixel(this_managers_screen,ii+i,jj+j,getpixel(buffer,ii+i,jj+j));
                        rest(Show_Effect_Param);
                            
                    }
               }
            break;
            case W_EFFECT_PRERECT:
                for (int i=0;i<MAX(xw,yh);i+=Show_Effect_Param)
                {
                    if (Vsynced && this_managers_screen == screen)
                        vsync();
                    rectfill(this_managers_screen,xp,yp,xp+MIN(i,xw-1),yp + MIN(i,yh-1),0);
                }

            // fall through
            default:
                   if (Vsynced && this_managers_screen == screen)
                            vsync();

                   blit(buffer,this_managers_screen,xp,yp,xp,yp,xw,yh);
            break;

            case W_EFFECT_BUNCH:
                    for (int i = 0;i<xw*yh*Show_Effect_Param;i++)
                    {
                        if (Vsynced && this_managers_screen == screen)
                            vsync();
                        xx = xp + (rand() % xw);
                        yy = yp + (rand() % yh);
                        putpixel(this_managers_screen,xx,yy,getpixel(buffer,xx,yy));
                    }    
                    if (Vsynced && this_managers_screen == screen)
                            vsync();

                    blit(buffer,this_managers_screen,xp,yp,xp,yp,xw,yh);
            break;
        }
        mouse_on();
    }
    if (parent)
        parent->update(xp,yp,xw+1,yh+1); //wqaarom +1, nog eens uitzoeken
}


/*************************************************************
 * draws all objects above from_wobject in the specifield area
 *************************************************************/
void WinSystem::rebuild_screen(BITMAP * s,int from_x,int from_y,int to_x,int to_y, int from_wobject)
{
    Wobject *the_obj;
    GeoInfo g;
    if (!_Buffered)
      mouse_off(from_x,from_y,to_x,to_y);

    //trace("rebuild screen: set clip %p",s);
    set_clip(s,from_x,from_y,to_x-1,to_y-1);
    for (int i= from_wobject;i>=0;i--)
    {
        the_obj = wobject_list[order_wobject_list[i]];
        g = the_obj->geoinfo;
        if (g.is_displayed 
            && !the_obj->frozen && 
            !the_obj->is_being_deleted
            )   // if is_being_deleted of the object is set to true, the destructor is active, which means all functions from derived classes
        {                                                             // are already destroyed
            if (_ShadowStyle)
            {
                the_obj->draw_shadow
                (
                    s,
                    from_x,
                    from_y,
                    to_x,
                    to_y
                );
            }
            
            if (the_obj->draw_border)
            {
                the_obj->draw_border
                (
                    s,
                    wobject_list[order_wobject_list[i]],
                    from_x,
                    from_y,
                    to_x,
                    to_y
                );
            }

            the_obj->draw
            (
                s,
                from_x,
                from_y,
                to_x,
                to_y
            );
        }
    }
    //trace("rebuild screen: set clip %p",s);
    set_clip(s,0,0,TheScreen_size_x - 1,TheScreen_size_y - 1);
    if (!_Buffered)
      mouse_on();
    
}
/*************************************************************
 * draws all objects below from_wobject in the specifield area using boxlists
 *************************************************************/
void WinSystem::rebuild_screen_using_boxes(BITMAP * s,int from_x,int from_y,int to_x,int to_y, int from_wobject, int to_wobject)
{
          
    Wobject *the_obj;
    GeoInfo g;

//    trace("rebuild screen using boxes %d %d %d %d from %d  to %d",
//          from_x,from_y,to_x,to_y,from_wobject,to_wobject);

    from_x = MAX(0,from_x);
    from_y = MAX(0,from_y);
    to_x = MIN(to_x, TheScreen->w);
    to_y = MIN(to_y, TheScreen->h);
    if (from_x >= to_x || from_y >= to_y)
    {
//       trace("rebuild sc %d %d %d %d  NOT",from_x,from_y,to_x,to_y);
       return;
    }



    //trace("rebuild screen ub: set clip %p",s);
    set_clip(s,from_x,from_y,to_x-1,to_y-1);

    for (int i= from_wobject;i>= to_wobject;i--)
    {
        the_obj = wobject_list[order_wobject_list[i]];
        g = the_obj->geoinfo;
        if (g.is_displayed 
            && !the_obj->frozen && 
            !the_obj->is_being_deleted
            )   // if is_being_deleted of the object is set to true, the destructor is active, which means all functions from derived classes
        {                          // are already destroyed
//            trace("draw i");
            the_obj->my_boxlist->draw_the_boxes(s,from_x,from_y,to_x,to_y);
//            trace("i drawn");
        }    
    }
  redraw_higher_lying_see_through_objects(s,from_x,from_y,to_x,to_y,from_wobject);


    //trace("rebuild screen ub: set clip %p",s);
    set_clip(s,0,0,TheScreen_size_x - 1,TheScreen_size_y - 1);
}

/*************************************************************
 * updates all visible object on the whole screen
 *************************************************************/

void WinSystem::refresh_all_visible_objects()
{
    if (!_Buffered)
       mouse_off();
    BITMAP *s = _Buffered ? buffer : this_managers_screen;

    #ifdef DEBUGBOXLIST
//       clear(TheScreen);
    #endif

    if (WinSystem::Full_Remove)
    {
       drawing_mode(DRAW_MODE_SOLID,0,0,0);
    //if (_mouse_screen == screen) if (wham_trace) trace("2");
       rectfill(s,0,0,TheScreen_size_x -1,TheScreen_size_y -1,_BackgroundColor);
    }
    for (int i=0;i<number_of_wobjects;i++)
        wobject_list[order_wobject_list[i]]->refresh_border_info();
        
    
    if (!UseBoxes)
       rebuild_screen(s,0,0,TheScreen_size_x ,TheScreen_size_y ,number_of_wobjects -  1);
    else
    {
//       update_boxlists(0);
       rebuild_screen_using_boxes(s,0,0,TheScreen_size_x ,TheScreen_size_y ,number_of_wobjects -1, 0);
    }
    if (!_Buffered)
        mouse_on();

    update_screen(0,0,TheScreen_size_x ,TheScreen_size_y);

}

/*************************************************************
 * returns pointer to bitmap used for output
 *************************************************************/


BITMAP *WinSystem::get_TheScreen()
{
    return _Buffered ? buffer : this_managers_screen;
}

/*************************************************************
 * sends a message to 'to', returns the value this object returns.
 * if 'to' is NULL  
 * it sends it to every object from the top down
 * until one accepts it, returns the value that one returns
 * and sets 'accpeted' to the object that responded(if accepted is not NULL)
 *************************************************************/


int WinSystem::send_message(Wobject *from, int message, Wobject *to, Wobject **accepted)
{
    #ifdef DEBUGMESSAGESYS
        if (from && to)
            if (wham_trace) trace("send_message: message %s from %s to %s",message_string(message),from->get_geoinfo()->name,to->get_geoinfo()->name);
        else if (from)
            if (wham_trace) trace("send_message: message %s from %s to first who wants it",message_string(message),from->get_geoinfo()->name);
        else if (to)
            if (wham_trace) trace("send_message: message %s from unknown to %s",message_string(message),to->get_geoinfo()->name);
        else
            if (wham_trace) trace("send_message: message %s from unknown to first who wants it",message_string(message));
    #endif

    Wham_Secretary.process_message(from,message,to);

    if (to)
        return to->process_message(from,message);

    int ret = FALSE;
    Wobject *obj;

    for (int i=0;i<number_of_wobjects;i++)
    {
        obj = wobject_list[order_wobject_list[i]];
        // try all but sender
        if ((obj != from) && (ret = obj->process_message(from,message)))
        {
            if (accepted)
                *accepted = wobject_list[order_wobject_list[i]];
            return ret;
        }
        if (obj->Child_manager)
            obj->Child_manager->send_message(from,message,to);

    }
    // if no object wants it,  and there is a user hook, send it there
    if (user_process_message)
        return user_process_message(from,message);

    return 0;
}

/*************************************************************
 * sends the message to all objects, ignoring their return values
 * but returning TRUE if one of them returned something nonzero
 *************************************************************/

int WinSystem::broadcast_message(Wobject *from, int message)
{
    #ifdef DEBUGMESSAGESYS
        if (from)
            if (wham_trace) trace("broadcast_message: message %s from %s to all",message_string(message),from->get_geoinfo()->name);
        else
            if (wham_trace) trace("broadcast_message: message %s from unknown to all",message_string(message));
    #endif


    Wham_Secretary.process_message(from,message,NULL);

    if (!Message_System_Active)
        return 0;
    int is_response = FALSE;
    Wobject *obj;
    for (int i=0;i<number_of_wobjects;i++)
    {
        obj = wobject_list[order_wobject_list[i]];
        // send to all but the sender
        if (obj !=from && (obj->process_message(from,message)))
        {
            is_response = TRUE;
        }
        if (obj->Child_manager)
            obj->Child_manager->broadcast_message(from,message);
    }
    if (user_process_message && user_process_message(from,message))
            is_response = TRUE;
    if (message == W_MSG_MANAGER_DELETE)
        Message_System_Active = FALSE;
        
    return is_response;
}


void WinSystem::broadcast_delayed_message(Wobject *from, int message)
{
    if (!Message_System_Active)
        return;
        
    nr_of_messages_in_mailbin++;
    
//    if (from)
//       trace("broadcast delayed: %s : %s (%d msgs)",from->get_geoinfo()->name,message_string(message),nr_of_messages_in_mailbin);
//    else
//       trace("broadcast delayed: (?) : %s(%d msgs)",message_string(message),nr_of_messages_in_mailbin);
       
    if (nr_of_messages_in_mailbin > mailbin_size)
    {
      mailbin_size += Mailbin_Increment;
      mailbin = (Message *)realloc(mailbin, sizeof(Message)* mailbin_size);
      if (!mailbin)
         fatal("Could not enlarge Message collection buffer");
    }
    mailbin[nr_of_messages_in_mailbin-1].message = message;
    mailbin[nr_of_messages_in_mailbin-1].from = from;
    mailbin[nr_of_messages_in_mailbin-1].to = NULL;
    mailbin[nr_of_messages_in_mailbin-1].to_all = TRUE;

}

void WinSystem::send_delayed_message(Wobject *from, int message, Wobject *to)
{
    if (!Message_System_Active)
        return;
    nr_of_messages_in_mailbin++;

//    if (from)
//       trace("send delayed message %s to %s: %s (msgs %d)",
 //            from->get_geoinfo()->name,
//             to->get_geoinfo()->name,
 //            message_string(message),
//             nr_of_messages_in_mailbin
//             );
//    else
//       trace("send delayed message (?) to %s: %s (msgs %d)",
//             to->get_geoinfo()->name,
//             message_string(message),
//             nr_of_messages_in_mailbin
//             );
    
    if (nr_of_messages_in_mailbin > mailbin_size)
    {
      mailbin_size += Mailbin_Increment;
      mailbin = (Message *)realloc(mailbin, sizeof(Message)* mailbin_size);
      if (!mailbin)
         fatal("Could not enlarge Message collection buffer");
    }
    mailbin[nr_of_messages_in_mailbin-1].message = message;
    mailbin[nr_of_messages_in_mailbin-1].from = from;
    mailbin[nr_of_messages_in_mailbin-1].to = to;
    mailbin[nr_of_messages_in_mailbin-1].to_all =FALSE;
}


void WinSystem::dispatch_delayed_messages()
{
   Wobject *obj;
   if (!Message_System_Active)
     return;

   int until = nr_of_messages_in_mailbin; // don't process messages that come in during execution of this function
//   trace("dispatch delayed (%d msgs)",until);
   for (int i=0;i<until;i++)
   {

      Wham_Secretary.process_message((Wobject *)(mailbin[i].from),mailbin[i].message,(Wobject *)(mailbin[i].to));

      if (mailbin[i].to)
      {
//         if (mailbin[i].from)
//            trace("msg %d %s from %s to %s",i, message_string(mailbin[i].message),
//            ((Wobject *)(mailbin[i].from))->get_geoinfo()->name,
//            ((Wobject *)(mailbin[i].to))->get_geoinfo()->name
//            );
//         else
//            trace("msg %d %s from (?) to %s",i, message_string(mailbin[i].message),
//            ((Wobject *)(mailbin[i].to))->get_geoinfo()->name
//            );
         ((Wobject *)(mailbin[i].to))->process_message((Wobject *)(mailbin[i].from),mailbin[i].message);
      }
      else
      {
//         if (mailbin[i].from)
//            trace("msg %d %s from %s to all",i, message_string(mailbin[i].message),
//            ((Wobject *)(mailbin[i].from))->get_geoinfo()->name
//            );
//         else
//            trace("msg %d %s from (?) to all",i, message_string(mailbin[i].message));
            
       for (int j=0;j<number_of_wobjects;j++)
       {
           obj = wobject_list[order_wobject_list[j]];
           // send to all, even the sender
            obj->process_message((Wobject *)(mailbin[i].from),mailbin[i].message);
            
           if (obj->Child_manager)
           {
               obj->Child_manager->broadcast_delayed_message((Wobject *)(mailbin[i].from),mailbin[i].message);
           }
       }
       if (user_process_message)
         user_process_message((Wobject *)(mailbin[i].from),mailbin[i].message);
      }

   }

   for (int i=0;i<number_of_wobjects;i++)
   {
     obj = wobject_list[order_wobject_list[i]];
     if (obj->Child_manager)
         obj->Child_manager->dispatch_delayed_messages();
   }

   for (int i=until;i<nr_of_messages_in_mailbin;i++)
   {
      mailbin[i - until] = mailbin[i];
   }

   nr_of_messages_in_mailbin-=until; // some more messages could have arrived in the mean time

}

int (*WinSystem::user_process_message)(Wobject *from, int message) = NULL;

//mouse pointers


void WinSystem::set_user_msg_processor(int (*user_msg_proc)(Wobject *from,int message))
{
     user_process_message = user_msg_proc;
}

Mouse_Cursor
    *WinSystem::Mouse_Cursor0 = NULL,
    *WinSystem::Mouse_Cursor1 = NULL,
    *WinSystem::Mouse_Cursor2 = NULL,
    *WinSystem::Mouse_Cursor3 = NULL,
    *WinSystem::Mouse_Cursor4 = NULL,
    *WinSystem::Mouse_Cursor5 = NULL,
    *WinSystem::Mouse_Cursor6 = NULL,
    *WinSystem::Mouse_Cursor7 = NULL,
    *WinSystem::Mouse_Cursor8 = NULL,
    *WinSystem::Mouse_Cursor9 = NULL;
                            
int WinSystem::get_number_of_wobjects()
{
    return number_of_wobjects;
}


Wobject *WinSystem::get_top_wobject()
{
    int number_of_top_locked = get_number_of_locked_top_wobjects();
    if (number_of_top_locked >= number_of_wobjects)
        return NULL;
    //return the first non top-locked window
    return wobject_list[order_wobject_list[number_of_top_locked]];
}



void WinSystem::set_shift_all_range(int min_x,int min_y,int max_x, int max_y) // pass all 0 to set to infinite
{
   minshiftx = min_x;
   minshifty = min_y;
   maxshiftx = max_x;
   maxshifty = max_y;
}


/****************************************************************************
 * shifts all windows, except those that are shift_locked
 ***************************************************************************/
void WinSystem::shift_all(int x, int y)
{
    int lastxshifted, lastyshifted;
    BITMAP *s = _Buffered ? buffer : this_managers_screen;
    Wobject *obj;
    if (!y && !x)
    {
        x = - x_shifted;
        y = - y_shifted;
    }
    lastyshifted = y_shifted;
    lastxshifted = x_shifted;
    y_shifted += y;
    x_shifted += x;

    if (minshiftx<=maxshiftx)
      x_shifted = MID(x_shifted,maxshiftx,minshiftx);

    if (minshifty<=maxshifty)
      y_shifted = MID(y_shifted,maxshifty,minshifty);

    x = x_shifted - lastxshifted;
    y = y_shifted - lastyshifted;

    for (int i=0;i<number_of_wobjects;i++)
    {
        obj = wobject_list[order_wobject_list[i]];
        if (!obj->geoinfo.shift_is_locked)
        {
            obj->geoinfo.pos_x += x;
            obj->geoinfo.pos_y += y;
        }
    }
    if (UseBoxes)
        update_boxlists(0);
        
    if (Full_Remove)
       clear_to_color(s,_BackgroundColor);

    
    if (!UseBoxes)
      rebuild_screen(s,0,0,TheScreen_size_x , TheScreen_size_y, number_of_wobjects -1);
    else
      rebuild_screen_using_boxes(s,0,0,TheScreen_size_x , TheScreen_size_y , number_of_wobjects -1,0);
    
    update_screen(0,0,TheScreen_size_x ,TheScreen_size_y );
    broadcast_message(NULL,W_MSG_SHIFTED);
}


Wobject *WinSystem::get_wobject_from_hierarchy(int nr)
{
    if (nr >= 0 && nr < number_of_wobjects)
        return wobject_list[order_wobject_list[nr]];

    return NULL;
}

void WinSystem::update_boxlists(Wobject *from)
{
    update_boxlists(wobject_order_list[from->geoinfo.tag]);
}


/************************************************************
 * updates all boxlists in windows lying under hierachr nr from
 * 
 ****************************************************************/


void WinSystem::update_boxlists(int from)
{

   #ifdef DEBUGBOXLIST
   static BITMAP *tmpbuffer = create_bitmap(SCREEN_W,SCREEN_H);
   if (!tmpbuffer)
      out_of_mem("cannot debug");
   #endif
    Wobject *this_object,*that_object;
    for (int i= from; i< number_of_wobjects;i++)
    {
    #ifdef DEBUGBOXLIST_LISTPART
        if (wham_trace) trace("going to update %d",i);
    #endif
        this_object = wobject_list[order_wobject_list[i]];

        if (!this_object->geoinfo.is_displayed)
            continue;

        this_object->re_init_boxlist();
        for (int j=0;j<i;j++)
        {
            that_object = wobject_list[order_wobject_list[j]];
            if (that_object->geoinfo.is_displayed && !that_object->see_through)
            {

               #ifdef DEBUGBOXLIST
                blit(screen,tmpbuffer,0,0,0,0,SCREEN_W,SCREEN_H);
                arced_rect
                (
                  TheScreen,
                  this_object->geoinfo.pos_x - this_object->geoinfo.Left_Border_Size,
                  this_object->geoinfo.pos_y - this_object->geoinfo.Upper_Border_Size,
                  this_object->geoinfo.pos_x + this_object->geoinfo.Right_Border_Size + this_object->geoinfo.screen_size_x -1,
                  this_object->geoinfo.pos_y + this_object->geoinfo.Lower_Border_Size + this_object->geoinfo.screen_size_y -1,
                  makecol(0,255,0)
                );
                arced_rect
                (
                  TheScreen,
                  that_object->geoinfo.pos_x - that_object->geoinfo.Left_Border_Size,
                  that_object->geoinfo.pos_y - that_object->geoinfo.Upper_Border_Size,
                  that_object->geoinfo.pos_x + that_object->geoinfo.Right_Border_Size + that_object->geoinfo.screen_size_x -1,
                  that_object->geoinfo.pos_y + that_object->geoinfo.Lower_Border_Size + that_object->geoinfo.screen_size_y -1,
                  makecol(255,0,0)
                );
                if (wham_trace) trace("overlay object %d over %d",order_wobject_list[j],order_wobject_list[i]);
                blit(tmpbuffer,screen,0,0,0,0,SCREEN_W,SCREEN_H);
               #endif

                this_object->my_boxlist->overlay
                (
                    that_object->geoinfo.pos_x - that_object->geoinfo.Left_Border_Size,
                    that_object->geoinfo.pos_y - that_object->geoinfo.Upper_Border_Size,
                    that_object->geoinfo.pos_x + that_object->geoinfo.Right_Border_Size + that_object->geoinfo.screen_size_x,
                    that_object->geoinfo.pos_y + that_object->geoinfo.Lower_Border_Size + that_object->geoinfo.screen_size_y
                );
            }
        }

    }
}


/************************************************************
 * for debugging
 * 
 ****************************************************************/
void WinSystem::show_all_boxlist_boxes()
{
    if (!UseBoxes)
        return;

    PALETTE p;
    get_palette(p);
    set_palette(desktop_palette);
    mouse_off();
    for (int i=0;i<number_of_wobjects;i++)
    {
        if (wobject_list[i]->geoinfo.is_displayed)
        {
            clear_to_color(screen,makecol(0,0,0));
            wobject_list[i]->my_boxlist->show_the_boxes();
            trace("wobject %d",i);
        }
    }
    set_palette(p);
    mouse_on();
}


/************************************************************
 * not very useful, speedgain nearly zero
 * could be useful once for something, so I'll let it here
 ****************************************************************/
void WinSystem::defrag_all_boxlist_boxes()
{    
    if (!UseBoxes)
        return;
    for (int i=0;i<number_of_wobjects;i++)
    {
        if (wobject_list[i]->geoinfo.is_displayed)
        {
            wobject_list[i]->my_boxlist->defrag_the_boxes();
        }
    }
}


/************************************************************
 * because the see_through objects are not overlaid (with boxlists)
 * to allow the underlying windows to be visible
 * the will be draw over by underlying objects
 * hence this function to restore them
 ****************************************************************/
void WinSystem::redraw_higher_lying_see_through_objects(BITMAP *s, int from_x, int from_y, int to_x, int to_y, Wobject* from_wobject)
{
    redraw_higher_lying_see_through_objects(s,from_x,from_y,to_x,to_y,wobject_order_list[from_wobject->geoinfo.tag]);
}

void WinSystem::redraw_higher_lying_see_through_objects(BITMAP *s, int from_x, int from_y, int to_x, int to_y, int from_wobject)
{
    // update higher lying see_through objects
    Wobject *the_obj;
    GeoInfo g;
    from_x = MAX(from_x,0);
    from_y = MAX(from_y,0);
    to_x = MIN(to_x, TheScreen->w);
    to_y = MIN(to_y, TheScreen->w);

    if (from_x >= to_x || from_y >= to_y)
    {
//       trace("redraw higher st %d %d %d %d  NOT",from_x,from_y,to_x,to_y);
       return;
    }

    for (int i= from_wobject - 1;i>= 0;i--)
    {
        the_obj = wobject_list[order_wobject_list[i]];
        g = the_obj->geoinfo;
        if (g.is_displayed 
            && !the_obj->frozen 
            && !the_obj->is_being_deleted
            && the_obj->see_through
            )   // if is_being_deleted of the object is set to true, the destructor is active, which means all functions from derived classes
        {                          // are already destroyed
            the_obj->my_boxlist->draw_the_boxes(s,from_x,from_y,to_x,to_y);
        }    
    }
}
// because shadows are not overlaid, shadows mean all the higer lying boxes 
// need redrawing
// thus also this function for opaque objects

void WinSystem::redraw_higher_lying_objects(BITMAP *s, int from_x, int from_y, int to_x, int to_y, Wobject* from_wobject)
{
    redraw_higher_lying_objects(s,from_x,from_y,to_x,to_y,wobject_order_list[from_wobject->geoinfo.tag]);
}

void WinSystem::redraw_higher_lying_objects(BITMAP *s, int from_x, int from_y, int to_x, int to_y, int from_wobject)
{
    // update higher lying see_through objects
    Wobject *the_obj;
    GeoInfo g;
    from_x = MAX(from_x,0);
    from_y = MAX(from_y,0);
    to_x = MIN(to_x, TheScreen->w);
    to_y = MIN(to_y, TheScreen->w);

    if (from_x >= to_x || from_y >= to_y)
    {
//       trace("redraw highr lying %d %d %d %d  NOT",from_x,from_y,to_x,to_y);
       return;
    }

    for (int i= from_wobject - 1;i>= 0;i--)
    {
        the_obj = wobject_list[order_wobject_list[i]];
        g = the_obj->geoinfo;
        if (g.is_displayed 
            && !the_obj->frozen 
            && !the_obj->is_being_deleted
            )   // if is_being_deleted of the object is set to true, the destructor is active, which means all functions from derived classes
        {                          // are already destroyed
            the_obj->my_boxlist->draw_the_boxes(s,from_x,from_y,to_x,to_y);
        }    
    }
}


BITMAP *WinSystem::get_scratch_bmp(int size_x, int size_y, int color_depth)
{
        if (!color_depth)
                color_depth = bitmap_color_depth(screen);

        if (!scratch_bmp)
        {
                scratch_bmp = create_bitmap_ex(color_depth,size_x,size_y);
                if (!scratch_bmp)
                        out_of_mem("allocating scratch buffer");
                return scratch_bmp;
        }
        if (size_x > scratch_bmp->w || size_y > scratch_bmp->h || color_depth != bitmap_color_depth(scratch_bmp))
        {
                destroy_bitmap(scratch_bmp);
                scratch_bmp = create_bitmap_ex(color_depth,size_x,size_y);
                if (!scratch_bmp)
                        out_of_mem("allocating scratch buffer");
                return scratch_bmp;
        }
        //trace("scrtch %p",scratch_bmp);
        set_clip(scratch_bmp, 0, 0, size_x - 1, size_y - 1);
        return scratch_bmp;

}

int WinSystem::must_delete_colormap = FALSE;



void WinSystem::claim_coordinates(Wobject *w)
{
#ifdef WTRACE
   if (wham_trace) trace("%s claims coordinates",w->get_geoinfo()->name);
#endif
   
   coordinate_claimer_stack->push((void *)current_coordinate_claimer);
   current_coordinate_claimer = w;
}

void WinSystem::claim_keyboard(Wobject *w)
{
#ifdef WTRACE
   if (wham_trace) trace("%s claims keyboard",w->get_geoinfo()->name);
#endif
   keyboard_claimer_stack->push((void *)current_keyboard_claimer);
   current_keyboard_claimer = w;
}

void WinSystem::release_coordinates(Wobject *w)
{
#ifdef WTRACE
   if (wham_trace) trace("release coordinates");
#endif
   if (current_coordinate_claimer == w)
      current_coordinate_claimer = (Wobject *)coordinate_claimer_stack->pop();
   else
       coordinate_claimer_stack->remove(w);
}

void WinSystem::release_keyboard(Wobject *w)
{
#ifdef WTRACE
   if (wham_trace) trace("release keyboard");
#endif
   if (current_keyboard_claimer == w)
      current_keyboard_claimer = (Wobject *)keyboard_claimer_stack->pop();
   else
      keyboard_claimer_stack->remove(w);
}


Wobject *WinSystem::get_keyboard_claimer()
{
   return current_keyboard_claimer;
}

Wobject *WinSystem::get_coordinate_claimer()
{
   return current_coordinate_claimer;
}

void WinSystem::output_debug_info()
{
   PACKFILE * debtxt = pack_fopen("debug.txt","w");
   char buffer[100];
   if (!debtxt)
      fatal("could not open 'debug.txt'");

   pack_fputs("Debug_Info\n",debtxt);
   sprintf(buffer, "Number of objects:%d\n",number_of_wobjects);
   

   pack_fclose(debtxt);
}


void WinSystem::do_requests()
{
  Wobject *the_obj;
  #ifdef WTRACE
      int check = FALSE; // check for requests
      for (int i=0;i<number_of_wobjects;i++)
      {
        the_obj = wobject_list[order_wobject_list[i]];
        if (the_obj->the_request)
            check = TRUE;
      }
  
      if (wham_trace && check) trace("WinSystem::do_requests");
  #endif



   for (int i=0;i<number_of_wobjects;i++)
   {
        the_obj = wobject_list[order_wobject_list[i]];
        if (the_obj->Child_manager)
          the_obj->Child_manager->do_requests();
        
        switch(the_obj->the_request)
        {
         default:
            break;
         case W_REQUEST_UPDATE:
            the_obj->update();
            break;
         case W_REQUEST_SHOW:
            the_obj->show();
            break;
         case W_REQUEST_HIDE:
            the_obj->hide();
            break;
        }
        if (the_obj && the_obj->the_request != W_REQUEST_DELETE)
           the_obj->the_request = 0;
   }

   int done = 0;
   while (!done)
   {
      done = -1;
      for (int i=0;i<number_of_wobjects;i++)
      {
         the_obj = wobject_list[order_wobject_list[i]];
         if (the_obj->the_request == W_REQUEST_DELETE)
         {
            the_obj->the_request = 0;
            delete the_obj;
            the_obj = NULL;
            done = 0;
            break;  // break out of for loop & start again
         }
      }
   }
   
}



void WinSystem::set_background_color(int color)
{

   _BackgroundColor = BackgroundColor = color;
   refresh_all_visible_objects();
}



