/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.                                                     *
 *  for contact info see contact.txt in the docs directory.                 *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include "../include/helpers.h"
#include "../include/wobject.h"
#include <string.h>

/*************************************************************
 * constructors of a Wobject
 *************************************************************/
extern BITMAP *_mouse_screen;

Wobject::Wobject(char *name,int _size_x,int _size_y, int _pos_x, int _pos_y)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s Wobject constructor 1",name);
    #endif

    if (!WHAM)
    {
        WHAM = new WinSystem();
        if (!WHAM)
            out_of_mem("cannot create window manager");
        delete_its_manager = TRUE;
    }
    its_manager = WHAM;
    draw_border = NULL;
    create(name,_size_x,_size_y, _pos_x, _pos_y);
}



Wobject::Wobject(char *name,int _size_x,int _size_y, int _pos_x, int _pos_y, WinSystem *_its_manager)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s Wobject constructor 2",name);
    #endif
    
    its_manager = _its_manager;
    draw_border = NULL;
    create(name,_size_x,_size_y, _pos_x, _pos_y);
}


Wobject::Wobject(char *name,int _size_x,int _size_y, int _pos_x, int _pos_y,int (*_draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y))
{
    #ifdef WTRACE
        if (wham_trace) trace("%s Wobject constructor 3",name);
    #endif

    if (!WHAM)
    {
        WHAM = new WinSystem();
        if (!WHAM)
            out_of_mem("cannot create window manager");
        delete_its_manager = TRUE;
    }
    its_manager = WHAM;
    draw_border = _draw_border;
    create(name,_size_x, _size_y, _pos_x, _pos_y);
}



Wobject::Wobject(char *name,int _size_x,int _size_y, int _pos_x, int _pos_y, WinSystem *_its_manager,int (*_draw_border)(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y))
{
    #ifdef WTRACE
        if (wham_trace) trace("%s Wobject constructor 4",name);
    #endif

    its_manager = _its_manager;
    draw_border = _draw_border;
    create(name,_size_x,_size_y, _pos_x, _pos_y);
}


/*************************************************************
 * destructor
 *************************************************************/

Wobject::~Wobject()
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject destructor",geoinfo.name,geoinfo.tag);
    #endif

    if (destruct_callback)
    {
//      trace("call callback");
       destruct_callback(this);
       
    }
    is_being_deleted = TRUE;
    destroy();
}


/*************************************************************
 * some helper functions for constructor/destructor
 *************************************************************/
void Wobject::create(char *_name,int _size_x, int _size_y, int _pos_x, int _pos_y )
{
    #ifdef WTRACE
        if (wham_trace) trace("Wobject::create '%s'",_name);
    #endif
    _size_x = MAX(1,_size_x);
    _size_y = MAX(1,_size_y);
    // get the dimensions of some borders
    geoinfo.name = strdup(_name);
    if (!geoinfo.name)
        out_of_mem("cannot allocate name");
    is_being_deleted = FALSE;
    refresh_border_info();
    see_through = FALSE;
    minimum_size_x = MIN(its_manager->Minimum_Size_x,_size_x);
    minimum_size_y = MIN(its_manager->Minimum_Size_y,_size_y);
    Child_manager = NULL;    
    geoinfo.is_locked = FALSE;
    geoinfo.pos_is_locked = FALSE;
    geoinfo.scale_is_locked = FALSE;
    geoinfo.viewport_pos_is_locked = FALSE;
    geoinfo.viewport_size_is_locked = FALSE;
    geoinfo.size_is_locked = TRUE;
    geoinfo.visibility_is_locked = FALSE;
    geoinfo.shift_is_locked = FALSE;
    count++;
    geoinfo.tag = its_manager->get_new_tag(this);
    geoinfo.pos_x = _pos_x;
    geoinfo.pos_y = _pos_y;
    geoinfo.viewport_size_x = geoinfo.user_size_x = geoinfo.screen_size_x = geoinfo.size_x = _size_x;
    geoinfo.viewport_size_y = geoinfo.user_size_y = geoinfo.screen_size_y = geoinfo.size_y = _size_y;
    geoinfo.user_pos_x =  geoinfo.user_pos_y = geoinfo.viewport_pos_x = geoinfo.viewport_pos_y = 0;
    geoinfo.scale_x = geoinfo.scale_y = 1;

    geoinfo.is_displayed = FALSE;

    user_data = NULL;
    magic_number = 0;
    frozen = 0;
    its_manager->broadcast_message(this, W_MSG_CREATED);
    attached = NULL;


    my_boxlist = NULL;

    #ifdef WTRACE
        if (wham_trace) trace("Wobject::created '%s' tag %d",_name,geoinfo.tag);
    #endif

    the_request = 0;
    
    belongings = NULL;


    destruct_callback = NULL;
}





void Wobject::destroy()
{
    #ifdef WTRACE
        if (wham_trace) trace("Wobject::destroy %d %s",geoinfo.tag,geoinfo.name);
    #endif

    if (belongings)
    {
      Wobject *w;
      while(belongings->get_size())
      {
         w = (Wobject *)(belongings->pop());
         if (!its_manager->is_being_deleted)
                  delete w;
      }
      delete belongings;
      belongings = NULL;
    }

    its_manager->broadcast_message(this, W_MSG_DESTROYED);

    if (Child_manager)
    {
        delete Child_manager;
        
    }
    if (geoinfo.is_displayed)
    {
        its_manager->remove_wobject(geoinfo.tag);  // remove the wobject from the screen
        if (WinSystem::Show_Effect)
        {
            its_manager->update_screen
            (
                geoinfo.pos_x - geoinfo.Left_Border_Size,
                geoinfo.pos_y - geoinfo.Upper_Border_Size,
                geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize + geoinfo.Left_Border_Size ,
                geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize + geoinfo.Upper_Border_Size,
                WinSystem::Show_Effect
            );
        }
        else
        {
            its_manager->update_screen
            (
                geoinfo.pos_x - geoinfo.Left_Border_Size,
                geoinfo.pos_y - geoinfo.Upper_Border_Size,
                geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize + geoinfo.Left_Border_Size ,
                geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize + geoinfo.Upper_Border_Size 
            );
        }
    }
    its_manager->destroy_tag(geoinfo.tag);    // destroy it from the list
    if (!count && delete_its_manager && !its_manager->is_being_deleted)
    {
        if (its_manager == WHAM)
            WHAM = NULL;
        delete its_manager;           // after the last wobject, close off the system
        its_manager = NULL;
    }
//    trace("destroy boxlist");
    if (my_boxlist)
    {
        while (my_boxlist->next)
        {
            delete my_boxlist->next;
        }
        delete my_boxlist;
    }
//    trace("done destroying");
    count--;

}




/*************************************************************
 *  counter of the number of wobjects
 *************************************************************/
int Wobject::count = -1;

/*************************************************************
 * pointer to the wobject manager
 *************************************************************/
WinSystem *WHAM= NULL;

/*************************************************************
 * update the wobject ( show if it is already visible)
 *************************************************************/
void Wobject::update()
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::update",geoinfo.name,geoinfo.tag);
    #endif
    if (geoinfo.is_displayed)
        show();
}

/*************************************************************
 * partial update the wobject ( update part of the object 
 * if it is already visible)
 *************************************************************/
void Wobject::update(int from_x,int from_y,int width,int height)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::update(%d,%d,%d,%d)",geoinfo.name,geoinfo.tag, from_x,from_y,width,height);
    #endif

    if (!geoinfo.is_displayed)
        return;
    int to_x = from_x + width - 1;
    int to_y = from_y + height - 1;
    
    from_x -= geoinfo.viewport_pos_x;
    from_y -= geoinfo.viewport_pos_y;
    to_x -= geoinfo.viewport_pos_x;
    to_y -= geoinfo.viewport_pos_y;
    if (to_x < 0 || to_y <0) 
 	   return;

    if (geoinfo.screen_size_x != geoinfo.viewport_size_x)
        to_x++; // don't know why, but this is neccecary, probably some rounding error
    if (geoinfo.screen_size_y != geoinfo.viewport_size_y)
        to_y++; // don't know why, but this is neccecary, probably some rounding error same needs to be done for from_* but after scaling ?? 8-O))) ???

    to_x = MIN(geoinfo.viewport_size_x, to_x);
    to_y = MIN(geoinfo.viewport_size_x, to_y);
    from_x = MAX(0, from_x);
    from_y = MAX(0, from_y);
    if (geoinfo.screen_size_x != geoinfo.viewport_size_x) // skip floating point multiply if possible
    {
	    from_x = (int)(from_x * geoinfo.scale_x);
	    to_x = (int)((to_x ) * geoinfo.scale_x );
       from_x++; // don't know why, but this is neccecary, probably some rounding error
    }
    if (geoinfo.screen_size_y != geoinfo.viewport_size_y)
    {
	    from_y = (int)(from_y * geoinfo.scale_y);
   	 to_y = (int)((to_y ) * geoinfo.scale_y );
       from_y++; // don't know why, but this is neccecary, probably some rounding error same needs to be done for from_* but after scaling ?? 8-O))) ???
    }
    from_x += geoinfo.pos_x;
    from_y += geoinfo.pos_y;
    to_x += geoinfo.pos_x;
    to_y += geoinfo.pos_y;


    its_manager->display_wobject(geoinfo.tag,from_x,from_y,to_x,to_y);

    its_manager->broadcast_message(this, W_MSG_UPDATED);
    return;

}



int Wobject::refresh_border()
{
    #ifdef WTRACE
        if (wham_trace) trace("Wobject::refresh_border %s (%d)",geoinfo.name,geoinfo.tag);
    #endif
    if (!geoinfo.is_displayed || !draw_border)
       return FALSE;

// trace("going to display %d %d %d %d",
//       geoinfo.pos_x - geoinfo.Left_Border_Size,
//       geoinfo.pos_y - geoinfo.Upper_Border_Size,
//       geoinfo.pos_x,
//       geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size
//      );

   its_manager->display_wobject
   (
       geoinfo.tag,
       geoinfo.pos_x - geoinfo.Left_Border_Size,
       geoinfo.pos_y - geoinfo.Upper_Border_Size,
       geoinfo.pos_x,
       geoinfo.pos_y + geoinfo.screen_size_y
   );

   its_manager->display_wobject
   (
       geoinfo.tag,
       geoinfo.pos_x,
       geoinfo.pos_y - geoinfo.Upper_Border_Size,
       geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size,
       geoinfo.pos_y
   );

   its_manager->display_wobject
   (
       geoinfo.tag,
       geoinfo.pos_x + geoinfo.screen_size_x ,
       geoinfo.pos_y,
       geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size,
       geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size
   );
   its_manager->display_wobject
   (
       geoinfo.tag,
       geoinfo.pos_x - geoinfo.Left_Border_Size,
       geoinfo.pos_y + geoinfo.screen_size_y,
       geoinfo.pos_x + geoinfo.screen_size_x,
       geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size
   );

    
   its_manager->update_screen
   (
       geoinfo.pos_x - geoinfo.Left_Border_Size,
       geoinfo.pos_y - geoinfo.Upper_Border_Size,
       geoinfo.screen_size_x + geoinfo.Right_Border_Size + geoinfo.Left_Border_Size + its_manager->_ShadowSize,
       geoinfo.screen_size_y + geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size + its_manager->_ShadowSize
   );

   return TRUE;
}


/*************************************************************
 * show/update the wobject
 *************************************************************/
int Wobject::show()
{

    #ifdef WTRACE
        if (wham_trace) trace("Wobject::show %s (%d)",geoinfo.name,geoinfo.tag);
    #endif
    int was_displayed = geoinfo.is_displayed;
    
    if (!was_displayed)
    {
        if (geoinfo.visibility_is_locked) // it was not visible and it will not be
            return FALSE;
        its_manager->display_wobject(geoinfo.tag,DISPLAY_MODE_FULL);
        geoinfo.is_displayed = TRUE;
    }
    else
        its_manager->display_wobject(geoinfo.tag,DISPLAY_MODE_NOBORDER);
    
    
    if (!WinSystem::Show_Effect || was_displayed)
    {
        its_manager->update_screen
        (
            geoinfo.pos_x - geoinfo.Left_Border_Size,
            geoinfo.pos_y - geoinfo.Upper_Border_Size,
            geoinfo.screen_size_x + geoinfo.Right_Border_Size + geoinfo.Left_Border_Size + its_manager->_ShadowSize,
            geoinfo.screen_size_y + geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size + its_manager->_ShadowSize
        );
    }
    else
    {
        its_manager->update_screen
        (
            geoinfo.pos_x - geoinfo.Left_Border_Size,
            geoinfo.pos_y - geoinfo.Upper_Border_Size,
            geoinfo.screen_size_x + geoinfo.Right_Border_Size + geoinfo.Left_Border_Size + its_manager->_ShadowSize,
            geoinfo.screen_size_y + geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size + its_manager->_ShadowSize,
            WinSystem::Show_Effect
        );
    }

    if (!was_displayed)
        its_manager->broadcast_delayed_message(this, W_MSG_SHOWN);
    else
        its_manager->broadcast_delayed_message(this, W_MSG_UPDATED);
    return TRUE;
}

/*************************************************************
 * hide the wobject
 *************************************************************/
int Wobject::hide()
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::hide",geoinfo.name,geoinfo.tag);
    #endif
    if (geoinfo.visibility_is_locked && geoinfo.is_displayed)
        return FALSE; //return failure

    if (!geoinfo.is_displayed)
        return TRUE; //return succes

    its_manager->remove_wobject(geoinfo.tag);
    if (!WinSystem::Show_Effect)
    {
        its_manager->update_screen
        (
            geoinfo.pos_x - geoinfo.Left_Border_Size,
            geoinfo.pos_y - geoinfo.Upper_Border_Size,
            geoinfo.screen_size_x + geoinfo.Right_Border_Size + geoinfo.Left_Border_Size + its_manager->_ShadowSize,
            geoinfo.screen_size_y + geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size + its_manager->_ShadowSize
        );
    }
    else
    {
        its_manager->update_screen
        (
            geoinfo.pos_x - geoinfo.Left_Border_Size,
            geoinfo.pos_y - geoinfo.Upper_Border_Size,
            geoinfo.screen_size_x + geoinfo.Right_Border_Size + geoinfo.Left_Border_Size + its_manager->_ShadowSize,
            geoinfo.screen_size_y + geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size + its_manager->_ShadowSize,
            WinSystem::Show_Effect
        );
    }
    geoinfo.is_displayed = FALSE;
    its_manager->broadcast_delayed_message(this, W_MSG_HIDDEN);
    return TRUE;
}



/*************************************************************
 * make this wobject the bottom object
 *************************************************************/
int Wobject::lower()
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::lower",geoinfo.name,geoinfo.tag);
    #endif

    if (geoinfo.is_locked) return FALSE;

    int
        new_pos = its_manager->get_first_free_bottom_place();
     int its_position_in_order = its_manager->get_order(geoinfo.tag);
    
    if (its_position_in_order >= new_pos)
        return TRUE; // this already is the bottom wobject, so return success
    
    its_manager->broadcast_message(this, W_MSG_GOING_TO_LOWER);


    its_manager->lower(geoinfo.tag);

    its_manager->update_screen
    (
        geoinfo.pos_x - geoinfo.Left_Border_Size,
        geoinfo.pos_y - geoinfo.Upper_Border_Size,
        geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize + geoinfo.Left_Border_Size,
        geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize + geoinfo.Upper_Border_Size
    );

    its_manager->broadcast_message(this, W_MSG_LOWERED);
    return TRUE; // return success
}

/*************************************************************
 * make this wobject the top wobject
 *************************************************************/
int Wobject::raise()
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::raise",geoinfo.name,geoinfo.tag);
    #endif


    if (its_manager->parent)
        its_manager->parent->raise();

    if (geoinfo.is_locked) return FALSE;

    int 
        its_position_in_order = its_manager->get_order(geoinfo.tag);
    int
        number_locked = its_manager->get_number_of_locked_top_wobjects();

    
    if (its_position_in_order <= number_locked)
        return TRUE; // this already is the top wobject, so return success
    

    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::raise", geoinfo.name, geoinfo.tag);
    #endif
    its_manager->broadcast_message(this, W_MSG_GOING_TO_RAISE);


    if (its_manager->_ShadowStyle > 0)
        remove_shadow();

    its_manager->raise(geoinfo.tag);
    if (its_manager->UseBoxes && geoinfo.is_displayed)
    {
        its_manager->update_boxlists(this);
    }

    if (geoinfo.is_displayed)
        its_manager->display_wobject(geoinfo.tag,DISPLAY_MODE_FULL);
    if (geoinfo.is_displayed)
    {
        its_manager->update_screen
        (
            geoinfo.pos_x - geoinfo.Left_Border_Size,
            geoinfo.pos_y - geoinfo.Upper_Border_Size,
            geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize + geoinfo.Left_Border_Size,
            geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize + geoinfo.Upper_Border_Size 
        );
    }

    its_manager->broadcast_message(this, W_MSG_RAISED);
    return TRUE; // return success
}


/*************************************************************
 * make this wobject the top wobject, even if there are top-locked
 * objects, and lock it so the top locked will stay top locked
 *************************************************************/
int Wobject::raise_and_lock()
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::raise_and_lock",geoinfo.name,geoinfo.tag);
    #endif
    if (its_manager->parent)
        its_manager->parent->raise();

    int 
        its_position_in_order = its_manager->get_order(geoinfo.tag);
    int
        number_locked = its_manager->get_number_of_locked_top_wobjects();

    
    if (its_position_in_order <= number_locked && geoinfo.is_locked)
        return TRUE; // this already is the top wobject, so return success
    


    if (its_manager->_ShadowStyle && geoinfo.is_displayed)
        remove_shadow();

    its_manager->raise_guaranteed(geoinfo.tag);

    if (its_manager->UseBoxes && geoinfo.is_displayed)
    {
        its_manager->update_boxlists(this);
    }
    if (geoinfo.is_displayed)
    {
        its_manager->display_wobject(geoinfo.tag,DISPLAY_MODE_FULL);
        its_manager->update_screen
        (
            geoinfo.pos_x - geoinfo.Left_Border_Size,
            geoinfo.pos_y - geoinfo.Upper_Border_Size,
            geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize + geoinfo.Left_Border_Size,
            geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize + geoinfo.Upper_Border_Size 
        );
    }

    its_manager->broadcast_message(this, W_MSG_RAISED);
    geoinfo.is_locked = TRUE;
    return TRUE; // return success
}



/*************************************************************
 * return a pointer to the user bitmap
 *************************************************************/
BITMAP *Wobject::bmp()
{
    trace("object %s (%d) does not contain a bitmap\n"
          "if name is (null) this probably means\n"
          "a deleted object is accessed\n",geoinfo.name, geoinfo.tag);
    return NULL; //bare  object doesn't contain bitmap
}


/*************************************************************
 * bookkeeping flag
 *************************************************************/
int Wobject::delete_its_manager = 0;





/*************************************************************
 * move the wobject
 *************************************************************/
int Wobject::move(int nw_x,int nw_y)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::move",geoinfo.name,geoinfo.tag);
    #endif
    int refrx,refry,refrw,refrh;
    if (geoinfo.pos_is_locked) 
        return FALSE;
    if (nw_x == geoinfo.pos_x && nw_y == geoinfo.pos_y)
        return TRUE; //return success

    if (geoinfo.is_displayed)
    {
        if (see_through)
        {
            its_manager->remove_wobject(geoinfo.tag);
        }
        else
        {
            if (its_manager->_ShadowStyle)
            {
                remove_shadow();
            }
            if (nw_x > geoinfo.pos_x)
            {
                    its_manager->remove_wobject
                    (
                        geoinfo.tag,
                        geoinfo.pos_x - geoinfo.Left_Border_Size,
                        geoinfo.pos_y - geoinfo.Upper_Border_Size,
                        nw_x - geoinfo.Left_Border_Size,
                        MIN(geoinfo.pos_y,nw_y)  + geoinfo.screen_size_y+ geoinfo.Lower_Border_Size
                    );
            }
            else if (nw_x < geoinfo.pos_x)
            {
                    its_manager->remove_wobject
                    (
                        geoinfo.tag,
                        nw_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size,
                        MAX(geoinfo.pos_y,nw_y) - geoinfo.Upper_Border_Size,
                        geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize,
                        geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                    );
            }
        
            if (nw_y > geoinfo.pos_y)
            {
                    its_manager->remove_wobject
                    (
                        geoinfo.tag,
                        MAX(geoinfo.pos_x,nw_x) - geoinfo.Left_Border_Size,
                        geoinfo.pos_y - geoinfo.Upper_Border_Size,
                        geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize,
                        nw_y - geoinfo.Upper_Border_Size
                    );
            }
            else if (nw_y < geoinfo.pos_y)
            {
                    its_manager->remove_wobject
                    (
                        geoinfo.tag,
                        geoinfo.pos_x - geoinfo.Left_Border_Size,
                        nw_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size,
                        MIN(geoinfo.pos_x,nw_x)  + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize ,
                        geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                    );
            }
        }
    }
    if (geoinfo.pos_x<nw_x)
    {
        refrx = geoinfo.pos_x;
        refrw = nw_x + geoinfo.screen_size_x - geoinfo.pos_x;
    }
    else
    {
        refrx = nw_x;
        refrw = geoinfo.pos_x + geoinfo.screen_size_x - nw_x;
    }
    
    if (geoinfo.pos_y<nw_y)
    {
        refry = geoinfo.pos_y;
        refrh = nw_y + geoinfo.screen_size_y - geoinfo.pos_y;
    }
    else
    {
        refry = nw_y;
        refrh = geoinfo.pos_y + geoinfo.screen_size_y - nw_y;
    }


    geoinfo.pos_x = nw_x;
    geoinfo.pos_y = nw_y;
    if (geoinfo.is_displayed)
    {

        if (its_manager->UseBoxes)
          its_manager->update_boxlists(this);

        refrw += its_manager->_ShadowSize;
        refrh += its_manager->_ShadowSize;
        refrw += geoinfo.Right_Border_Size + geoinfo.Left_Border_Size;
        refrh += geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size;
        refrx -= geoinfo.Left_Border_Size;
        refry -= geoinfo.Upper_Border_Size;
        its_manager->display_wobject(geoinfo.tag,DISPLAY_MODE_FULL);
        its_manager->update_screen(refrx,refry,refrw,refrh);
    }
    its_manager->broadcast_message(this, W_MSG_MOVED);
    return TRUE; //return success
}

int Wobject::move_rescale(int nw_x,int nw_y, int nw_size_x, int nw_size_y)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::move_rescale",geoinfo.name,geoinfo.tag);
    #endif

    if (geoinfo.scale_is_locked) return FALSE;
    if (geoinfo.pos_is_locked) return FALSE;
    
    int refrx,refry,refrw,refrh;
    nw_size_x = MAX( nw_size_x , minimum_size_x);
    nw_size_y = MAX( nw_size_y , minimum_size_y);
    int was_displayed = geoinfo.is_displayed;
    if (was_displayed)
    {
        if (see_through)
        {
            its_manager->remove_wobject(geoinfo.tag);
        }
        else
        {
            if (its_manager->_ShadowStyle)
                remove_shadow();

            if (nw_x > geoinfo.pos_x)
            {
                its_manager->remove_wobject
                (
                    geoinfo.tag,
                    geoinfo.pos_x - geoinfo.Left_Border_Size,
                    geoinfo.pos_y - geoinfo.Upper_Border_Size,
                    nw_x - geoinfo.Left_Border_Size,
                    MIN(geoinfo.pos_y + geoinfo.screen_size_y,nw_y + nw_size_y) + geoinfo.Lower_Border_Size
                );
            }
            if (nw_y > geoinfo.pos_y)
            {
                its_manager->remove_wobject
                (
                    geoinfo.tag,
                    MAX(geoinfo.pos_x - geoinfo.Left_Border_Size,nw_x - geoinfo.Left_Border_Size),
                    geoinfo.pos_y - geoinfo.Upper_Border_Size,
                    geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize,
                    nw_y - geoinfo.Upper_Border_Size
                );
            }
            if (nw_x + nw_size_x < geoinfo.pos_x + geoinfo.screen_size_x)
            {   
                its_manager->remove_wobject
                (
                    geoinfo.tag,
                    nw_x + nw_size_x + geoinfo.Right_Border_Size,
                    MAX(geoinfo.pos_y,nw_y) - geoinfo.Upper_Border_Size,
                    geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize,
                    geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                );
            }
            if  (nw_y + nw_size_y < geoinfo.pos_y + geoinfo.screen_size_y)
            {
                its_manager->remove_wobject
                (
                    geoinfo.tag,
                    geoinfo.pos_x - geoinfo.Left_Border_Size,
                    nw_y + nw_size_y + geoinfo.Lower_Border_Size,
                    MIN(geoinfo.pos_x + geoinfo.screen_size_x,nw_x + nw_size_x) + geoinfo.Right_Border_Size + its_manager->_ShadowSize,
                    geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                );
            }   
        }
    }
    refrw = geoinfo.screen_size_x;
    refrh = geoinfo.screen_size_y;
    refrx = geoinfo.pos_x - geoinfo.Left_Border_Size;
    refry = geoinfo.pos_y - geoinfo.Upper_Border_Size;
    //rescale
    geoinfo.screen_size_x = nw_size_x;
    geoinfo.screen_size_y = nw_size_y;

    refrw = MAX(geoinfo.screen_size_x, refrw);
    refrh = MAX(geoinfo.screen_size_y, refrh);
    refrw += its_manager->_ShadowSize;
    refrh += its_manager->_ShadowSize;
    refrw += geoinfo.Right_Border_Size + geoinfo.Left_Border_Size;
    refrh += geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size;


    if (geoinfo.pos_x<nw_x)
    {
        refrw += nw_x -geoinfo.pos_x;
    }
    else
    {
        refrx -= geoinfo.pos_x  - nw_x;
        refrw += geoinfo.pos_x  - nw_x;
    }
    
    if (geoinfo.pos_y<nw_y)
    {
        refrh += nw_y - geoinfo.pos_y;
    }
    else
    {
        refry -= geoinfo.pos_y - nw_y;
        refrh += geoinfo.pos_y - nw_y;
    }


    geoinfo.pos_x = nw_x;
    geoinfo.pos_y = nw_y;

    geoinfo.scale_x = (float)geoinfo.screen_size_x/(float)geoinfo.viewport_size_x;
    geoinfo.scale_y = (float)geoinfo.screen_size_y/(float)geoinfo.viewport_size_y;

    if (was_displayed)
    {
        if (its_manager->UseBoxes)
            its_manager->update_boxlists(this);
        its_manager->display_wobject(geoinfo.tag,DISPLAY_MODE_FULL);
        its_manager->update_screen(refrx,refry,refrw,refrh);
    
    }
    its_manager->broadcast_message(this, W_MSG_MOVE_RESCALED);
    return TRUE; //return success


}




/*************************************************************
 * rescale the object on the screen
 * note that the size of the memory bitmap
 * stays the same 
 * this only influences the way it'l look on screen
 *************************************************************/
int Wobject::rescale(int nw_size_x, int nw_size_y)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::rescale",geoinfo.name,geoinfo.tag);
    #endif

    if (geoinfo.scale_is_locked) return FALSE;
    int refrw,refrh;
    if (nw_size_x == geoinfo.screen_size_x && nw_size_y == geoinfo.screen_size_y)
        return TRUE;
    nw_size_x = MAX(nw_size_x,minimum_size_x);
    nw_size_y = MAX(nw_size_y,minimum_size_y);
    int was_displayed = geoinfo.is_displayed;

    // it needs to be removed if the new window is smaller than the old one or it contains see_through parts
    if (was_displayed && (( nw_size_x < geoinfo.screen_size_x || nw_size_y < geoinfo.screen_size_y ) || see_through) )
    {
        if (see_through)
        {
            its_manager->remove_wobject(geoinfo.tag);
        }
        else
        {
            if (its_manager->_ShadowStyle)
                remove_shadow();
                
            if (nw_size_x < geoinfo.screen_size_x)
            {

                its_manager->remove_wobject
                (
                    geoinfo.tag,
                    geoinfo.pos_x + nw_size_x + geoinfo.Right_Border_Size , 
                    geoinfo.pos_y - geoinfo.Upper_Border_Size,
                    geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize, 
                    geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                );
            }
    #ifdef SLOW
        rest(500);
    #endif

            if (nw_size_y < geoinfo.screen_size_y)
            {
                its_manager->remove_wobject
                (
                    geoinfo.tag,
                    geoinfo.pos_x - geoinfo.Left_Border_Size, 
                    geoinfo.pos_y + nw_size_y + geoinfo.Lower_Border_Size,
                    geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize, 
                    geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                );
            }
        }
    }
    #ifdef SLOW
        rest(500);
    #endif

    refrw = geoinfo.screen_size_x;
    refrh = geoinfo.screen_size_y;
    geoinfo.screen_size_x = nw_size_x;
    geoinfo.screen_size_y = nw_size_y;
    refrw = MAX(geoinfo.screen_size_x,refrw);
    refrh = MAX(geoinfo.screen_size_y,refrh);
    refrw += its_manager->_ShadowSize;
    refrh += its_manager->_ShadowSize;
    refrw += geoinfo.Right_Border_Size + geoinfo.Left_Border_Size;
    refrh += geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size;

    geoinfo.scale_x = (float)geoinfo.screen_size_x/(float)geoinfo.viewport_size_x;
    geoinfo.scale_y = (float)geoinfo.screen_size_y/(float)geoinfo.viewport_size_y;
    if (was_displayed)
    {
        if (its_manager->UseBoxes)
            its_manager->update_boxlists(this);
        its_manager->display_wobject(geoinfo.tag,DISPLAY_MODE_FULL);
        its_manager->update_screen(geoinfo.pos_x - geoinfo.Left_Border_Size,geoinfo.pos_y - geoinfo.Upper_Border_Size,refrw,refrh);
    }   
    its_manager->broadcast_message(this, W_MSG_RESCALED);
    return TRUE; // return success
}



/*************************************************************
 * some info functions
 *************************************************************/
GeoInfo const *Wobject::get_geometric_info() 
{
    return &geoinfo;
}

int Wobject::get_tag()
{
    return geoinfo.tag;
}


WinSystem *Wobject::get_manager()
{
    return its_manager;
}


/*************************************************************
 * locking functions
 *************************************************************/

void Wobject::set_lock(int lock)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::set_lock(%d)",geoinfo.name,geoinfo.tag,lock);
    #endif
    int was = geoinfo.is_locked;
    geoinfo.is_locked = lock;
    if (was != lock) 
    {
        if (lock)
            its_manager->broadcast_message(this, W_MSG_LOCKED);
        else
            its_manager->broadcast_message(this, W_MSG_UNLOCKED);
    }
}

void Wobject::set_lock_shift(int lock)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::set_lock_shift(%d)",geoinfo.name,geoinfo.tag,lock);
    #endif
    geoinfo.shift_is_locked = lock;
}

void Wobject::set_lock_visibility(int lock)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::set_lock_visibility(%d)",geoinfo.name,geoinfo.tag,lock);
    #endif
    geoinfo.visibility_is_locked = lock;
}


void Wobject::set_lock_position(int lock)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::set_lock_position(%d)",geoinfo.name,geoinfo.tag,lock);
    #endif
    geoinfo.pos_is_locked = lock;
}

void Wobject::set_lock_size(int lock)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::set_lock_size(%d)",geoinfo.name,geoinfo.tag,lock);
    #endif
    geoinfo.size_is_locked = lock;
}
void Wobject::set_lock_scale(int lock)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::set_lock_scale(%d)",geoinfo.name,geoinfo.tag,lock);
    #endif
    geoinfo.scale_is_locked = lock;
}
void Wobject::set_lock_viewport_size(int lock)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::set_lock_viewport_size(%d)",geoinfo.name,geoinfo.tag,lock);
    #endif
    geoinfo.viewport_size_is_locked = lock;
}
void Wobject::set_lock_viewport_position(int lock)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::set_lock_viewport_position(%d)",geoinfo.name,geoinfo.tag,lock);
    #endif
    geoinfo.viewport_pos_is_locked = lock;
}

void Wobject::transform_coordinates(int *x,int *y)
{
    #ifdef WTRACE
        //if (wham_trace) trace("%s (%d) Wobject::transform_coordinates",geoinfo.name,geoinfo.tag);
    #endif
    if (its_manager->parent)
        its_manager->parent->transform_coordinates(x,y);
    *x -= geoinfo.pos_x;
    *y -= geoinfo.pos_y;
    *x *= geoinfo.viewport_size_x;
    *y *= geoinfo.viewport_size_y;
    *x /= geoinfo.screen_size_x;
    *y /= geoinfo.screen_size_y;
    *x -= geoinfo.user_pos_x;
    *y -= geoinfo.user_pos_y;
    *x += geoinfo.viewport_pos_x;
    *y += geoinfo.viewport_pos_y;
}

int Wobject::move_viewport(int nw_x, int nw_y)
{

    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::move_viewport",geoinfo.name,geoinfo.tag);
    #endif
    if (get_geoinfo()->viewport_pos_is_locked) 
        return FALSE;

    if (nw_x == get_geoinfo()->viewport_pos_x && nw_y == get_geoinfo()->viewport_pos_y)
        return TRUE; //return success

    nw_x = MIN(nw_x,get_geoinfo()->size_x - get_geoinfo()->viewport_size_x);
    nw_y = MIN(nw_y,get_geoinfo()->size_y - get_geoinfo()->viewport_size_y);
    nw_x = MAX(nw_x,0);
    nw_y = MAX(nw_y,0);

    nw_x = MIN(geoinfo.size_x - geoinfo.viewport_size_x, nw_x);
    nw_y = MIN(geoinfo.size_y - geoinfo.viewport_size_y, nw_y);


    if (nw_x != geoinfo.viewport_pos_x || nw_y !=geoinfo.viewport_pos_y)
    {
    	geoinfo.viewport_pos_x = nw_x;
	   geoinfo.viewport_pos_y = nw_y;


      if (get_geoinfo()->is_displayed)
        show();
      its_manager->broadcast_message(this, W_MSG_VIEWPORT_MOVED);
    }


    return TRUE; 
}



void Wobject::point(BITMAP *onto,int x, int y)
{
   onto = NULL;
   x = y = 0; // silence compiler 'bout unused parameters
	// derived classes sould do something usefule with it
	// like putpixel(onto,x,y, getpixel(object)); this one just returns
   return;
}


/*************************************************************
 * draws the portion of the shadow that is inside the rectangle
 * given by the coordinates
 *************************************************************/
void Wobject::draw_shadow(BITMAP *onto, int from_x, int from_y, int to_x, int to_y)
{
    #ifdef WTRACE
        if (wham_trace) trace("Wobject::draw_shadow of %s (%d)",geoinfo.name, geoinfo.tag);
    #endif

    //adjust coordinates on screen
    int shadow_from_x = MAX(from_x,get_geoinfo()->pos_x + get_geoinfo()->screen_size_x + get_geoinfo()->Right_Border_Size);
    int shadow_from_y = MAX(from_y,get_geoinfo()->pos_y + get_geoinfo()->screen_size_y + get_geoinfo()->Lower_Border_Size);
    int shadow_to_x = MIN(to_x,get_geoinfo()->pos_x + get_geoinfo()->screen_size_x + get_manager()->_ShadowSize + get_geoinfo()->Right_Border_Size);
    int shadow_to_y = MIN(to_y,get_geoinfo()->pos_y + get_geoinfo()->screen_size_y + get_manager()->_ShadowSize + get_geoinfo()->Lower_Border_Size);
    
    //adjust coordinates on screen
    from_x = MAX(from_x,get_geoinfo()->pos_x - get_geoinfo()->Left_Border_Size  + get_manager()->_ShadowSize);
    from_y = MAX(from_y,get_geoinfo()->pos_y - get_geoinfo()->Upper_Border_Size + get_manager()->_ShadowSize);
    to_x = MIN(to_x,get_geoinfo()->pos_x + get_geoinfo()->screen_size_x + get_geoinfo()->Right_Border_Size);
    to_y = MIN(to_y,get_geoinfo()->pos_y + get_geoinfo()->screen_size_y + get_geoinfo()->Lower_Border_Size);
    
    

    // set up drawing style for shadow
    switch(its_manager->_ShadowStyle)
    {
        case WHAM_SLOW_SHADOW:
            drawing_mode(DRAW_MODE_TRANS,0,0,0);
            break;
        case WHAM_FAST_SHADOW:
            drawing_mode(DRAW_MODE_XOR,0,0,0);
            break;
        case WHAM_SIMPLE_SHADOW:
            drawing_mode(DRAW_MODE_SOLID,0,0,0);
            break;
        case WHAM_PATTERN_SHADOW:
            drawing_mode(DRAW_MODE_MASKED_PATTERN,get_manager()->shadow_pattern,0,0);
            break;

        default:
            fatal("WHAM:illegal shadow drawing style");
    }

    if (shadow_to_x > shadow_from_x && from_y < to_y)
    {
        rectfill
        (
            onto,
            shadow_from_x,
            from_y, 
            shadow_to_x - 1,
            to_y - 1,
            get_manager()->_ShadowColor
        );
    }
    if (shadow_to_y > shadow_from_y && from_x < to_x)
    {
        rectfill
        (
            onto,
            from_x,
            shadow_from_y, 
            to_x - 1,
            shadow_to_y - 1,
            get_manager()->_ShadowColor
        );
    }
    if (shadow_to_y > shadow_from_y && shadow_from_x < shadow_to_x)
    {
        rectfill
        (
            onto,
            shadow_from_x,
            shadow_from_y,
            shadow_to_x - 1,
            shadow_to_y - 1,
            get_manager()->_ShadowColor
        );
    }
    drawing_mode(DRAW_MODE_SOLID,0,0,0);

}    // end of shadow code

// removes the shadow
void Wobject::remove_shadow()
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::remove shadow",geoinfo.name,geoinfo.tag);
    #endif
                remove_shadow
                (
                    geoinfo.pos_x - geoinfo.Left_Border_Size,
                    geoinfo.pos_y - geoinfo.Upper_Border_Size,
                    geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize,
                    geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                );
                //shadow removed
}

/*************************************************************
 * removes the portion of the shadow that is inside the rectangle
 * given by the coordinates
 *************************************************************/
void Wobject::remove_shadow(int from_x, int from_y, int to_x, int to_y)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::remove_shadow(%d %d %d %d)",geoinfo.name,geoinfo.tag, from_x,from_y, to_x, to_y);
    #endif

    //adjust coordinates on screen
    int shadow_from_x = MAX(from_x,get_geoinfo()->pos_x + get_geoinfo()->screen_size_x + get_geoinfo()->Right_Border_Size) ;
    int shadow_from_y = MAX(from_y,get_geoinfo()->pos_y + get_geoinfo()->screen_size_y + get_geoinfo()->Lower_Border_Size);
    int shadow_to_x = MIN(to_x,get_geoinfo()->pos_x + get_geoinfo()->screen_size_x + get_manager()->_ShadowSize + get_geoinfo()->Right_Border_Size);
    int shadow_to_y = MIN(to_y,get_geoinfo()->pos_y + get_geoinfo()->screen_size_y + get_manager()->_ShadowSize + get_geoinfo()->Lower_Border_Size);
    
    //adjust coordinates on screen
    from_x = MAX(from_x,get_geoinfo()->pos_x - get_geoinfo()->Left_Border_Size + its_manager->_ShadowSize);
    from_y = MAX(from_y,get_geoinfo()->pos_y - get_geoinfo()->Upper_Border_Size + its_manager->_ShadowSize);
    to_x = MIN(to_x,get_geoinfo()->pos_x + get_geoinfo()->screen_size_x + get_geoinfo()->Right_Border_Size);
    to_y = MIN(to_y,get_geoinfo()->pos_y + get_geoinfo()->screen_size_y + get_geoinfo()->Lower_Border_Size);
    
    if (shadow_to_x > shadow_from_x && from_y < to_y)
    {
        its_manager->remove_wobject
        (
            geoinfo.tag,
            shadow_from_x,
            from_y, 
            shadow_to_x ,
            to_y
        );
    }
    if (shadow_to_y > shadow_from_y && from_x < to_x)
    {
        its_manager->remove_wobject
        (
            geoinfo.tag,
            from_x,
            shadow_from_y, 
            to_x ,
            shadow_to_y 
        );
    }
    if (shadow_to_y > shadow_from_y && shadow_from_x < shadow_to_x)
    {
        its_manager->remove_wobject
        (
            geoinfo.tag,
            shadow_from_x,
            shadow_from_y,
            shadow_to_x ,
            shadow_to_y 
        );
    }
}    // end of shadow removing code

/*************************************************************
 * set the size & position of the user area
 *************************************************************/
void Wobject::set_user_area(int _pos_x,int _pos_y,int _size_x, int _size_y)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::set_user_area",geoinfo.name,geoinfo.tag);
    #endif


    geoinfo.user_pos_x = _pos_x;
    geoinfo.user_pos_y = _pos_y;
    geoinfo.user_size_x = _size_x;
    geoinfo.user_size_y = _size_y;
    its_manager->broadcast_message(this, W_MSG_USER_AREA_SET);
}


int Wobject::resize(int nw_size_x,int nw_size_y)
{
    nw_size_x = MAX(nw_size_x,minimum_size_x);
    nw_size_y = MAX(nw_size_y,minimum_size_y);
    
    geoinfo.user_pos_x =  geoinfo.user_pos_x >= nw_size_x ? nw_size_x - 1 :  geoinfo.user_pos_x;
    geoinfo.user_pos_y =  geoinfo.user_pos_y >= nw_size_y ? nw_size_y - 1 :  geoinfo.user_pos_y;
    geoinfo.user_size_x = geoinfo.user_size_x +  geoinfo.user_pos_x >= nw_size_x ? nw_size_x -  geoinfo.user_pos_x - 1: geoinfo.user_size_x;
    geoinfo.user_size_y = geoinfo.user_size_y +  geoinfo.user_pos_y >= nw_size_y ? nw_size_y -  geoinfo.user_pos_y - 1: geoinfo.user_size_y;
    geoinfo.viewport_size_x = geoinfo.viewport_size_x +  geoinfo.viewport_pos_x >= nw_size_x ? nw_size_x -  geoinfo.viewport_pos_x - 1: geoinfo.viewport_size_x;
    geoinfo.viewport_size_y = geoinfo.viewport_size_y +  geoinfo.viewport_pos_y >= nw_size_y ? nw_size_y -  geoinfo.viewport_pos_y - 1: geoinfo.viewport_size_y;
    geoinfo.size_x = nw_size_x;
    geoinfo.size_y = nw_size_y;



    its_manager->broadcast_message(this, W_MSG_RESIZED);
    if (its_manager->UseBoxes && geoinfo.is_displayed)
         its_manager->update_boxlists(this);
    return TRUE;
 
}


// the ccordinates are screen-relative
int Wobject::resize_s_viewport(int nw_size_x,int nw_size_y)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::resize_s_viewport",geoinfo.name,geoinfo.tag);
    #endif

    int refrw,refrh;
    int nw_vp_pos_x = geoinfo.viewport_pos_x;
    int nw_vp_pos_y = geoinfo.viewport_pos_y;
    int nw_vp_size_x,nw_vp_size_y;
    if (geoinfo.viewport_size_is_locked)
        return FALSE;



//    nw_size_x = MIN(nw_size_x,(int)((geoinfo.size_x - geoinfo.viewport_pos_x) * geoinfo.scale_x));
//    nw_size_y = MIN(nw_size_y,(int)((geoinfo.size_y - geoinfo.viewport_pos_y) * geoinfo.scale_y));
    nw_size_x = MIN(nw_size_x,(int)(geoinfo.size_x  * geoinfo.scale_x));
    nw_size_y = MIN(nw_size_y,(int)(geoinfo.size_y  * geoinfo.scale_y));
    nw_size_x = MAX(nw_size_x,minimum_size_x);
    nw_size_y = MAX(nw_size_y,minimum_size_y);

    if (nw_size_x == geoinfo.screen_size_x && nw_size_y == geoinfo.screen_size_y)
        return TRUE;

    nw_vp_size_x = (int)(nw_size_x / geoinfo.scale_x);
    nw_vp_size_y = (int)(nw_size_y / geoinfo.scale_y);


    // hack to catch roundoff errors in scaling, so it's impossible to resize
    // the viewport so that it is exactly 1 pixel smaller than the window
    // if it is scaled smaller
    
    if (nw_vp_size_x == geoinfo.size_x - 1 && geoinfo.scale_x < 1.0)
       nw_vp_size_x++;

    if (nw_vp_size_y == geoinfo.size_y - 1 && geoinfo.scale_y < 1.0)
       nw_vp_size_y++;

    if (geoinfo.viewport_pos_x + nw_vp_size_x > geoinfo.size_x)
                nw_vp_pos_x = geoinfo.size_x  - nw_vp_size_x;

    if (geoinfo.viewport_pos_y + nw_vp_size_y > geoinfo.size_y )
                nw_vp_pos_y = geoinfo.size_y  - nw_vp_size_y;
    
        

    int was_displayed = geoinfo.is_displayed;

    // it needs to be removed if the new window is smaller than the old one or it contains see_through parts
    if (was_displayed && (( nw_size_x < geoinfo.screen_size_x || nw_size_y < geoinfo.screen_size_y ) || see_through) )
    {
        if (see_through)
        {
            its_manager->remove_wobject(geoinfo.tag);
        }
        else
        {
            if (its_manager->_ShadowStyle)
                remove_shadow();
            if (nw_size_x < geoinfo.screen_size_x)
            {
                its_manager->remove_wobject
                (
                    geoinfo.tag,
                    geoinfo.pos_x + nw_size_x + geoinfo.Right_Border_Size , 
                    geoinfo.pos_y - geoinfo.Upper_Border_Size,
                    geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize, 
                    geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                );
            }
    #ifdef SLOW
        rest(500);
    #endif

            if (nw_size_y < geoinfo.screen_size_y)
            {
                its_manager->remove_wobject
                (
                    geoinfo.tag,
                    geoinfo.pos_x - geoinfo.Left_Border_Size, 
                    geoinfo.pos_y + nw_size_y + geoinfo.Lower_Border_Size,
                    geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize, 
                    geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                );
            }
        }
    }
    #ifdef SLOW
        rest(500);
    #endif

    refrw = geoinfo.screen_size_x;
    refrh = geoinfo.screen_size_y;
    geoinfo.viewport_size_x = nw_vp_size_x;
    geoinfo.viewport_size_y = nw_vp_size_y;
    geoinfo.screen_size_x = nw_size_x;
    geoinfo.screen_size_y = nw_size_y;
    refrw = MAX(geoinfo.screen_size_x,refrw);
    refrh = MAX(geoinfo.screen_size_y,refrh);
    refrw += its_manager->_ShadowSize;
    refrh += its_manager->_ShadowSize;
    refrw += geoinfo.Right_Border_Size + geoinfo.Left_Border_Size;
    refrh += geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size;
    geoinfo.viewport_pos_x = nw_vp_pos_x;
    geoinfo.viewport_pos_y = nw_vp_pos_y;

    if (was_displayed)
    {
        if (its_manager->UseBoxes)
        {
            its_manager->update_boxlists(this);
        }
        its_manager->display_wobject(geoinfo.tag,DISPLAY_MODE_FULL);
        its_manager->update_screen(geoinfo.pos_x - geoinfo.Left_Border_Size,geoinfo.pos_y - geoinfo.Upper_Border_Size,refrw,refrh);
    }   
    its_manager->broadcast_message(this, W_MSG_VIEWPORT_RESIZED);
    return TRUE; // return success

}

int Wobject::resize_viewport(int nw_v_size_x,int nw_v_size_y)
{
    if (geoinfo.viewport_size_is_locked)
        return FALSE;

    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::resize_viewport",geoinfo.name,geoinfo.tag);
    #endif
    nw_v_size_x = MIN(nw_v_size_x,geoinfo.size_x);
    nw_v_size_y = MIN(nw_v_size_y,geoinfo.size_y);
    
    nw_v_size_x = MAX(nw_v_size_x,(int)(minimum_size_x / geoinfo.scale_x));
    nw_v_size_y = MAX(nw_v_size_y,(int)(minimum_size_y / geoinfo.scale_y));

    int nw_v_pos_x = MIN(geoinfo.viewport_pos_x, geoinfo.size_x - nw_v_size_x);
    int nw_v_pos_y = MIN(geoinfo.viewport_pos_y, geoinfo.size_y - nw_v_size_y);
    int nw_size_x =  (int)(nw_v_size_x * geoinfo.scale_x);
    int nw_size_y =  (int)(nw_v_size_y * geoinfo.scale_y);

    int refrw,refrh;
    if (nw_v_size_x == geoinfo.viewport_size_x && nw_v_size_y == geoinfo.viewport_size_y)
        return TRUE;
    int was_displayed = geoinfo.is_displayed;

    // it needs to be removed if the new window is smaller than the old one or it contains see_through parts
    if (was_displayed && (( nw_size_x < geoinfo.screen_size_x || nw_size_y < geoinfo.screen_size_y ) || see_through) )
    {
        if (see_through)
        {
            its_manager->remove_wobject(geoinfo.tag);
        }
        else
        {
            if (its_manager->_ShadowStyle)
                remove_shadow();
            if (nw_size_x < geoinfo.screen_size_x)
            {
                its_manager->remove_wobject
                (
                    geoinfo.tag,
                    geoinfo.pos_x + nw_size_x + geoinfo.Right_Border_Size , 
                    geoinfo.pos_y - geoinfo.Upper_Border_Size,
                    geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize, 
                    geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                );
            }
    #ifdef SLOW
        rest(500);
    #endif

            if (nw_size_y < geoinfo.screen_size_y)
            {
                its_manager->remove_wobject
                (
                    geoinfo.tag,
                    geoinfo.pos_x - geoinfo.Left_Border_Size, 
                    geoinfo.pos_y + nw_size_y + geoinfo.Lower_Border_Size,
                    geoinfo.pos_x + geoinfo.screen_size_x + geoinfo.Right_Border_Size + its_manager->_ShadowSize, 
                    geoinfo.pos_y + geoinfo.screen_size_y + geoinfo.Lower_Border_Size + its_manager->_ShadowSize
                );
            }
        }
    }
    #ifdef SLOW
        rest(500);
    #endif

    refrw = geoinfo.screen_size_x;
    refrh = geoinfo.screen_size_y;
    
    
    geoinfo.viewport_size_x =nw_v_size_x;
    geoinfo.viewport_size_y =nw_v_size_y;
    geoinfo.viewport_pos_x = nw_v_pos_x;
    geoinfo.viewport_pos_y = nw_v_pos_y;
    geoinfo.screen_size_x = nw_size_x;
    geoinfo.screen_size_y = nw_size_y;


    refrw = MAX(geoinfo.screen_size_x,refrw);
    refrh = MAX(geoinfo.screen_size_y,refrh);
    refrw += its_manager->_ShadowSize;
    refrh += its_manager->_ShadowSize;
    refrw += geoinfo.Right_Border_Size + geoinfo.Left_Border_Size;
    refrh += geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size;

    geoinfo.scale_x = (float)geoinfo.screen_size_x/(float)geoinfo.viewport_size_x;
    geoinfo.scale_y = (float)geoinfo.screen_size_y/(float)geoinfo.viewport_size_y;
    if (was_displayed)
    {
        if (its_manager->UseBoxes)
            its_manager->update_boxlists(this);
        its_manager->display_wobject(geoinfo.tag,DISPLAY_MODE_FULL);
        its_manager->update_screen(geoinfo.pos_x - geoinfo.Left_Border_Size,geoinfo.pos_y - geoinfo.Upper_Border_Size,refrw,refrh);
    }   
    its_manager->broadcast_message(this, W_MSG_VIEWPORT_RESIZED);
    return TRUE; // return success

}


void Wobject::notify_children()
{
    if (Child_manager)
        Child_manager->parent_bitmap_has_changed();
}

/*********************************************************************************
 * freezes the window, the window will not be updated until Wobject::thaw is called
 *********************************************************************************/
 
void Wobject::freeze()
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::freeze",geoinfo.name,geoinfo.tag);
    #endif
    frozen++;
    its_manager->broadcast_message(this, W_MSG_FROZEN);
    
}


/*********************************************************************************
 * thaws a frozen window (see freeze())
 *********************************************************************************/
void Wobject::thaw()
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::thaw",geoinfo.name,geoinfo.tag);
    #endif
    frozen--;
    if (geoinfo.is_displayed && !frozen)
    {
    
        its_manager->display_wobject(geoinfo.tag,DISPLAY_MODE_FULL);
        its_manager->update_screen
        (
            geoinfo.pos_x - geoinfo.Left_Border_Size,
            geoinfo.pos_y - geoinfo.Upper_Border_Size,
            geoinfo.screen_size_x + geoinfo.Right_Border_Size + geoinfo.Left_Border_Size + its_manager->_ShadowSize,
            geoinfo.screen_size_y + geoinfo.Lower_Border_Size + geoinfo.Upper_Border_Size + its_manager->_ShadowSize
        );

    }
    if (frozen < 0)
    {
     frozen = 0;
     trace("wobject::thaw() without wobject::freeze()");
    }
    its_manager->broadcast_message(this, W_MSG_THAWN);
}

/****************************************
 * returns temperature :)
 ****************************************/
int Wobject::is_frozen()
{
    return frozen;
}

int Wobject::process_keyboard()
{
    return 0;
}


Wobject *Wobject::process_mouse(int *x_out, int *y_out)
{
    Wobject *win;
    int out_x,out_y,flag;
    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);

    if (x_out)
        *x_out = out_x;
    if (y_out)
        *y_out = out_y;

    return win;
}

int Wobject::inquire_mouse(int x, int y)
{
    x= y=0;  // silence compiler
    return TRUE;    // since the manager calls this function only if the 
                    // coordinates lie within the window, return always true
}



int Wobject::process_message(Wobject  *from,int message)
{

    #ifdef DEBUGMESSAGESYS
        if (from)
        {
            if (wham_trace) trace("message %s received by %s from %s",message_string(message), geoinfo.name,from->geoinfo.name);
        }
        else
        {
            if (wham_trace) trace("message %s received by %s ",message_string(message), geoinfo.name);
        }
    #endif
    if (from == attached && from)
    {
        switch(message)
        {
            case W_MSG_HIDDEN:
                hide();
                break;
            case W_MSG_SHOWN:
                show();
                break;
            case W_MSG_GOING_TO_LOWER:
                lower();
                break;
            case W_MSG_GOING_TO_RAISE:
                raise();
                break;
            case W_MSG_VIEWPORT_RESIZED:
                //fall through
            case W_MSG_MOVE_RESCALED:
                //fall through
            case W_MSG_RESCALED:
                //fall through
            case W_MSG_RESIZED:
                //fall through
            case W_MSG_MOVED:
                move_to_attachpoint();
                break;
            case W_MSG_DESTROYED:
                attached = NULL;
                break;
            case W_MSG_ATTACHED:
                corner_this = from->corner_that;
                corner_that = from->corner_this;
                x_offset = - from->x_offset;
                y_offset = - from->y_offset;
                break;
            case W_MSG_LOCKED:
                set_lock(TRUE);
                break;
            case W_MSG_UNLOCKED:
                set_lock(FALSE);
                break;
            case W_MSG_RAISE_AND_LOCKED:
                raise_and_lock();
                break;
            default:
                break;
        }       

    }
    switch(message)
    {
        case W_MSG_RAISE:
            return raise();
            break;
        case W_MSG_LOWER:
            return lower();
            break;
        case W_MSG_HIDE:
            return hide();
            break;
        case W_MSG_UPDATE:
            update();
            return TRUE;
            break;
        case W_MSG_SHOW:
            return show();
            break;
        case W_MSG_DESTROY:
            request(W_REQUEST_DELETE);
            break;
        case W_MSG_GOTMOUSE:
            geoinfo.has_focus = TRUE;
            refresh_border();
            break;
        case W_MSG_LOSTMOUSE:
            geoinfo.has_focus = FALSE;
            refresh_border();
            break;
        default:
            return FALSE;
            break;
    }
                
   return TRUE;
}


void Wobject::attach_to(Wobject *to,int _corn_this,int _corn_that,int offsx,int offsy)
{
    #ifdef WTRACE
        if (wham_trace) trace("%s (%d) Wobject::atatch_to %s (%d)",geoinfo.name,geoinfo.tag,to->geoinfo.name,to->geoinfo.tag);
    #endif

    attached = to;
    corner_this = _corn_this;
    corner_that = _corn_that;
    x_offset = offsx;
    y_offset = offsy;
    its_manager->send_message(this,W_MSG_ATTACHED,to);
    move_to_attachpoint();
}

void Wobject::move_to_attachpoint()
{
    GeoInfo 
        const *gthat = attached->get_geoinfo();
    int xto = gthat->pos_x + x_offset;
    int yto = gthat->pos_y + y_offset;
    switch (corner_this)
    {
        case WINDOW_UL_CORNER:
            xto += geoinfo.Left_Border_Size;
            yto += geoinfo.Upper_Border_Size;
            break;
        case WINDOW_UR_CORNER:
            xto -= geoinfo.Right_Border_Size;
            xto -= geoinfo.screen_size_x;
            yto += geoinfo.Upper_Border_Size;
            break;
        case WINDOW_LL_CORNER:
            xto += geoinfo.Left_Border_Size;
            yto -= geoinfo.Lower_Border_Size;
            yto -= geoinfo.screen_size_y;
            break;
        case WINDOW_LR_CORNER:
            xto -= geoinfo.Right_Border_Size;
            xto -= geoinfo.screen_size_x;
            yto -= geoinfo.Lower_Border_Size;
            yto -= geoinfo.screen_size_y;

            break;
    }
    switch (corner_that)
    {
        case WINDOW_UL_CORNER:
            xto -= gthat->Left_Border_Size;
            yto -= gthat->Upper_Border_Size;
            break;
        case WINDOW_UR_CORNER:
            xto += gthat->Right_Border_Size;
            xto += gthat->screen_size_x;
            yto -= gthat->Upper_Border_Size;
            break;
        case WINDOW_LL_CORNER:
            xto -= gthat->Left_Border_Size;
            yto += gthat->Lower_Border_Size;
            yto += gthat->screen_size_y;
            break;
        case WINDOW_LR_CORNER:
            xto += gthat->Right_Border_Size;
            xto += gthat->screen_size_x;
            yto += gthat->Lower_Border_Size;
            yto += gthat->screen_size_y;

            break;
    }
    move(xto,yto);
}


void Wobject::re_init_boxlist()
{
    if (my_boxlist)
    {
        #ifdef DEBUGBOXLIST
            if (wham_trace) trace("%s (%d) re_init_boxlist:deleting old",geoinfo.name,geoinfo.tag);
        #endif
        while (my_boxlist->next)
        {
            #ifdef DEBUGBOXLIST
                if (wham_trace) trace("deleting %d",my_boxlist->next->tag);
            #endif

            delete my_boxlist->next;
        }
        #ifdef DEBUGBOXLIST
            if (wham_trace) trace("delete the last old one %d",my_boxlist->tag);
        #endif 
        delete my_boxlist;
    }
    #ifdef DEBUGBOXLIST
        if (wham_trace) trace("re_init_boxlist :making new");
    #endif    
    my_boxlist = new Boxlist
                    (
                        this,
                        geoinfo.pos_x - geoinfo.Left_Border_Size,
                        geoinfo.pos_y - geoinfo.Upper_Border_Size,
                        geoinfo.pos_x + geoinfo.Right_Border_Size + geoinfo.screen_size_x + its_manager->_ShadowSize,
                        geoinfo.pos_y + geoinfo.Lower_Border_Size + geoinfo.screen_size_y + its_manager->_ShadowSize
                    );

    if (!my_boxlist)
        out_of_mem("Try setting WinSystem::UseBoxes to FALSE");
}


void Wobject::request(int _request)
{
   #ifdef WTRACE
		if (wham_trace) trace("%s (%d) Wobject::request %d",geoinfo.name,geoinfo.tag,_request);
   #endif
	the_request = _request;
}

// all objects registered will be automagically deletd when this object is deleted
// this is useful when you attach windows to each other and do not want to keep track of them
// also, wobjects created in the constructor of another wobject should not be deleted
// from the destructor, but registered to let the system deal with it
// for the WinSystem class might need to delete all windows, and it gets confused
// if some objects delete other objects themselves
void Wobject::register_as_property(Wobject *w)
{
    #ifdef WTRACE
        if (wham_trace) trace("Wobject::register as property %s(%d) registers %s(%d)",geoinfo.name,geoinfo.tag,w->geoinfo.name,w->geoinfo.tag);
    #endif


   if (!belongings)
   {
      belongings = new Stack(20);
   }

   belongings->push((void *)w);
}

void Wobject::refresh_border_info()
{
    if (draw_border)
    {
        geoinfo.Upper_Border_Size = draw_border(NULL,NULL,WINDOW_UPPER_BORDER,0,0,0);
        geoinfo.Lower_Border_Size = draw_border(NULL,NULL,WINDOW_LOWER_BORDER,0,0,0);
        geoinfo.Right_Border_Size = draw_border(NULL,NULL,WINDOW_RIGHT_BORDER,0,0,0);
        geoinfo.Left_Border_Size = draw_border(NULL,NULL,WINDOW_LEFT_BORDER,0,0,0);
    }
    else
    {
        geoinfo.Upper_Border_Size = 0;
        geoinfo.Lower_Border_Size = 0;
        geoinfo.Right_Border_Size = 0;
        geoinfo.Left_Border_Size = 0;
    }

}


void Wobject::set_destruction_callback(void (*callback)(Wobject *))
{
   destruct_callback = callback;
}
