#include <stdio.h>
#include "s_wham.h"
#define THRESH 1000000
#define GRAV   -300
#define FRICT .99
#define radius 900
#define MAXSPEED 15
#define PULL 0

int NUMBER_OF_PLANETS = 14;


void init_planets(int size_x, int size_y);
void draw_planets_frame(BITMAP *onto);
void cleanup_planets();

char the_string[100] = ":)";


int exit_flag = FALSE;
int set_exit_flag()
{
	exit_flag = TRUE;
   return 0;
}

MENU my_menu[] =
{
    {"Exit",set_exit_flag,NULL,0,NULL},
    {"testing1",NULL,NULL,0,NULL},
    {"testing2",NULL,NULL,0,NULL},
    {NULL,NULL,NULL,0,NULL}
};


DIALOG dialog[] =
{
	/* (dialog proc)     (x)   (y)   (w)   (h)   (fg)  (bg)  (key) (flags)  (d1)  (d2)  (dp) */
	{ d_edit_proc,       80,   32,   512,  48,   255,  3,    0,    0,       16,   0,    the_string },
	{ d_button_proc,     80,   132,  160,  48,   255,  3,    't',  0,       0,    0,    const_cast<char *>("&Toggle Me") },
	{ d_button_proc,     80,   300,  160,  48,   255,  3,    0,    D_EXIT,  0,    0,    const_cast<char *>("OK") },
	{ NULL,              0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL }
};

unsigned int Utimer;

void user_timer()
{
	Utimer++;
}

int main()
{
	 //initialise allegro
	 allegro_init();
    
    // where to find config file?
    set_config_file("example.cfg");
    
    // set gfx mode , install keyboard, timer,mouse, windowing system

    int width,height,depth;
    width = get_config_int(NULL,"gfx_width",800);
    height = get_config_int(NULL,"gfx_height",600);
    depth = get_config_int(NULL,"gfx_depth",8);
    
	 install_s_wham(GFX_AUTODETECT, width,height,1000,800,depth);
    
    // add my_menu to the startbutton menu
	 s_wham_set_user_menu(my_menu);

	 install_int(user_timer, 33);

	 PALETTE pal;
	 generate_332_palette(pal);
    
	 pal[0].r = 0;
	 pal[0].g = 0;
	 pal[0].b = 0;

    // these will be displayed in the 'about' box accessible from the start menu
    s_wham_about_string = "S_Wham example, august 1998";
    s_wham_about_bitmap = create_bitmap(133,100);

	 fade_in(pal,2);       // set palette
    s_wham_restore_colors();

	 init_planets(600,400);



    // create window for the planets
	 S_Window *the_window = new S_Window("example",600,400);     // create a window
    
	 // create example dialog
	 S_Dialog *the_dialog = new S_Dialog("dialog", dialog,600,400);     // create a window
    // let dialog clean up itself when it's done
    the_dialog->delete_when_done = TRUE;
    
    // show planets window
	 the_window->show();                                 // show the window on the screen
    // start example dialog
	 the_dialog->start();
    
    // show example dialog
	 the_dialog->show();                                 // show the window on the screen
    

	 // get one frame of planets for about_bitmap
    draw_planets_frame(the_window->bmp());
    stretch_blit(the_window->bmp(), s_wham_about_bitmap,0,0,600,400,0,0,133,100);

    // main loop
    Utimer = 0;

	 while (!key[KEY_ESC] && !exit_flag)
	 {
    
		Wham_Secretary.process_mouse();
      
      if (Utimer > 0)
      {
         Utimer = 0;
			draw_planets_frame(the_window->bmp());
			the_window->update();
      }
      
	 }


	 cleanup_planets();

	 return 0;
}
END_OF_MAIN();

typedef struct PLANET
{
		BITMAP *sprite;
		int x,y;
		double vx, vy;
		int mass;
} PLANET;

double G;
PLANET *planet;

void init_planets(int size_x, int size_y)
{

	planet = (PLANET *)malloc(NUMBER_OF_PLANETS * sizeof(PLANET));


	// init
	BITMAP *sprite = create_bitmap(17,17);
	clear_to_color(sprite,0);
	for (int i=0; i<8;i++)
	{
		circlefill(sprite,8,8,8-i,makecol(255 * i/8,0,140 - 20 * i ));
	}
	G = GRAV;

	for (int i =0;i<NUMBER_OF_PLANETS;i++)
	{
		planet[i].sprite= sprite;
		planet[i].x= random() % size_x;
		planet[i].y= random() % size_y;
		planet[i].vx= 0;
		planet[i].vy= 0;
		planet[i].mass= (random() % 500) + 50;
//    planet[i].mass= 1000;
		
	}


}

void draw_planets_frame(BITMAP *onto)
{
		for (int i =0;i<NUMBER_OF_PLANETS;i++)
		{
			for (int j =0;j<NUMBER_OF_PLANETS;j++)
			{
				if (i !=j )
				{  // calc
					{
						PLANET *p1,*p2;
						static int sign = 1;
						p1 =&(planet[i]);
						p2 =&(planet[j]);
					
						int xdif,ydif;
					
						xdif = p1->x - p2->x;
						ydif = p1->y - p2->y;
						unsigned int dist = xdif * xdif + ydif * ydif;
						if (!dist)
							continue;
						if (dist > THRESH)
							continue;
					
					//   dist = dist < radius ? radius : dist;
						sign = dist < radius ? -2 : 1;
						double force = sign * (G * (p1->mass + p2->mass)) / dist;
					
						fixed dx,dy,dz;
					
						dx = itofix(xdif);
						dy = itofix(ydif);
						dz = 0;
						normalize_vector(&dx,&dy,&dz);
						double ax = fixtof(dx) * force / p1->mass;
						double ay = (fixtof(dy) * force + PULL) / p1->mass;
					
					
						p1->vx +=ax;
						p1->vy +=ay;
					
						if (p1->vx > MAXSPEED)
							p1->vx = MAXSPEED;
						if (p1->vy > MAXSPEED)
							p1->vy = MAXSPEED;
						if (p1->vx < -MAXSPEED)
							p1->vx = -MAXSPEED;
						if (p1->vy < -MAXSPEED)
							p1->vy = -MAXSPEED;
					  
					}
				}
			}
		}
		
		for (int i =0;i<NUMBER_OF_PLANETS;i++)
		{
			planet[i].x+=(int)(planet[i].vx);
			planet[i].y+=(int)(planet[i].vy);
			if (planet[i].x > onto->w || planet[i].x < 0)
			{
				planet[i].vx = - planet[i].vx;
				planet[i].x+=(int)(2*planet[i].vx);
			}
			if (planet[i].y > onto->h || planet[i].y < 0)
			{
				planet[i].vy = - planet[i].vy;
				planet[i].y+=(int)(2*planet[i].vy);
			}
	
		
			planet[i].vx *= FRICT;
			planet[i].vy *= FRICT;
		}

	
	
		clear_to_color(onto,0);
	
		for (int i =0;i<NUMBER_OF_PLANETS;i++)
		{
			draw_sprite(onto,planet[i].sprite,planet[i].x-8,planet[i].y-8);
		}
}



void cleanup_planets()
{
	free(planet);
//   ttrace("%d",ret);
}

