#include "s_task.h"
#include "s_border.h"
#include "s_wham.h"
#include "s_dialog.h"
#include <panel.h>

void s_callback(void *dp3, int d2);


static S_Dialog *about_box = 0;
static S_Dialog *sysinfo_box = 0;
static DIALOG *current_taskbar_dialog;
static int doing_menu = FALSE;
static char gfx_specs[80];
static char gfx_specs2[80];
static char gfx_specs3[80];
static char cpu_name[80];
static char sysinfo_string[1024];
static char help_string[] = "...Windowing Library Help...\n"
									 "Resizing windows:\n"
									 " Drag right or lower border with left mouse button .\n\n"
									 "Moving viewport:\n"
									 " Drag scrollbar with left mouse button or drag window contents with right mouse button.\n\n"
									 "Moving Window:\n"
									 " Drag Upper or Left border.\n\n"
									 "Roll up (minimise) Window:\n"
									 " Click first border button (bar).\n\n"
									 "Hide Window:\n"
									 " Click second border button (bar).\n\n"
									 "Maximise Window:\n"
									 " Click third border button (bar).\n\n"
                            "Display Window:\n"
									 " Click button on taskbar.\n\n"
									 "Raise window:\n"
									 " Click in window, or click on taskbar button.\n\n"
									 "\n"
                            ;
static unsigned int pack_color(int col)
{
   int r ,g,b;
   r = getr(col);
   g = getg(col);
   b = getb(col);
   return (r << 16) + (g << 8) + b;
}


static void calc_gui_mg_color()
{
	int r = getr(gui_fg_color) + getr(gui_bg_color)*2;
	int g = getg(gui_fg_color) + getg(gui_bg_color)*2;
	int b = getb(gui_fg_color) + getb(gui_bg_color)*2;
   r /=3;
   g /=3;
   b /=3;
   gui_mg_color = makecol(r,g,b);
   
}

static DIALOG system_info[]=
{
   /* (dialog proc)     (x)   (y)   (w)   (h)   (fg)  (bg)  (key) (flags)  (d1)  (d2)  (dp) */
   { d_textbox_proc,   10,   10,  380,	  70,	 255,    0,    0,    0,       0,    0,    sysinfo_string},
   { d_textbox_proc,   10,   90,  380,	  70,	 255,    0,    0,    0,       0,    0,    help_string},
   { d_button_proc,    10,  170,  380,   15,  255,  	3,    0,    D_EXIT,  0,    0,    const_cast<char *>("OK")},
   { NULL,              0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL }
};




static void get_sysinfo()
{

   AlGuiContainer::push_gui_screen();
   sprintf(gfx_specs, "%dx%d (%dx%d), %ldk vram", 
	   SCREEN_W, SCREEN_H, VIRTUAL_W, VIRTUAL_H, gfx_driver->vid_mem/1024);
   AlGuiContainer::pop_gui_screen();

   if (!gfx_driver->linear)
      sprintf(gfx_specs+strlen(gfx_specs), ", %ldk banks, %ldk granularity",
	      gfx_driver->bank_size/1024, gfx_driver->bank_gran/1024);

   switch (bitmap_color_depth(WHAM->TheScreen)) {

      case 8:
	 strcpy(gfx_specs2, "8 bit (256 color)");
	 break;

      case 15:
	 strcpy(gfx_specs2, "15 bit (32K HiColor)");
	 break;

      case 16:
	 strcpy(gfx_specs2, "16 bit (64K HiColor)");
	 break;

      case 24:
	 strcpy(gfx_specs2, "24 bit (16M TrueColor)");
	 break;

      case 32:
	 strcpy(gfx_specs2, "32 bit (16M TrueColor)");
	 break;

      default:
	 strcpy(gfx_specs2, "Unknown color depth!");
	 break;
   }

   strcpy(gfx_specs3, "Capabilities: ");
   int c = 0;

   if (gfx_capabilities & GFX_CAN_SCROLL) {
      strcat(gfx_specs3, "scroll");
      c++;
   }

   if (gfx_capabilities & GFX_CAN_TRIPLE_BUFFER) {
      if (c)
	 strcat(gfx_specs3, ", ");
      strcat(gfx_specs3, "triple buffer");
      c++;
   }


   AlGuiContainer::push_gui_screen();
   mouse_off();
   show_mouse(WHAM->TheScreen);

   if (gfx_capabilities & GFX_HW_CURSOR) {
      if (c)
	 strcat(gfx_specs3, ", ");
      strcat(gfx_specs3, "hardware cursor");
      c++;
   }

   show_mouse(NULL);
   mouse_on();
   AlGuiContainer::pop_gui_screen();

   if (gfx_capabilities & ~(GFX_CAN_SCROLL | GFX_CAN_TRIPLE_BUFFER | GFX_HW_CURSOR)) {
      if (c)
	 strcat(gfx_specs3, ", ");
      strcat(gfx_specs3, "hardware acceleration");
      c++;
   }

   if (!c)
      strcat(gfx_specs3, "0");



   sprintf
   (
   	sysinfo_string,
      "...System Info...\n\n"
      "Library versions:\n"
      "%s\n"
      "%s\n"
      "\n"
      "Graphics System:\n"
      "%s\n"
      "%s\n"
      "%s\n"
      "%s\n"
      "%s\n\n",
      allegro_id,
      WHAM_VERSION_STRING,
      gfx_driver->name,
      gfx_driver->desc,
      gfx_specs,
      gfx_specs2,
      gfx_specs3
   );


   check_cpu();
   sprintf(sysinfo_string + strlen(sysinfo_string), "CPU:");
   if (cpu_cpuid)
   	strcat(sysinfo_string, "\n");
   else
   	strcat(sysinfo_string, "(guesswork)\n");
      
   switch(cpu_family)
   {
		case 3:
        	strcat(sysinfo_string, "80386 ");
      	switch(cpu_model)
         {
         	case 0:
            	strcat(sysinfo_string, "DX\n");
               break;
            case 2:
            	strcat(sysinfo_string, "SX\n");
               break;
            case 4:
            	strcat(sysinfo_string, "SL\n");
               break;
         	default:
            	strcat(sysinfo_string, "\n");
            	break;
         }
      	break;
      case 4:
        	strcat(sysinfo_string, "80486");
      	switch(cpu_model)
         {
         	case 0:
            case 1:
            	strcat(sysinfo_string, "DX\n");
               break;
            case 3:
            	strcat(sysinfo_string, "SX/DX\n");
               break;
            case 2:
            	strcat(sysinfo_string, "SX\n");
               break;
            case 4:
            	strcat(sysinfo_string, "SL\n");
               break;
            case 5:
            	strcat(sysinfo_string, "SX2\n");
               break;
            case 7:
            	strcat(sysinfo_string, "DX2 write back enhanced.\n");
               break;
            case 8:
            	strcat(sysinfo_string, "DX4\n");
               break;
            case 14:
            	strcat(sysinfo_string, "Cyrix\n");
               break;
         	default:
            
            	strcat(sysinfo_string, "\n");
            	break;
         }
      	break;
      case 5:
        	strcat(sysinfo_string, "Pentium ");
      	switch(cpu_model)
         {
				case 2:
            	strcat(sysinfo_string,"P54C\n");
            	break;
				case 3:
            	strcat(sysinfo_string,"Overdrive\n");
            	break;
				case 5:
            	strcat(sysinfo_string,"Overdrive for IntelDX4\n");
            	break;
				case 14:
            	strcat(sysinfo_string,"Cyrix\n");
            	break;
         	case 1:
         	default:
            	strcat(sysinfo_string,"\n");
            	break;
               
         }
      	break;
      case 6:
      	strcat(sysinfo_string, "Pentium Pro\n");
	     	break;
      default:
      	strcat(sysinfo_string, "Unknown cpu type");
         if (cpu_family > 6)
         	strcat(sysinfo_string , "( > Pentium Pro).\n");
         else
				strcat(sysinfo_string, "\n");
      	break;
   }
   strcat(sysinfo_string, cpu_vendor);
	strcat(sysinfo_string, "\n");
   if (cpu_fpu)
   	strcat(sysinfo_string, "FPU present\n");
   else
   	strcat(sysinfo_string, "FPU absent\n");

	if (cpu_mmx)
   	strcat(sysinfo_string, "MMX: yes\n");
	else
   	strcat(sysinfo_string, "MMX: no\n");

 	strcat(sysinfo_string, "\n");

   strcat(sysinfo_string, "Operating System:\n");
   switch(os_type)
   {
   	default:
   	case OSTYPE_UNKNOWN:
      	strcat(sysinfo_string, "MSDOS (?)\n");
         break;
   	case OSTYPE_WIN3:
      	sprintf(sysinfo_string + strlen(sysinfo_string), "Windows 3.X\n");
         break;
   	case OSTYPE_WIN95:
      	strcat(sysinfo_string, "Windows '95\n");
         break;
   	case OSTYPE_WINNT:
      	strcat(sysinfo_string, "Windows NT\n");
         break;
   	case OSTYPE_OS2:
      	strcat(sysinfo_string, "OS/2\n");
         break;
   	case OSTYPE_WARP:
      	strcat(sysinfo_string, "OS/2 Warp 3\n");
         break;
   	case OSTYPE_DOSEMU:
      	strcat(sysinfo_string, "Linux DOSEMU\n");
         break;
   	case OSTYPE_OPENDOS:
      	strcat(sysinfo_string, "Caldera openDOS\n");
         break;

   }


}

char const *__s_wham_get_system_info_string()
{
   get_sysinfo();
   return sysinfo_string;
}



S_Dialog *col_slider_window;

static DIALOG pickcolor[]=
{
   /* (dialog proc)     (x)   (y)   (w)   (h)   (fg)  (bg)  (key) (flags)  (d1)  (d2)  (dp) */
   { d_slider_proc,     1,   10,  200,   10,    0,    0,    0,    0,		255,	  0,    NULL,(void *)s_callback, (void *)col_slider_window},
   { d_slider_proc,     1,   30,  200,   10,    0,    0,    0,    0,		255,	  0,    NULL,(void *)s_callback, (void *)col_slider_window},
   { d_slider_proc,     1,   50,  200,   10,    0,    0,    0,    0,		255,	  0,    NULL,(void *)s_callback, (void *)col_slider_window},
   { d_ctext_proc,    100,   80,  200,   10,    0,    0,    0,    0,        0,    0,  const_cast<char *>("COLOR")},
   { d_button_proc,     1,  110,  200,   18,   255,  3,    0,    D_EXIT,  0,    0,    const_cast<char *>("OK") },
   { NULL,              0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL }
};



void s_callback(void *dp3, int d2)
{
	pickcolor[3].fg = makecol(pickcolor[0].d2,pickcolor[1].d2,pickcolor[2].d2);

   if (active_dialog == pickcolor)
   {
	   SEND_MESSAGE(pickcolor + 3 , MSG_DRAW, 0);
   }
   AlGuiContainer::push_gui_screen();
	((Wobject*)dp3)->show();
   AlGuiContainer::pop_gui_screen();
   
}


static int d_selectcol_proc(int msg, DIALOG *d, int c)
{

    int ret = d_button_proc(msg,d,c);
    

    if (msg == MSG_CLICK || msg == MSG_KEY)
    {
        int color;
        while(mouse_b);

        switch(d->d1)
        {
      	case 0:
				color = S_Borderclass::Border_Color;
         	break;
      	case 1:
         	color = WinSystem::BackgroundColor;
         	break;
      	case 2:
				color = S_Borderclass::Title_Color;
         	break;
      	case 3:
				color = gui_fg_color;
         	break;
      	case 4:
				color = gui_bg_color;
         	break;
         default:
            color = 0;
            break;
        }
        pickcolor[0].d2 = getr(color);
        pickcolor[1].d2 = getg(color);
        pickcolor[2].d2 = getb(color);
        set_dialog_color(pickcolor, gui_fg_color, gui_bg_color);
        col_slider_window = new S_Dialog("Pick a Color",pickcolor,202,132,WHAM->get_TheScreen_size_x() / 2 - 101, WHAM->get_TheScreen_size_y()/2 -66,  TRUE);
        pickcolor[0].dp3 = pickcolor[1].dp3 = pickcolor[2].dp3 = col_slider_window;

      AlGuiContainer::push_gui_screen();
        col_slider_window->start();
        col_slider_window->show();
      AlGuiContainer::pop_gui_screen();

        d->flags &= ~D_SELECTED;
        ret = SEND_MESSAGE(d,MSG_DRAW,c);

      // wait here until it exits
        while(col_slider_window->is_running())
      	 Wham_Secretary.process_mouse();
		delete col_slider_window;

        color = makecol(pickcolor[0].d2,pickcolor[1].d2,pickcolor[2].d2);
      
        switch(d->d1)
        {
        	case 0:
		  		Panel::Background_Color = S_Borderclass::Border_Color = color;
		  		Panel::Light_Outline_Color = S_Borderclass::Outline_Color1 = makecol(MIN(255,pickcolor[0].d2 *2),MIN(255,pickcolor[1].d2 *2),MIN(255,pickcolor[2].d2 *2));
		  		Panel::Light_Outline_Color = S_Borderclass::Outline_Color2 = makecol(pickcolor[0].d2/2,pickcolor[1].d2/2,pickcolor[2].d2/2);
         	break;
      	case 1:
            WHAM->set_background_color(color);
            if (s_wham_root_window)
	            s_wham_root_window->set(NULL, color);
         	break;
      	case 2:
				S_Borderclass::Title_Color = color;
         	break;
      	case 3:
				gui_fg_color = color;
         	break;
      	case 4:
				gui_bg_color = color;
         	break;
         default:
            break;
        }
        set_config_hex("S_WHAMWINDOWS","BorderColor", pack_color(S_Borderclass::Border_Color));
        set_config_hex("S_WHAMWINDOWS","OutlineColor1", pack_color(S_Borderclass::Outline_Color1));
        set_config_hex("S_WHAMWINDOWS","OutlineColor2", pack_color(S_Borderclass::Outline_Color2));
        set_config_hex("S_WHAMWINDOWS","BackgroundColor",pack_color(WinSystem::BackgroundColor));
        set_config_hex("S_WHAMWINDOWS","TitleColor",pack_color(S_Borderclass::Title_Color));
        set_config_hex("S_WHAMWINDOWS","FgColor",pack_color(gui_fg_color));
        set_config_hex("S_WHAMWINDOWS","BgColor",pack_color(gui_bg_color));
		  ScrollBar::Light_Color = S_Borderclass::Outline_Color1;
   	  ScrollBar::Middle_Color = S_Borderclass::Border_Color;
   	  ScrollBar::Dark_Color = S_Borderclass::Outline_Color2;
		  calc_gui_mg_color();
        WHAM->refresh_all_visible_objects();
    }
    return ret;

}

static DIALOG colorsdialog[]=
{
   /* (dialog proc)     (x)   (y)   (w)   (h)   (fg)  (bg)  (key) (flags)  (d1)  (d2)  (dp) */
   { d_selectcol_proc,     1,    1,  200,  18,   255,  3,   0,  0,       0,    0,    const_cast<char *>("Border color") },
   { d_selectcol_proc,     1,   21,  200,  18,   255,  3,   0,  0,       1,    0,    const_cast<char *>("Desktop color") },
   { d_selectcol_proc,     1,   41,  200,  18,   255,  3,   0,  0,       2,    0,    const_cast<char *>("Title color") },
   { d_selectcol_proc,     1,   61,  200,  18,   255,  3,   0,  0,       3,    0,    const_cast<char *>("Dialog Foreground") },
   { d_selectcol_proc,     1,   81,  200,  18,   255,  3,   0,  0,       4,    0,    const_cast<char *>("Dialog Background") },
   { d_button_proc,     	1,  111,  200,  18,   255,  3,    0,    D_EXIT,  0,    0,    const_cast<char *>("OK") },
   { NULL,              0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL }
};







static int do_colors_menu()
{
    S_Dialog *colmenu;

    
 	  while(mouse_b);
     set_dialog_color(colorsdialog, gui_fg_color, gui_bg_color);
     colmenu = new S_Dialog("Colors Menu",colorsdialog,202,132,WHAM->get_TheScreen_size_x() / 2 - 101, WHAM->get_TheScreen_size_y() / 2 - 66, TRUE);
     colmenu->delete_when_done = TRUE; // this is needed because wel loose control when we leave this function
     AlGuiContainer::push_gui_screen();
     colmenu->start();
     colmenu->show();
     AlGuiContainer::pop_gui_screen();

    return 0;
}

static int closer_func(Wobject *w, void *param)
{
	w = NULL;
   param = NULL;
	return -1;
}

static int do_sysinfo()
{
    S_Dialog *menu;
    if (sysinfo_box)
    {
    	sysinfo_box->show();
    	return 0;
    }
	 while(mouse_b);
    get_sysinfo();
    set_dialog_color(system_info, gui_fg_color, gui_bg_color);
    menu = new S_Dialog("System info",system_info,400,190,(WHAM->get_TheScreen_size_x() - 400) /2, (WHAM->get_TheScreen_size_y() - 190)/2);
	 menu->raise_and_lock();
    if (WHAM->get_TheScreen_size_x() < 400 || WHAM->get_TheScreen_size_y() < 190)
    {
        		menu->resize_s_viewport(WHAM->get_TheScreen_size_x() - 30, WHAM->get_TheScreen_size_y() - 30);
            menu->move(15,15);
    }
    menu->add_close_button(closer_func,NULL);
    menu->delete_when_done = TRUE; // this is needed because wel loose control when we leave this function
    menu->start();

    AlGuiContainer::push_gui_screen();
    menu->show();
    AlGuiContainer::pop_gui_screen();

    sysinfo_box = menu;
   return 0;
}

static DIALOG about[20];


static int do_about()
{
    if (about_box)
    {
    	about_box->show();
    	return 0;
    }

    S_Dialog *menu;
    
    if (s_wham_about_string && s_wham_about_bitmap)
    {
        	about[0].proc = d_textbox_proc;
         about[0].x = 10;
         about[0].y = 2;
         about[0].w = 380;
         about[0].h = 21;
         about[0].flags = 0;
         about[0].d1 = about[0].d2 = 0;
         about[0].dp = s_wham_about_string;
        	about[1].proc = d_bitmap_proc;
         about[1].x = (400 - s_wham_about_bitmap->w)/2;
         about[1].y = 25;
         about[1].w = s_wham_about_bitmap->w;
         about[1].h = s_wham_about_bitmap->h;
         about[1].flags = 0;
         about[1].d1 = about[1].d2 = 0;
         about[1].dp = s_wham_about_bitmap;
        	about[2].proc = d_ctext_proc;
         about[2].x = 200;
         about[2].y = 30 + s_wham_about_bitmap->h;
         about[2].w = about[2].h = 0;
         about[2].flags = 0;
         about[2].d1 = about[2].d2 = 0;
         about[2].dp = const_cast<char *>("This program uses the Allegro game library:");
        	about[3].proc = d_ctext_proc;
         about[3].x = 200;
         about[3].y = 40 + s_wham_about_bitmap->h;
         about[3].w = about[3].h = 0;
         about[3].flags = 0;
         about[3].d1 = about[3].d2 = 0;
         about[3].dp = const_cast<char *>("(http://www.talula.demon.co.uk/allegro)");
        	about[4].proc = d_ctext_proc;
         about[4].x = 200;
         about[4].y = 50 + s_wham_about_bitmap->h;
         about[4].w = about[4].h = 0;
         about[4].flags = 0;
         about[4].d1 = about[4].d2 = 0;
         about[4].dp = const_cast<char *>("And the WHAM windowing library:");
        	about[5].proc = d_ctext_proc;
         about[5].x = 200;
         about[5].y = 60 + s_wham_about_bitmap->h;
         about[5].w = about[5].h = 0;
         about[5].flags = 0;
         about[5].d1 = about[5].d2 = 0;
         about[5].dp = const_cast<char *>("(http://home.hccnet.nl/m.versteegh)");
        	about[6].proc = d_ctext_proc;
         about[6].x = 200;
         about[6].y = 70 + s_wham_about_bitmap->h;
         about[6].w = about[6].h = 0;
         about[6].flags = 0;
         about[6].d1 = about[6].d2 = 0;
         about[6].dp = const_cast<char *>("And is probably compiled with DJGPP:");
        	about[7].proc = d_ctext_proc;
         about[7].x = 200;
         about[7].y = 80 + s_wham_about_bitmap->h;
         about[7].w = about[7].h = 0;
         about[7].flags = 0;
         about[7].d1 = about[7].d2 = 0;
         about[7].dp = const_cast<char *>("(http://www.delorie.com/djgpp/)");
        	about[8].proc = d_button_proc;
         about[8].x = 10;
         about[8].y = 175;
         about[8].w = 30;
         about[8].h = 12;
         about[8].flags = D_EXIT;
         about[8].d1 = about[8].d2 = 0;
         about[8].dp = const_cast<char *>("OK");
        	about[9].proc = NULL;

         
    }
    else if (s_wham_about_string)
    {
         about[0].proc = d_textbox_proc;
         about[0].x = 10;
         about[0].y = 2;
         about[0].w = 380;
         about[0].h = 50;
         about[0].flags = 0;
         about[0].d1 = about[0].d2 = 0;
         about[0].dp = s_wham_about_string;
        	about[1].proc = d_ctext_proc;
         about[1].x = 200;
         about[1].y = 60;
         about[1].w = about[1].h = 0;
         about[1].flags = 0;
         about[1].d1 = about[1].d2 = 0;
         about[1].dp = const_cast<char *>("This program uses the Allegro game library:");
        	about[2].proc = d_ctext_proc;
         about[2].x = 200;
         about[2].y = 70 ;
         about[2].w = about[2].h = 0;
         about[2].flags = 0;
         about[2].d1 = about[2].d2 = 0;
         about[2].dp = const_cast<char *>("(http://www.talula.demon.co.uk/allegro)");
        	about[3].proc = d_ctext_proc;
         about[3].x = 200;
         about[3].y = 80 ;
         about[3].w = about[3].h = 0;
         about[3].flags = 0;
         about[3].d1 = about[3].d2 = 0;
         about[3].dp = const_cast<char *>("And the WHAM windowing library:");
        	about[4].proc = d_ctext_proc;
         about[4].x = 200;
         about[4].y = 90 ;
         about[4].w = about[4].h = 0;
         about[4].flags = 0;
         about[4].d1 = about[4].d2 = 0;
         about[4].dp = const_cast<char *>("(http://home.hccnet.nl/m.versteegh");
        	about[5].proc = d_ctext_proc;
         about[5].x = 200;
         about[5].y = 100 ;
         about[5].w = about[5].h = 0;
         about[5].flags = 0;
         about[5].d1 = about[5].d2 = 0;
         about[5].dp = const_cast<char *>("And is probably compiled with DJGPP:");
        	about[6].proc = d_ctext_proc;
         about[6].x = 200;
         about[6].y = 110 ;
         about[6].w = about[6].h = 0;
         about[6].flags = 0;
         about[6].d1 = about[6].d2 = 0;
         about[6].dp = const_cast<char *>("(http://www.delorie.com/djgpp/)");
        	about[7].proc = d_button_proc;
         about[7].x = 10;
         about[7].y = 175 ;
         about[7].w = 30;
         about[7].h = 12;
         about[7].flags = D_EXIT;
         about[7].d1 = about[7].d2 = 0;
         about[7].dp = const_cast<char *>("OK");
        	about[8].proc = NULL;
    }
    else if (s_wham_about_bitmap)
    {
        	about[0].proc = d_ctext_proc;
         about[0].x = 200;
         about[0].y = 10;
         about[0].w = about[0].h = 0;
         about[0].flags = 0;
         about[0].d1 = about[0].d2 = 0;
         about[0].dp = const_cast<char *>("   ");
        	about[1].proc = d_bitmap_proc;
         about[1].x = (400 - s_wham_about_bitmap->w)/2;
         about[1].y = 20;
         about[1].w = s_wham_about_bitmap->w;
         about[1].h = s_wham_about_bitmap->h;
         about[1].flags = 0;
         about[1].d1 = about[1].d2 = 0;
         about[1].dp = s_wham_about_bitmap;
        	about[2].proc = d_ctext_proc;
         about[2].x = 200;
         about[2].y = 30 + s_wham_about_bitmap->h;
         about[2].w = about[2].h = 0;
         about[2].flags = 0;
         about[2].d1 = about[2].d2 = 0;
         about[2].dp = const_cast<char *>("This program uses the Allegro game library:");
        	about[3].proc = d_ctext_proc;
         about[3].x = 200;
         about[3].y = 30 + s_wham_about_bitmap->h;
         about[3].w = about[3].h = 0;
         about[3].flags = 0;
         about[3].d1 = about[3].d2 = 0;
         about[3].dp = const_cast<char *>("(http://www.talula.demon.co.uk/allegro)");
        	about[4].proc = d_ctext_proc;
         about[4].x = 200;
         about[4].y = 40 + s_wham_about_bitmap->h;
         about[4].w = about[4].h = 0;
         about[4].flags = 0;
         about[4].d1 = about[4].d2 = 0;
         about[4].dp = const_cast<char *>("And the WHAM windowing library:");
        	about[5].proc = d_ctext_proc;
         about[5].x = 200;
         about[5].y = 50 + s_wham_about_bitmap->h;
         about[5].w = about[5].h = 0;
         about[5].flags = 0;
         about[5].d1 = about[5].d2 = 0;
         about[5].dp = const_cast<char *>("(http://home.hccnet.nl/m.versteegh");
        	about[6].proc = d_ctext_proc;
         about[6].x = 200;
         about[6].y = 60 + s_wham_about_bitmap->h;
         about[6].w = about[6].h = 0;
         about[6].flags = 0;
         about[6].d1 = about[6].d2 = 0;
         about[6].dp = const_cast<char *>("And is probably compiled with DJGPP:");
        	about[7].proc = d_ctext_proc;
         about[7].x = 200;
         about[7].y = 70 + s_wham_about_bitmap->h;
         about[7].w = about[7].h = 0;
         about[7].flags = 0;
         about[7].d1 = about[7].d2 = 0;
         about[7].dp = const_cast<char *>("(http://www.delorie.com/djgpp/)");
        	about[8].proc = d_button_proc;
         about[8].x = 10;
         about[8].y = 175;
         about[8].w = 30;
         about[8].h = 12;
         about[8].flags = D_EXIT;
         about[8].d1 = about[8].d2 = 0;
         about[8].dp = const_cast<char *>("OK");
        	about[9].proc = NULL;
    }
    else
    {
        	about[0].proc = d_ctext_proc;
         about[0].x = 200;
         about[0].y = 10;
         about[0].w = about[0].h = 0;
         about[0].flags = 0;
         about[0].d1 = about[0].d2 = 0;
         about[0].dp = const_cast<char *>("   ");
        	about[1].proc = d_ctext_proc;
         about[1].x = 200;
         about[1].y = 30;
         about[1].w = about[1].h = 0;
         about[1].flags = 0;
         about[1].d1 = about[1].d2 = 0;
         about[1].dp = const_cast<char *>("This program uses the Allegro game library:");
        	about[2].proc = d_ctext_proc;
         about[2].x = 200;
         about[2].y = 40 ;
         about[2].w = about[2].h = 0;
         about[2].flags = 0;
         about[2].d1 = about[2].d2 = 0;
         about[2].dp = const_cast<char *>("(http://www.talula.demon.co.uk/allegro)");
        	about[3].proc = d_ctext_proc;
         about[3].x = 200;
         about[3].y = 50 ;
         about[3].w = about[3].h = 0;
         about[3].flags = 0;
         about[3].d1 = about[3].d2 = 0;
         about[3].dp = const_cast<char *>("And the WHAM windowing library:");
        	about[4].proc = d_ctext_proc;
         about[4].x = 200;
         about[4].y = 60 ;
         about[4].w = about[4].h = 0;
         about[4].flags = 0;
         about[4].d1 = about[4].d2 = 0;
         about[4].dp = const_cast<char *>("(http://home.hccnet.nl/m.versteegh)");
        	about[5].proc = d_ctext_proc;
         about[5].x = 200;
         about[5].y = 70 ;
         about[5].w = about[5].h = 0;
         about[5].flags = 0;
         about[5].d1 = about[5].d2 = 0;
         about[5].dp = const_cast<char *>("And is probably compiled with DJGPP:");
        	about[6].proc = d_ctext_proc;
         about[6].x = 200;
         about[6].y = 80 ;
         about[6].w = about[6].h = 0;
         about[6].flags = 0;
         about[6].d1 = about[6].d2 = 0;
         about[6].dp = const_cast<char *>("(http://www.delorie.com/djgpp/)");
        	about[7].proc = d_button_proc;
         about[7].x = 10;
         about[7].y = 175 ;
         about[7].w = 30;
         about[7].h = 12;
         about[7].flags = D_EXIT;
         about[7].d1 = about[7].d2 = 0;
         about[7].dp = const_cast<char *>("OK");
        	about[8].proc = NULL;
    }

    set_dialog_color(about, gui_fg_color, gui_bg_color);

    menu = new S_Dialog("About",about,400,190,(WHAM->get_TheScreen_size_x() - 400) /2, (WHAM->get_TheScreen_size_y() - 190)/2);
	 menu->raise_and_lock();
    if (WHAM->get_TheScreen_size_x() < 400 || WHAM->get_TheScreen_size_y() < 190)
    {
        		menu->resize_s_viewport(WHAM->get_TheScreen_size_x() - 30, WHAM->get_TheScreen_size_y() - 30);
            menu->move(15,15);
    }

    menu->start();
    AlGuiContainer::push_gui_screen();
    menu->show();
    AlGuiContainer::pop_gui_screen();
    menu->delete_when_done = TRUE;

  	 about_box = menu;
    
    return 0;
}

static S_Dialog *taskbar_settings_window;

static DIALOG taskbar_settings[]=
{
   /* (dialog proc)     (x)   (y)   (w)   (h)   (fg)  (bg)  (key) (flags)  (d1)  (d2)  (dp) */
   { d_check_proc,         1,    1,  200,  10,   255,  3,   0,  	0,       0,    0,    const_cast<char *>("Autohide Taskbar") },
   { d_text_proc,          1,   20,  200,  10,   255,  3,   0,  	0,       0,    0,    const_cast<char *>("Taskbar location") },
   { d_radio_proc,         1,   30,  200,   8,   255,  3,   0,D_SELECTED,  0,    0,    const_cast<char *>("Top") },
   { d_radio_proc,         1,   40,  200,   8,   255,  3,   0,  	0,       0,    0,    const_cast<char *>("Bottom") },
   { d_text_proc,          1,   50,  200,  10,   255,  3,   0,  	0,       0,    0,    const_cast<char *>("Windows raise...") },
   { d_radio_proc,         1,   60,  200,   8,   255,  3,   0,D_SELECTED,  1,    0,    const_cast<char *>("when clicked.") },
   { d_radio_proc,         1,   70,  200,   8,   255,  3,   0,  	0,       1,    0,    const_cast<char *>("when the mouse is over.") },
   { d_text_proc,          1,   85,  200,  10,   255,  3,   0,  	0,       0,    0,    const_cast<char *>("Taskbar slide speed") },
   { d_slider_proc,        1,	  95,	 200,  10,	 255,  3,   0,	 	0,     200,		0,		NULL},
   { d_button_proc,     	1,  111,  200,  18,   255,  3,   0,    D_EXIT,  0,    0,    const_cast<char *>("OK") },
   { NULL,              0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL }
};

static int do_settings_menu()
{


   if (_s_wham_task_bar->do_shift)
   	taskbar_settings[0].flags |= D_SELECTED;
	else
   	taskbar_settings[0].flags &= ~D_SELECTED;

   if (_s_wham_task_bar->location)
   {
   	taskbar_settings[3].flags |= D_SELECTED;
   	taskbar_settings[2].flags &= ~D_SELECTED;
   }
   else
   {
   	taskbar_settings[3].flags &= ~D_SELECTED;
   	taskbar_settings[2].flags |= D_SELECTED;
   }


   if (whamgui_raise_style == RAISE_IMMEDIATE)
   {
   	taskbar_settings[6].flags |= D_SELECTED;
   	taskbar_settings[5].flags &= ~D_SELECTED;
   }
   else
   {
   	taskbar_settings[5].flags |= D_SELECTED;
   	taskbar_settings[6].flags &= ~D_SELECTED;
   }

   taskbar_settings[8].d2 = 20*(11 - _s_wham_task_bar->shift_slowness);

 	while(mouse_b);
   
   set_dialog_color(taskbar_settings, gui_fg_color, gui_bg_color);
   
   taskbar_settings_window = new S_Dialog("Taskbar settings",taskbar_settings,202,132,WHAM->get_TheScreen_size_x() / 2 - 101, WHAM->get_TheScreen_size_y() / 2 - 66, TRUE);
//   taskbar_settings[8].dp3 = taskbar_settings_window;
   
//   taskbar_settings_window->delete_when_done = TRUE; // this is needed because wel loose control when we leave this function
   AlGuiContainer::push_gui_screen();
   taskbar_settings_window->show();
   taskbar_settings_window->start();
   AlGuiContainer::pop_gui_screen();


   while(taskbar_settings_window->is_running())
   	Wham_Secretary.process_mouse();

   delete taskbar_settings_window;


   _s_wham_task_bar->do_shift = taskbar_settings[0].flags & D_SELECTED;

	_s_wham_task_bar->location = taskbar_settings[3].flags & D_SELECTED;

   _s_wham_task_bar->shift_slowness = 11 - taskbar_settings[8].d2/20;

   if (!_s_wham_task_bar->do_shift)
		_s_wham_task_bar->shift_in();
   else
		_s_wham_task_bar->shift_out();

   whamgui_raise_style = (taskbar_settings[5].flags & D_SELECTED) ? RAISE_CLICK : RAISE_IMMEDIATE;

   set_config_int("S_WHAMWINDOWS","whamgui_raise_style",whamgui_raise_style);

   return 0;

}



static MENU base_sysmenu[]=
{
	// text	 proc				  child flags data
   {const_cast<char *>("Colors"), do_colors_menu, NULL, 0, NULL},
   {const_cast<char *>("Settings"), do_settings_menu, NULL, 0, NULL},
   {const_cast<char *>("System Help"), do_sysinfo, NULL, 0, NULL},
   {const_cast<char *>("About"), do_about, NULL, 0, NULL},
   {NULL,NULL, NULL, 0, NULL}
};

static MENU *sysmenu = base_sysmenu;

static int menu_size(MENU *m)
{
	int i =0;

   while(m[i].text)
   	i++;
      
   return i;
}

void s_wham_set_user_menu(MENU *user_menu)
{
	if (sysmenu != base_sysmenu)
   	free(sysmenu);
      
   sysmenu = (MENU *)malloc(sizeof(MENU) * (menu_size(base_sysmenu) + menu_size(user_menu)+2));
   for (int i=0;i<menu_size(user_menu);i++)
   {
   	sysmenu[i] = user_menu[i];
   }
   sysmenu[menu_size(user_menu)].text = "";
   sysmenu[menu_size(user_menu)].proc = NULL;
   sysmenu[menu_size(user_menu)].dp = NULL;
   sysmenu[menu_size(user_menu)].flags = 0;
   for (int i=0;i<menu_size(base_sysmenu);i++)
   {
      sysmenu[i + menu_size(user_menu) + 1] = base_sysmenu[i];
   }
   sysmenu[menu_size(user_menu) + menu_size(base_sysmenu)+1].text = NULL;

   
}


static int d_sysmenu_proc(int msg, DIALOG *d, int c)
{

    int ret = d_button_proc(msg,d,c);
	 int ypos;
    if (msg == MSG_CLICK || msg == MSG_KEY)
    {
	     while(mouse_b);
    	  ypos
        =
        _s_wham_task_bar->get_geoinfo()->pos_y > WHAM->get_TheScreen_size_y() / 2
        ?
        _s_wham_task_bar->get_geoinfo()->pos_y - 8
        :
        _s_wham_task_bar->get_geoinfo()->pos_y + _s_wham_task_bar->get_geoinfo()->screen_size_y + 8
        ;
        _s_wham_task_bar->update(); // to show the button flip
        doing_menu = TRUE;
        wham_do_menu(sysmenu,_s_wham_task_bar->get_geoinfo()->pos_x,ypos);
        doing_menu = FALSE;
        // d is now no longer valid, because
        // the menu window opened & closed again
        // causing the taskbar dialog to be changed
        // so re-assign d

        d = current_taskbar_dialog; // button is first object of dialog
        
        d->flags &= ~D_SELECTED;
        ret = SEND_MESSAGE(d,MSG_DRAW,c);
    }
    return ret;
}

static char buttontext[100];

static void print_buttontext(char *t, int maxlen)
{
   unsigned int i = 0;
	for (i=0;t[i];i++)
   {
   	buttontext[i] = t[i];
   	buttontext[i+1] = 0;
      if (text_length(_s_wham_font,buttontext) >= maxlen)
      	break;
   }

   buttontext[i] = 0;
   if (i > 5 && i < strlen(t))
   	buttontext[i-1] = buttontext[i-2] = buttontext[i-3] = '.';
   
}

static int d_objbutton_proc(int msg, DIALOG *d, int c)
{

    if (msg == MSG_DRAW)
    	print_buttontext((char *)d->dp, d->w);


    char *origdp = (char *)d->dp;
    d->dp = buttontext;
    int ret = d_button_proc(msg,d,c);
    d->dp = origdp;
    int fx,fy,fw,fh;
    int x,y,w,h;
    int tx,ty,tw,th;
    if (msg == MSG_CLICK || msg == MSG_KEY)
    {
        AlGuiContainer::push_gui_screen();
        _s_wham_task_bar->update(); // to show the button flip
	     while(mouse_b);
        Wobject *wobj = (Wobject *)(d->dp2);
        wobj->raise();
        if (!wobj->get_geoinfo()->is_displayed)
        {
            xor_mode(TRUE);
            fx = d->x + _s_wham_task_bar->get_geoinfo()->pos_x;
            fy = d->y + _s_wham_task_bar->get_geoinfo()->pos_y;
            fw = d->w;
            fh = d->h;
            tx = (WHAM->get_TheScreen_size_x() - wobj->get_geoinfo()->screen_size_x)/2;
            ty = (WHAM->get_TheScreen_size_y() - wobj->get_geoinfo()->screen_size_y)/2;
            ty = MAX(ty,14);
            tw = wobj->get_geoinfo()->screen_size_x;
            th = wobj->get_geoinfo()->screen_size_y;
            for (int i=0;i<20;i++)
            {
             	x = (tx * i+ fx * (20 -i))/20;
             	y = (ty * i+ fy * (20 -i))/20;
             	w = (tw * i+ fw * (20 -i))/20;
             	h = (th * i+ fh * (20 -i))/20;
               mouse_off();
               rect(WHAM->TheScreen,x,y,x+w-1,y+h-1,s_wham_xor_color);
               mouse_on();
               rest(5);
               mouse_off();
               rect(WHAM->TheScreen,x,y,x+w-1,y+h-1,s_wham_xor_color);
               mouse_on();
               

            }
            solid_mode();

        
            wobj->move(	tx, ty);
            wobj->show();
        }
        AlGuiContainer::pop_gui_screen();

        d->flags &= ~D_SELECTED;
        ret = SEND_MESSAGE(d,MSG_DRAW,c);
    }
    return ret;
}

void S_TaskBar::freshen()
{
   DIALOG *dialog = (DIALOG *)malloc(sizeof(DIALOG) * (nr+2));

   dialog[0].proc = d_sysmenu_proc;
   dialog[0].x = 5;
   dialog[0].y = 3;
   dialog[0].w = 45;
   dialog[0].h = 14;
   dialog[0].fg = gui_fg_color;
   dialog[0].bg = gui_bg_color;
   dialog[0].key = 0;
   dialog[0].flags = 0;
   dialog[0].d1 = 0;
   dialog[0].d2 = 0;
   dialog[0].dp = const_cast<char *>("*");
   
   for (int i=1;i<nr+1;i++)
   {
      dialog[i].proc = d_objbutton_proc;
      dialog[i].x = (get_geoinfo()->size_x - 60) * (i - 1)/ nr + 55;
      dialog[i].y = 3;
      dialog[i].w = (get_geoinfo()->size_x - 60) / nr - 4;
      dialog[i].h = 14;
      dialog[i].fg = gui_fg_color;
      dialog[i].bg = gui_bg_color;
      dialog[i].key = objects[i-1]->get_geoinfo()->name[0];
      dialog[i].flags = 0;
      dialog[i].d1 = 0;
      dialog[i].d2 = 0;
      dialog[i].dp =  objects[i-1]->get_geoinfo()->name;
      dialog[i].dp2 = objects[i-1];
   }
   dialog[nr+1].proc = NULL;
   
   clear_to_color(bmp(),gui_bg_color);
   DIALOG *old = change_dialog(dialog);
   start();
   free(old);

   current_taskbar_dialog = dialog;
}



Wobject *S_TaskBar::process_mouse(int *x, int *y)
{
   if (!is_running()) // have we stopped (esc pressed for instance, just restart)
      start();



   
   if (get_geoinfo()->pos_y < WHAM->get_TheScreen_size_y() /2)
   	location = FALSE;
   else
   	location = TRUE;
   
   
   return AlGuiContainer::process_mouse(x,y);
   
}

static int shift_pos=0;

static void shifter()
{
	shift_pos++;
}

void S_TaskBar::shift_in()
{
  int s = 0;
  
  install_int(shifter,shift_slowness);

  shift_pos = 0;

  if (location)
  {
   if (get_geoinfo()->pos_y < WHAM->get_TheScreen_size_y())
   	return;
   while(s < 22)
   {
   	s = shift_pos;
      move(2, WHAM->get_TheScreen_size_y()  - MIN(s,21));
   }
  }
  else
  {
   if (get_geoinfo()->pos_y > 0)
   	return;
   while(s<22)
   {
   	s = shift_pos;
      move(2,MIN(21,s)-20);
   }
  }
  remove_int(shifter);
}


void S_TaskBar::shift_out()
{
	if (!do_shift || doing_menu)
   	return;

  int s = 0;
  
  install_int(shifter,shift_slowness);

  shift_pos = 0;

  if (location)
  {
   if (get_geoinfo()->pos_y == WHAM->get_TheScreen_size_y())
   	return;
   while(s<22)
   {
   	s = shift_pos;
      move(2, WHAM->get_TheScreen_size_y() - 21 + MIN(21,s));
   }
  }
  else
  {
   if (get_geoinfo()->pos_y == 0)
   	return;
	while(s<22)
   {
   	s = shift_pos;
      move(2, 1 - MIN(s,21));
   }

  }

  remove_int(shifter);
}


S_TaskBar::S_TaskBar()
:
   AlGuiContainer(WHAM->get_TheScreen_size_x()-4, 20, 2, WHAM->get_TheScreen_size_y(), "TaskBar", NULL, S_Borderclass::draw_box)
{
   objects = NULL;
   objects_to_add = NULL;
   objects_to_del = NULL;
   maxnr = nr = nr_to_add = maxnr_to_add = nr_to_del = maxnr_to_del = 0;
   clear_to_color(bmp(),gui_bg_color);
   set_lock_scale(TRUE);
   freshen();
   raise_and_lock(); //
	set_lock_scale(TRUE);
	set_lock_viewport_size(TRUE);
   set_lock_shift(TRUE);
   location = get_config_int("S_TASKBAR","location",0);
   do_shift = get_config_int("S_TASKBAR","do_shift",TRUE);
   shift_slowness = get_config_int("S_TASKBAR","shift_slowness",10);
   shift_slowness = MAX(1,shift_slowness);
   if (!location)
   	move(2, -20);
	calc_gui_mg_color();
}

S_TaskBar::~S_TaskBar()
{
	if (objects)
		free(objects);
   set_config_int("S_TASKBAR","shift_slowness",shift_slowness);
   set_config_int("S_TASKBAR","do_shift",do_shift);
   set_config_int("S_TASKBAR","location",location);
   
}

void S_TaskBar::add_obj(Wobject *obj)
{
	// check if not already in
   // check addlist
   for (int i=0;i<nr_to_add;i++)
   	if (obj == objects_to_add[i])
      	return;
	// check current content
   for (int i=0;i<nr;i++)
   	if (obj == objects[i])
      	return;

   nr_to_add++;
   if (nr_to_add > maxnr_to_add)
   {
   	maxnr_to_add += 50;
   	objects_to_add = (Wobject **)realloc(objects_to_add, sizeof(Wobject *) * maxnr_to_add);
		if (!objects_to_add)
      	out_of_mem("taskbar administration");
   }
   
   objects_to_add[nr_to_add -1] = obj;
//   trace("add to addlist %p %s",obj, obj->get_geoinfo()->name);
   WHAM->send_delayed_message(this, MSG_UPDATE_LISTS, this);
}

void S_TaskBar::update_lists()
{

   if (in_proc) // can't do it now, delay until we're out
   {
   	WHAM->send_delayed_message(this, MSG_UPDATE_LISTS, this);
      return;
   }

//   trace("update lists %d %d",nr_to_add,nr_to_del);
   // check lists for overlap, and remove overlapping calls
	for (int i=0;i<nr_to_add;i++)
   {
   	for (int j=0;j<nr_to_del;j++)
      {
      	if (objects_to_add[i] == objects_to_del[j])
         {

//   			trace("remove %p %s",objects_to_del[j], objects_to_del[j]->get_geoinfo()->name);
         	for (int k=i;k<nr_to_add-1;k++)
            	objects_to_add[k] = objects_to_add[k+1];
               
         	for (int k=j;k<nr_to_del-1;k++)
            	objects_to_del[k] = objects_to_del[k+1];
         	nr_to_add--;
         	nr_to_del--;
         }
      }
   }


	// do delete list
   for (int i=0;i<nr_to_del;i++)
   {
//   	trace("del %p %s",objects_to_del[i], objects_to_del[i]->get_geoinfo()->name);
   	int found = 0;
   	for (int j=0;j<nr;j++)
   	{
      	if (objects[j] == objects_to_del[i])
      	{
         	found = -1;
         	for (int k=j;k<nr;k++)
            	objects[k] = objects[k+1];
            
         	break;
      	}
   	}
   	if (found)
   	{
      	nr--;
   	}
   }

   // do add list
   for (int i=0;i<nr_to_add;i++)
   {
      nr++;
   	if (nr > maxnr)
   	{
   		maxnr += 50;
   		objects = (Wobject **)realloc(objects, sizeof(Wobject *) * maxnr);
   		if (!objects)
         	out_of_mem("taskbar administration");
   	}
//   	trace("add %p %s",objects_to_add[i], objects_to_add[i]->get_geoinfo()->name);
		objects[nr - 1] = objects_to_add[i];
   }

   if (nr_to_add || nr_to_del)
   {
//   	trace("going to FRESHEN in proc %d", in_proc);
	   freshen();
   }
   nr_to_add = nr_to_del = 0;
   

}

void S_TaskBar::delete_obj(Wobject *obj)
{

   nr_to_del++;
   if (nr_to_del > maxnr_to_del)
   {
   	maxnr_to_del += 50;
   	objects_to_del = (Wobject **)realloc(objects_to_del, sizeof(Wobject *) * maxnr_to_del);
	   if (!objects_to_del)
        	out_of_mem("taskbar administration");
   	      
   }
   objects_to_del[nr_to_del -1] = obj;
//   trace("add to dellist %p %s",obj, obj->get_geoinfo()->name);

   WHAM->send_delayed_message(this, MSG_UPDATE_LISTS, this);

}


int S_TaskBar::process_message(Wobject *from, int msg)
{
   if (from == this)
		AlGuiContainer::process_message(from,msg);

      
   switch(msg)
   {
      case W_MSG_GOTMOUSE:
         shift_in();
         break;
      case W_MSG_LOSTMOUSE:
         shift_out();
         break;
      case W_MSG_CREATED:
         if (from->magic_number == S_WHAM_MAGIC)
         {
//         	trace("got msg created from %s",from->get_geoinfo()->name);
            add_obj(from);
         }
         break;
      case MSG_UPDATE_LISTS:
      	update_lists();
         break;
      case W_MSG_DESTROYED:
         if (from->magic_number == S_WHAM_MAGIC)
         	delete_obj(from);

         if (from == about_box)
         	about_box = NULL;
         if (from == sysinfo_box)
         	sysinfo_box = NULL;
         break;
      default:
         break;
   }
   return AlGuiContainer::process_message(from,msg);
}

void S_TaskBar::get_coordinates_of_button(Wobject const *obj, int *x, int *y, int *w, int *h) const
{
	DIALOG const *d = get_dialog();
   for (int i=0;i<nr;i++)
   {
      if (objects[i] == obj)
      {
 		  *x = d[i+1].x;
 		  *y = d[i+1].y;
 		  *w = d[i+1].w;
 		  *h = d[i+1].h;
        return;
      }
   }
}




S_TaskBar *_s_wham_task_bar = NULL;

