/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <strings.h>
#include <winsys.h>
#include <helpers.h>
#include "file.h"


TXfile::TXfile(char *name)
{
    FILE *infile;
    if (!(infile = fopen(name,"r")))
        fatal("unable to open file %s",name);

    number_of_lines = 0;
    the_file = NULL;
    number_of_keys = 0;
    Toc = NULL;
    int chapterflag = FALSE;
    attribute = NULL;

    while(fgets(buffer,2045,infile))
    {
        if (buffer[0] != '@' || (buffer[0] == '@' && buffer[1] == '@')||(buffer[0] == '@' && buffer[1] == '\\'))
        {
            if (buffer[1] == '\\') buffer[1] = '@';
            number_of_lines++;
            the_file =(char **)realloc(the_file, number_of_lines * sizeof(char *));
            attribute=(int *)realloc(attribute, number_of_lines * sizeof(int));
            strcat(buffer,"  ");
            the_file[number_of_lines -1] = strdup(buffer);
            attribute[number_of_lines -1] = 0;
            if (chapterflag)
            {
                chapterflag = FALSE;
                attribute[number_of_lines -1] |= ATTRIB_CHAPTER;
            }
        }
        else
        {
            if (!strncmp(buffer,"@heading",8))
            {
                chapterflag = TRUE;
            }
        }
    }

    fclose(infile);
    build_toc();
}

TXfile::~TXfile()
{
    for (int i=0;i<number_of_lines;i++)
    {
        free(the_file[i]);
    }
    free(the_file);
    for (int i=0;i<number_of_keys;i++)
    {
        free(Toc[i].key);
    }
    free(Toc);
}

int qsort_helper_function(void const *a, void const *b)
{
    return strcasecmp(((Toc_entry *)a)->key,((Toc_entry *)b)->key);
}


void TXfile::build_toc()
{
    for(int i = 0;i<number_of_lines;i++)
    {
        if (the_file[i][0] == '@')   // only the @ is enough for all other lines starting with @ are filtered out
        {
            strcpy(buffer,the_file[i] + 2);
            for (char *s = buffer; *s; s++)
            {

                if (*s == '@')
                {
                    find_end_of_word(s + 1);
                    add_to_toc(s + 1,i);
                    attribute[i] |= ATTRIB_KEYWORD;
                    break;
                }
            }
        }
    }
    qsort(Toc,number_of_keys,sizeof(Toc_entry),qsort_helper_function);
}

void TXfile::find_end_of_word(char *s)
{
    while(s && (isalnum(*s) || *s == '_' ))
        s++;
    *s = 0;
}

void TXfile::add_to_toc(char *s, int i)
{
    number_of_keys++;
    Toc = (Toc_entry *)realloc(Toc,number_of_keys * sizeof(Toc_entry));
    Toc[number_of_keys - 1].key = strdup(s);
    Toc[number_of_keys - 1].line = i;
}


char *TXfile::get_line(int nr)
{
    buffer[0] = 0;
    char *t = buffer;
    if (nr < number_of_lines && nr >= 0)
    {
        for (char *s = the_file[nr];*s ; s++)
        {
            while (*s == '@' && *s) //filter out @'s
                s++;
            
            while (*s == 13 && *s)   // filter out CR's
                s++;
            
            while (*s == 10 && *s)   // filter out LF's
                s++;

            if (*s=='&' && *(s+1) == 'l' && *(s+2) == 't')
            {
                *t++='<';
                s+=2;
                continue;
            }
            if (*s=='&' && *(s+1) == 'g' && *(s+2) == 't')
            {
                *t++='>';
                s+=2;
                continue;
            }

            if (*s == '<')
            {
                while(*s && (*s != '>'))
                    s++;

                s++;  // one extra

            }
            *t++ = *s;
        }
        *t = 0;
    }
    return buffer;
}

int TXfile::get_toc_line(char *key)
{
    unsigned len = strlen(key);
    for (int i=0;i<number_of_keys;i++)
    {
        if (!(strncasecmp(key,Toc[i].key,len)))
        {
            return Toc[i].line;
        }

    }
    return 0;
}

int TXfile::get_toc_line(int nr)
{
    if (nr < number_of_keys)
        return Toc[nr].line;
    else
        return 0;
}



int TXfile::get_number_of_toc_entries()
{
    return number_of_keys;
}

char  *TXfile::get_toc_entry(int nr)
{
    if (nr < number_of_keys)
        return Toc[nr].key;
    else 
        return NULL;
}



int TXfile::get_attrib(int nr)
{
    if (nr >= 0 && nr < number_of_lines)
        return attribute[nr];

    return 0;
}


