/****************************************************************************
 * WHAM windowing system for Allegro & DJGPP                                *
 * Copyright 1997 Martijn Versteegh                                         *
 * WHAM is swapware                                                         *
 * that means if you use it you must send me something in return            *
 * If you use WHAM in a commercial product, 'something' should preferrable  *
 * be money. A program you made, a postcard or something like that will do  *
 * in most other cases.
 * I can be reached at                                                      *
 *  m.versteegh@cpedu.rug.nl                                                *
 *    or , if the above is expired                                          *
 *  vsteegh@dds.nl (my parents)                                             *
 *                                                                          *
 *  postbox 1443 9701BK Groningen, The Netherlands.                         *
 *                                                                          *
 *  Disclaimer:                                                             *
 *      I am not responsible for any effect this software may have on you,  *
 *      your properties or anything else connected to you.                  *
 *      Use it at your own risk.                                            *
 ****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <allegro.h>
#define WOBJECT 1
#define WINDOW  2

int main(int argc, char **argv)
{
    int Base = WOBJECT;

    if ((argc < 3) || (argc > 4))
    {
        printf("Usage:\n"
               "    WCLASS <class name> <file name> [w]"
               " creates a bare object derived from Wobject\n"
               " which can be atached to a winsys manager\n"
               " <class name> = the name of your object\n"
               " <file name> = name of the files to save in\n"
               "   (no extension)\n produces:\n"
               " <file name>.h  : header, include in your program header\n"
               " <filename>.cc  : source, include in your makefile or project\n"
               " edit .h and .cc files to add specifiec properties\n"
               " If you specify 'w' as third parameter the class will be \n"
               " derived from Window instead of Wobject\n"
               );
        exit(1);
    }
    if (argc == 4) 
    {
        if (argv[3][0] != 'w')
        {
            printf("Unknown bass class specifier %s\n",argv[3]);
            exit(1);
        }
        else
            Base = WINDOW;
    }

    char 
        headername[512],
        sourcename[512];

    strcpy(headername, argv[2]);
    strcpy(sourcename, argv[2]);
    strcat(headername, ".h");
    strcat(sourcename, ".cc");

    if (exists(headername))
    {
        printf("%s already exists",headername);
        exit(1);
    }

    if (exists(sourcename))
    {
        printf("%s already exists",sourcename);
        exit(1);
    }
    char buffer[1024];
    
    FILE *outfile;
    if (!(outfile = fopen(headername,"w")))
    {
        printf("unable to open file %s",headername);
        exit(1);
    }
    sprintf(buffer, "_Wham_%s_h_" , argv[2]);
    fprintf(outfile, "#ifndef %s\n", buffer);
    fprintf(outfile, "#define %s\n", buffer);
    fprintf(outfile, "\n");
    fprintf(outfile, "#include <helpers.h>");
    fprintf(outfile, "\n");
    
    if (Base == WOBJECT)
    {
        fprintf(outfile, "#include <wobject.h>\n");
    }
    else if (Base == WINDOW)
    {
        fprintf(outfile, "#include <window.h>\n");
    }
    fprintf(outfile, "\n");
    if (Base == WOBJECT)
    {
        fprintf(outfile, "class %s : public Wobject\n",argv[1]);
    }
    else if (Base == WINDOW)
    {
        fprintf(outfile, "class %s : public Window\n",argv[1]);
    }

    fprintf(outfile, "{\n");
    fprintf(outfile, "\n");
    fprintf(outfile, "  public:\n");
    fprintf(outfile, "\n");
    fprintf(outfile, "      // ADD CONSTRUCTORS & DESTRUCTORS IF NEEDED\n");
    fprintf(outfile, "      %s(int size_x,int size_y,int pos_x,int pos_y,char *name);\n",argv[1]);
    fprintf(outfile, "\n");
    fprintf(outfile, "      // this function handles the mouse for this class\n");
    fprintf(outfile, "      Wobject *process_mouse(int *x = NULL, int *y = NULL);\n", argv[1]);
    fprintf(outfile, "  private:\n");
    fprintf(outfile, "\n");
    if (Base == WOBJECT)
        fprintf(outfile, "      void draw(BITMAP *onto,int from_x, int from_y, int to_x, int to_y);\n");
    fprintf(outfile, "\n");
    fprintf(outfile, "      //IF BORDERS ARE NOT NEEDED REMOVE LINE BELOW, AND CHANGE CONSTRUCTOR\n");
    fprintf(outfile, "      static int draw_border(BITMAP *onto, Wobject *the_object,int from_x, int from_y, int to_x, int to_y);\n");
    fprintf(outfile, "\n");
    fprintf(outfile, "};\n");
    fprintf(outfile, "\n");
    fprintf(outfile, "#endif // end of headerfile\n");
    fclose(outfile);
    if (!(outfile = fopen(sourcename,"w")))
    {
        printf("unable to open file %s",sourcename);
        exit(1);
    }
    fprintf(outfile, "#include %c%s%c\n",34,headername,34);
    fprintf(outfile, "%s::%s(int size_x, int size_y,int pos_x,int pos_y,  char *name)\n",argv[1],argv[1]);
    if (Base == WOBJECT)
    {
        fprintf(outfile, "      : Wobject(name,size_x,size_y,pos_x,pos_y,%s::draw_border)\n",argv[1]);
    }
    else if (Base == WINDOW)
    {
        fprintf(outfile, "      : Window(name,size_x,size_y,pos_x,pos_y,%s::draw_border)\n",argv[1]);
    }
    fprintf(outfile, "{\n");
    fprintf(outfile, " //ADD OBJECT INITIALISING CODE HERE\n");
    fprintf(outfile, "}\n");

    fprintf(outfile, "\n");
    if (Base == WOBJECT)
    {
        fprintf(outfile, "// DRAW FUNCTION, ALTER THIS CODE SO THAT IT THE PART OF YOUR OBJECT \n");
        fprintf(outfile, "// WITHIN THE RECTANGLE SUPPLIED(COORDINATES ARE INCLUSIVE)\n");

        fprintf(outfile, "void %s::draw(BITMAP *onto, int f_x, int f_y,int t_x, int t_y)\n"
                         "{\n"
                      "      //EXAMPLE, CODE REPLACE WITH OWN CODE\n"
                      "      f_x = MAX(f_x,get_geoinfo()->pos_x);\n"
                      "      f_y = MAX(f_y,get_geoinfo()->pos_y);\n"
                      "      t_x = MIN(t_x,get_geoinfo()->pos_x + get_geoinfo()->screen_size_x);\n"
                     "      t_y = MIN(t_y,get_geoinfo()->pos_y + get_geoinfo()->screen_size_y);\n"
                     "\n"
                     "      if (f_x <= t_x && f_y <= t_y)\n"
                     "          rectfill(onto,f_x,f_y,t_x,t_y,128);\n"
                     "      // END EXAMPLE CODE\n"
                     "}\n\n",argv[1]
            );
    }

   fprintf(outfile, "\n");
   fprintf(outfile, "// IF NO BORDER USED REMOVE FUNCTION BELOW\n");
   fprintf(outfile,"int %s::draw_border(BITMAP *onto,Wobject *the_object,int from_x,int from_y,int to_x,int to_y)\n"
                   "{\n"
                   "    //EXAMPLE CODE, REPLACE THIS WITH OWN CODE\n"
                   "    int\n"
                   "        f_x,f_y,t_x,t_y;\n"
                   "\n"
                   "    if (!onto) // return border dimensions to manager\n"
                   "    {\n"
                   "        switch(from_x)\n"
                   "        {\n"
                   "            case WINDOW_UPPER_BORDER:\n"
                   "                return 10;\n"
                   "            break;\n"
                   "            case WINDOW_LOWER_BORDER:\n"
                   "                return 10;\n"
                   "            break;\n"
                   "            case WINDOW_LEFT_BORDER:\n"
                   "                return 10;\n"
                   "            break;\n"
                   "            case WINDOW_RIGHT_BORDER:\n"
                   "                return 10;\n"
                   "            break;\n"
                   "        }\n"
                   "\n"
                   "    }\n"
                   "\n"
                   "    GeoInfo const *g = the_object->get_geoinfo();\n"
                   "\n"
                   "    //upper border\n"
                   "    f_x = g->pos_x - 10;\n"
                   "    t_x = g->pos_x + g->screen_size_x + 9;\n"
                   "    f_y = g->pos_y - 10;      \n"
                   "    t_y = g->pos_y;           \n"
                   "                              \n"
                   "    f_x = MAX(f_x, from_x);   \n"
                   "    t_x = MIN(t_x, to_x);     \n"
                   "    f_y = MAX(f_y, from_y);   \n"
                   "    t_y = MIN(t_y, to_y);     \n"
                   "    if (f_x <= t_x && f_y <= t_y) \n"
                   "        rectfill(onto,f_x,f_y,t_x,t_y,255);\n"
                   "                                           \n"
                   "    //lower border                         \n"
                   "    f_x = g->pos_x - 10;                   \n"
                   "    t_x = g->pos_x + g->screen_size_x + 9; \n"
                   "    f_y = g->pos_y + g->screen_size_y;     \n"
                   "    t_y = g->pos_y + g->screen_size_y + 9; \n"
                   "                                           \n"
                   "    f_x = MAX(f_x, from_x);                \n"
                   "    t_x = MIN(t_x, to_x);                  \n"
                   "    f_y = MAX(f_y, from_y);                \n"
                   "    t_y = MIN(t_y, to_y);                  \n"
                   "    if (f_x <= t_x && f_y <= t_y)          \n"
                   "        rectfill(onto,f_x,f_y,t_x,t_y,255);\n"
                   "                                           \n"
                   "    //left border                          \n"
                   "    f_x = g->pos_x - 10;                   \n"
                   "    t_x = g->pos_x;                        \n"
                   "    f_y = g->pos_y;                        \n"
                   "    t_y = g->pos_y + g->screen_size_y;     \n"
                   "                                           \n"
                   "    f_x = MAX(f_x, from_x);                \n"
                   "    t_x = MIN(t_x, to_x);                  \n"
                   "    f_y = MAX(f_y, from_y);                \n"
                   "    t_y = MIN(t_y, to_y);                  \n"
                   "    if (f_x <= t_x && f_y <= t_y)          \n"
                   "        rectfill(onto,f_x,f_y,t_x,t_y,255);\n"
                   "                                           \n"
                   "    //right border                         \n"
                   "    f_x = g->pos_x + g->screen_size_x;     \n"
                   "    t_x = g->pos_x + g->screen_size_x + 9; \n"
                   "    f_y = g->pos_y;                        \n"
                   "    t_y = g->pos_y + g->screen_size_y;     \n"
                   "                                           \n"
                   "    f_x = MAX(f_x, from_x);                \n"
                   "    t_x = MIN(t_x, to_x);                  \n"
                   "    f_y = MAX(f_y, from_y);                \n"
                   "    t_y = MIN(t_y, to_y);                  \n"
                   "    if (f_x <= t_x && f_y <= t_y)          \n"
                   "        rectfill(onto,f_x,f_y,t_x,t_y,255);\n"
                   "                                           \n"
                   "                                           \n"
                   "    //END OF EXAMPLE CODE\n"
                   "    return 0; //returnvalue currently ignored \n"
                   "}\n",argv[1]                               
            );                                                 

    fprintf(outfile, "// MOUSE FUNCTION EDIT FOR DESIRED ACTIONS\n");

    fprintf(outfile,"Wobject *%s::process_mouse(int *x_out = NULL, int *y_out = NULL)\n",argv[1]);
    fprintf(outfile,"{\n");
    fprintf(outfile,"    Wobject *win; \n");
    fprintf(outfile,"    int out_x,out_y,mouse_was_x,mouse_was_y,pos_was_x,pos_was_y,mouse_is_x,mouse_is_y,size_was_x,size_was_y; // helper variables\n");
    fprintf(outfile,"    int flag; // is the border clicked?\n");
    fprintf(outfile,"    GeoInfo const *g; // to store geoinfo from object in\n");
    fprintf(outfile,"\n");
    fprintf(outfile,"    //get the object the mouse is currently in( if flag is set it is in a border)\n");
    fprintf(outfile,"    //out_x, out_y will be the coordinates relative to that object \n");
    fprintf(outfile,"    win =WHAM->process_coordinates(mouse_x,mouse_y,&out_x,&out_y,&flag);\n");
    fprintf(outfile,"\n");
    fprintf(outfile,"    //store coordinates relative to this object if the user wants them\n");
    fprintf(outfile,"    if (x_out)\n");
    fprintf(outfile,"        *x_out = out_x;\n");
    fprintf(outfile,"    if (y_out)\n");
    fprintf(outfile,"        *y_out = out_y;\n");
    fprintf(outfile,"    // check if got the right object , it could be wrong if the mouse moved very fast and moved out of the object since Wham_secretary.process_mouse  \n");
    fprintf(outfile,"    if (win != this) return win;\n");
    fprintf(outfile,"\n");
    fprintf(outfile,"    if (flag)\n");
    fprintf(outfile,"    {\n");
    fprintf(outfile,"       if (out_x <0 || out_x <0)\n");
    fprintf(outfile,"          set_Mouse_Cursor(WinSystem::Mouse_Cursor1);\n");
    fprintf(outfile,"       else\n");
    fprintf(outfile,"          set_Mouse_Cursor(WinSystem::Mouse_Cursor2);\n");
    fprintf(outfile,"    }\n");
    fprintf(outfile,"    else\n");
    fprintf(outfile,"       set_Mouse_Cursor(WinSystem::Mouse_Cursor0);\n");
    fprintf(outfile,"\n");
    fprintf(outfile,"    // mouse button 1 is currently pressed down\n");
    fprintf(outfile,"    if (mouse_b & 1)\n");
    fprintf(outfile,"    {\n");
    fprintf(outfile,"       g = win->get_geometric_info(); // get info on this object into g \n");
    fprintf(outfile,"       mouse_was_x = mouse_x; //store current mouse position\n");
    fprintf(outfile,"       mouse_was_y = mouse_y;\n");
    fprintf(outfile,"       pos_was_x = g->pos_x; //store current object location \n");
    fprintf(outfile,"       pos_was_y = g->pos_y;\n");
    fprintf(outfile,"       size_was_x = g->screen_size_x;//store current object size\n");
    fprintf(outfile,"       size_was_y = g->screen_size_y;\n");
    fprintf(outfile,"\n");   
    fprintf(outfile,"       //if the window has a parent manager, the parent window could be rescaled   \n");
    fprintf(outfile,"       //so calc the mouse position relative to the parent window\n");
    fprintf(outfile,"       if (win->get_manager()->get_parent())\n");
    fprintf(outfile,"           win->get_manager()->get_parent()->transform_coordinates(&mouse_was_x,&mouse_was_y);\n");
    fprintf(outfile,"\n");   
    fprintf(outfile,"       if (flag) // mouse is on a border \n");
    fprintf(outfile,"       {\n");
    fprintf(outfile,"           if (out_x < 0 || out_y <0) // left or upper border -> move\n");
    fprintf(outfile,"           {\n");
    fprintf(outfile,"               while (mouse_b & 1)\n");
    fprintf(outfile,"               {\n");
    fprintf(outfile,"                   mouse_is_x = mouse_x; // get current mouse pos\n");
    fprintf(outfile,"                   mouse_is_y = mouse_y;\n");
    fprintf(outfile,"                   // calc current mouse pos relative to parent win\n");
    fprintf(outfile,"                   if (win->get_manager()->get_parent())\n");
    fprintf(outfile,"                       win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);\n");
    fprintf(outfile,"                   // and move the window\n");
    fprintf(outfile,"                   win->move(pos_was_x + mouse_is_x - mouse_was_x,pos_was_y + mouse_is_y - mouse_was_y);\n");
    fprintf(outfile,"               }\n");
    fprintf(outfile,"           }\n");
    fprintf(outfile,"           else // left or lower border -> rescale\n");
    fprintf(outfile,"           {\n");
    fprintf(outfile,"               while (mouse_b & 1)\n");
    fprintf(outfile,"               {\n");
    fprintf(outfile,"                   mouse_is_x = mouse_x;// get current mosue pos\n");
    fprintf(outfile,"                   mouse_is_y = mouse_y;\n");
    fprintf(outfile,"                   // make mouse pos relative to parent window\n");
    fprintf(outfile,"                   if (win->get_manager()->get_parent())\n");
    fprintf(outfile,"                       win->get_manager()->get_parent()->transform_coordinates(&mouse_is_x,&mouse_is_y);\n");
    fprintf(outfile,"                   //rescale the object\n");
    fprintf(outfile,"                   win->rescale(size_was_x + mouse_is_x - mouse_was_x,size_was_y + mouse_is_y - mouse_was_y);\n");
    fprintf(outfile,"               }\n");
    fprintf(outfile,"           }\n");
    fprintf(outfile,"       }\n");
    fprintf(outfile,"       else // flag =0 -> clicked in the window\n");
    fprintf(outfile,"            win->raise();\n");
    fprintf(outfile,"    }\n");

    fprintf(outfile,"    if (!flag) \n");
    fprintf(outfile,"        return (Wobject *)win;\n");
    fprintf(outfile,"    else\n");
    fprintf(outfile,"        return NULL;\n");
    fprintf(outfile,"}\n");
    fprintf(outfile,"\n");




    fclose(outfile);
}
END_OF_MAIN();
