# makefile for gnu make for the WHAM windowing library
# requires DJGPPv2.02+ and Allegro3.1+

#for yamd memory debugger use, and change flags
#CXX=yamd-gcc
CXX = gcc

COREOBJS = obj/winsys.o obj/wobject.o obj/helpers.o obj/window.o obj/stack.o obj/secretar.o obj/boxlist.o obj/scrolbar.o obj/geoinfo.o obj/scrsav.o obj/alguicon.o obj/separatr.o obj/backgrnd.o obj/panel.o obj/whammenu.o
OBJS = $(COREOBJS) obj/whamwin.o obj/bordered.o obj/billwin.o obj/frame.o obj/whaminit.o obj/framewin.o obj/pager.o obj/sunglass.o obj/oclock.o obj/wtext.o
CFLAGS = -O3 -Wall -ffast-math  -finline-functions -I./include  -g -Werror -Wno-deprecated

.PHONY :examples clean

#Debugging flags
#uncomment this if you want to be able to see the hiding/displaying of windows (only works in unbuffered mode)
#CFLAGS += -DSLOW
#shows internal function trace of the library
#CFLAGS += -DWTRACE
#to debug the boxlist algorithm (try this one in unbuffered mode :)
#CFLAGS += -DDEBUGBOXLIST
#to debug message system
#CFLAGS += -DDEBUGMESSAGESYS


#headers used by all sourcefiles
BASEHEADERS = include/winsys.h include/wobject.h include/geoinfo.h include/helpers.h include/boxlist.h include/topinc.h include/wversion.h include/stack.h

# headers used by sourcefiles derived from Window
BASEHEADERSW = include/winsys.h include/wobject.h include/geoinfo.h include/helpers.h include/window.h include/topinc.h include/wversion.h include/stack.h

#headers used by sourcefiles derived from BillWin
BASEHEADERSB = include/winsys.h include/wobject.h include/geoinfo.h include/helpers.h include/billwin.h include/topinc.h include/wversion.h /include/stack.h


all: lib/libwham.a examples

lib/libwham.a : $(OBJS)
	ar -rs ./lib/libwham.a $(OBJS)   
	make -f wham.mak core	

examples:
	make -C examples

core: $(COREOBJS)
	ar -rs ./lib/libwcore.a $(COREOBJS)
   
obj/pager.o: src/pager.cc include/pager.h $(BASEHEADERS)
	$(CXX) -c -o obj/pager.o $(CFLAGS) src/pager.cc

obj/panel.o: src/panel.cc include/panel.h $(BASEHEADERSW)
	$(CXX) -c -o obj/panel.o $(CFLAGS) src/panel.cc

obj/framewin.o: src/framewin.cc include/framewin.h $(BASEHEADERSW)
	$(CXX) -c -o obj/framewin.o $(CFLAGS) src/framewin.cc

obj/whaminit.o: src/whaminit.cc include/whaminit.h include/billwin.h include/bordered.h \
	include/window.h include/whamwin.h $(BASEHEADERS)
	$(CXX) -c -o obj/whaminit.o $(CFLAGS) src/whaminit.cc

obj/billwin.o: src/billwin.cc include/billwin.h $(BASEHEADERS)
	$(CXX) -c -o obj/billwin.o $(CFLAGS) src/billwin.cc

obj/alguicon.o: src/alguicon.cc include/alguicon.h $(BASEHEADERSW)
	$(CXX) -c -o obj/alguicon.o $(CFLAGS) src/alguicon.cc

obj/secretar.o: src/secretar.cc include/secretar.h include/scrsav.h $(BASEHEADERS)
	$(CXX) -o obj/secretar.o $(CFLAGS) -c src/secretar.cc

obj/frame.o: src/frame.cc include/frame.h $(BASEHEADERS)
	$(CXX) -o obj/frame.o $(CFLAGS) -c src/frame.cc

obj/window.o: src/window.cc include/window.h $(BASEHEADERS)
	$(CXX) -o obj/window.o $(CFLAGS) -c src/window.cc

obj/winsys.o : src/winsys.cc $(BASEHEADERS)
	$(CXX) -o obj/winsys.o $(CFLAGS) -c src/winsys.cc
   
obj/wobject.o : src/wobject.cc $(BASEHEADERS)
	$(CXX) -o obj/wobject.o $(CFLAGS) -c src/wobject.cc

obj/helpers.o : src/helpers.cc $(BASEHEADERS)
	$(CXX) -o obj/helpers.o $(CFLAGS) -c src/helpers.cc

obj/whamwin.o : src/whamwin.cc include/whamwin.h $(BASEHEADERSW)
	$(CXX) -o obj/whamwin.o $(CFLAGS) -c src/whamwin.cc

obj/bordered.o: src/bordered.cc include/bordered.h $(BASEHEADERSW)
	$(CXX) -o obj/bordered.o $(CFLAGS) -c src/bordered.cc

obj/boxlist.o : src/boxlist.cc $(BASEHEADERS)
	$(CXX) -o obj/boxlist.o $(CFLAGS) -c src/boxlist.cc

obj/scrolbar.o : src/scrolbar.cc include/scrolbar.h $(BASEHEADERS)
	$(CXX) -o obj/scrolbar.o $(CFLAGS) -c src/scrolbar.cc

obj/sunglass.o : src/sunglass.cc include/sunglass.h $(BASEHEADERS)
	$(CXX) -o obj/sunglass.o $(CFLAGS) -c src/sunglass.cc

obj/oclock.o : src/oclock.cc include/oclock.h $(BASEHEADERS)
	$(CXX) -o obj/oclock.o $(CFLAGS) -c src/oclock.cc

obj/geoinfo.o: src/geoinfo.cc include/geoinfo.h $(BASEHEADERS)
	$(CXX) -o obj/geoinfo.o $(CFLAGS) -c src/geoinfo.cc

obj/scrsav.o: src/scrsav.cc include/screensv.h include/scrsav.h include/wversion.h
	$(CXX) -o obj/scrsav.o $(CFLAGS) -c src/scrsav.cc

obj/wtext.o: src/wtext.cc include/wtext.h $(BASEHEADERS)
	$(CXX) -c -o obj/wtext.o $(CFLAGS) src/wtext.cc

obj/stack.o: src/stack.cc include/stack.h $(BASEHEADERS)
	$(CXX) -c -o obj/stack.o $(CFLAGS) src/stack.cc

obj/separatr.o : src/separatr.cc include/separatr.h $(BASEHEADERS)
	$(CXX) -c -o obj/separatr.o $(CFLAGS) src/separatr.cc   

obj/backgrnd.o : src/backgrnd.cc include/backgrnd.h $(BASEHEADERS)
	$(CXX) -c -o obj/backgrnd.o $(CFLAGS) src/backgrnd.cc

obj/whammenu.o : src/whammenu.cc include/whammenu.h $(BASEHEADERS)
	$(CXX) -c -o obj/whammenu.o $(CFLAGS) src/whammenu.cc
